/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  display.c  -  Display class
 *
 *  (c) 2001-2004 Carsten Siebholz <c.siebholz AT t-online de>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <ctype.h>
#include <unistd.h>
#include <sys/io.h>
#include <sys/time.h>

#include <algorithm>
using namespace std;

#include "display.h"
#include "setup.h"
#include "global.h"
#include "strfct.h"
#include "i18n.h"

#include <vdr/tools.h>
#include <vdr/menu.h>

#define MAXLINES_MSG  4
#define MAXLINES_TEXT 16
#define FILENAME_EXTERNAL_TRIGGERED_SYMBOLS "/tmp/graphlcd_symbols"


// tiny:     0..48
#define MINY_T  0
#define MAXY_T  48

// small:   49..61
#define MINY_S  (MAXY_T+1)
#define MAXY_S  61

// medium:  62..127
#define MINY_M  (MAXY_S+1)
#define MAXY_M  127

// large:  128..
#define MINY_L  (MAXY_M+1)
#define MAXY_L  9999



int     FRAME_SPACE_X;
int     FRAME_SPACE_XB;
int     FRAME_SPACE_Y;
int     FRAME_SPACE_YB;
int     TEXT_OFFSET_X;
int     TEXT_OFFSET_Y_TIME;
int     TEXT_OFFSET_Y_CHANNEL;
int     TEXT_OFFSET_Y_TITLE;
int     SYMBOL_SPACE;
int     TIMEBAR_HEIGHT;



cGraphLCDDisplay::cGraphLCDDisplay()
{
  active = false;

  pLCDDevice  = NULL;
  pGraLib     = NULL;
  largeFont = NULL;
  normalFont = NULL;
  smallFont = NULL;
  symbols = NULL;

  cfgDir = "";
  fontDir = "";
  logoDir = "";

  CurrTime = time(NULL);
  LastTime = CurrTime-58;
  CurrTimeval.tv_sec = 0;
  CurrTimeval.tv_usec = 0;
  timerclear(&UpdateAt);
  LastTimeCheckSym = CurrTime;

  State = Normal;
  LastState = Normal;

  currentChannel = "";
  currentChannelTmp = "";
  currentChannelNr = 0;

  presentTime = 0;
  followingTime = 0;
  presentTitle = "";
  presentSubtitle = "";
  followingTitle = "";
  followingSubtitle = "";

  logo = NULL;
  logoList = NULL;

  replayName = "";
  replayControl = NULL;
  replayCurrent = 0;
  replayCurrentLast = FRAMESPERSEC;
  replayTotal = 0;
  replayProgress = 0;
  replayProgressLast = 1;
  replayMode = eReplayNormal;

  osdTitle = "";
  osdCurrentItem = "";
  menuTop = 0; menuCurrent = 0; menuCount = 0;

  osdMessage = "";

  osdTextItemComplete = "";

  for (int i = 0; i < LCDMAXCARDS; i++)
  {
    cardIsRecording[i] = 0;
    cardNameRecording[i] = "";
  }
  recCounter1 = 0;
  recCounter2 = 0;

  iAnzTabs = 0;
  for(int i=0; i<MAX_TABS; i++) iTab[i] = 0;
  iMaxTab[0] = 0;
  iMaxTab[1] = 66;
  iMaxTab[2] = 100;
  iMaxTab[3] = 100;
  iMaxTab[4] = 100;
  iMaxTab[5] = 100;
  iMaxTab[6] = 100;
  iMaxTab[7] = 100;
  iMaxTab[8] = 100;
  iMaxTab[9] = 100;

  for (int i = 0; i < 4; i++)
    osdColorButton[i] = "";

  strcpy (szETSymbols, "");

  iVolume = -1;
  iLastVolChange = -1;
}


cGraphLCDDisplay::~cGraphLCDDisplay()
{
  active = false;
  Cancel(3);

  if (pLCDDevice)
    delete pLCDDevice;

  if (pGraLib)
    delete pGraLib;

  if (largeFont)
    delete largeFont;
  if (normalFont)
    delete normalFont;
  if (smallFont)
    delete smallFont;
  if (symbols) delete symbols;

  if (logoList)
    delete logoList;
}


int cGraphLCDDisplay::Init(const char * CfgDir)
{
  string font_l;
  string font_n;
  string font_s;
  string font_sym;

  if (!CfgDir)
    return 2;
  cfgDir = CfgDir;
  fontDir = cfgDir + "/fonts";
  logoDir = cfgDir + "/logos";

  pLCDDevice = CreateDriver(GraphLCDSetup.DriverID, &GraphLCDSetup);


  if (!pLCDDevice)
  {
    esyslog("graphlcd plugin: ERROR loading device ID %d\n",
        GraphLCDSetup.DriverID);
    return 2;
  }

  pGraLib = new cGraphLCDGraphLib(pLCDDevice);
  if (!pGraLib)
  {
    esyslog("graphlcd plugin: ERROR loading graphlib\n");
    return 2;
  }

  if (pLCDDevice->SizeY() <= MAXY_T)
  {
    font_l = fontDir + "/f_t_l.fnt";
    font_n = fontDir + "/f_t_n.fnt";
    font_s = fontDir + "/f_t_s.fnt";
    font_sym = fontDir + "/f_t_sym.fnt";
  }
  else if (pLCDDevice->SizeY() <= MAXY_S)
  {
    font_l = fontDir + "/f_s_l.fnt";
    font_n = fontDir + "/f_s_n.fnt";
    font_s = fontDir + "/f_s_s.fnt";
    font_sym = fontDir + "/f_s_sym.fnt";
  }
  else if (pLCDDevice->SizeY() <= MAXY_M)
  {
    font_l = fontDir + "/f_m_l.fnt";
    font_n = fontDir + "/f_m_n.fnt";
    font_s = fontDir + "/f_m_s.fnt";
    font_sym = fontDir + "/f_m_sym.fnt";
  }
  else
  {
    font_l = fontDir + "/f_l_l.fnt";
    font_n = fontDir + "/f_l_n.fnt";
    font_s = fontDir + "/f_l_s.fnt";
    font_sym = fontDir + "/f_l_sym.fnt";
  }


  largeFont = new cGraphLCDFont();
  if (!largeFont->Load(font_l))
  {
    esyslog("graphlcd plugin: ERROR loading Font %s\n", font_l.c_str());
    return 1;
  }

  normalFont = new cGraphLCDFont();
  if (!normalFont->Load(font_n))
  {
    esyslog("graphlcd plugin: ERROR loading Font %s\n", font_n.c_str());
    return 1;
  }

  smallFont = new cGraphLCDFont();
  if (!smallFont->Load(font_s))
  {
    esyslog("graphlcd plugin: ERROR loading Font %s\n", font_s.c_str());
    return 1;
  }

  symbols = new cGraphLCDFont();
  if (!symbols->Load(font_sym))
  {
    esyslog("graphlcd plugin: ERROR loading Font %s\n", font_sym.c_str());
    return 1;
  }
	
  logoList = new cGraphLCDLogoList(logoDir.c_str(), cfgDir.c_str());
  if (!logoList)
  {
    esyslog("graphlcd plugin: ERROR out of memory\n");
    return 1;
  }


  if (pLCDDevice->SizeX() < 240)
  {
    FRAME_SPACE_X  = 0;
    FRAME_SPACE_XB = 1;
    TEXT_OFFSET_X  = 2;
  }
  else
  {
    FRAME_SPACE_X  = 3;
    FRAME_SPACE_XB = 3;
    TEXT_OFFSET_X  = 5;
  }

  if (pLCDDevice->SizeY() <= MAXY_T)
  {
    // very small display
    FRAME_SPACE_Y         = 0;
    FRAME_SPACE_YB        = 1;
    TEXT_OFFSET_Y_TIME    = 1;
    TEXT_OFFSET_Y_CHANNEL = 1;
    TEXT_OFFSET_Y_TITLE   = 1;
    SYMBOL_SPACE          = 1;
    TIMEBAR_HEIGHT        = 3;
  }
  else if (pLCDDevice->SizeY() <= MAXY_S)
  {
    // small display
    FRAME_SPACE_Y         = 0;
    FRAME_SPACE_YB        = 1;
    TEXT_OFFSET_Y_TIME    = 1;
    TEXT_OFFSET_Y_CHANNEL = 3;
    TEXT_OFFSET_Y_TITLE   = 1;
    SYMBOL_SPACE          = 1;
    TIMEBAR_HEIGHT        = 3;
  }
  else if (pLCDDevice->SizeY() <= MAXY_M)
  {
    // medium display
    FRAME_SPACE_Y         = 0;
    FRAME_SPACE_YB        = 1;
    TEXT_OFFSET_Y_TIME    = 1;
    TEXT_OFFSET_Y_CHANNEL = 3;
    TEXT_OFFSET_Y_TITLE   = 3;
    SYMBOL_SPACE          = 1;
    TIMEBAR_HEIGHT        = 3;
  }
  else
  {
    // large display
    FRAME_SPACE_Y         = 3;
    FRAME_SPACE_YB        = 3;
    TEXT_OFFSET_Y_TIME    = 5;
    TEXT_OFFSET_Y_CHANNEL = 8;
    TEXT_OFFSET_Y_TITLE   = 7;
    SYMBOL_SPACE          = 2;
    TIMEBAR_HEIGHT        = 5;
  }

  Start();
  return 0;
}


void
cGraphLCDDisplay::Action(void) {
  dsyslog("graphlcd plugin: Display update thread started (pid=%d)", getpid());

  if(pLCDDevice && pGraLib) {
    if(pLCDDevice->Init()==0) {

      active = true;
      while(active) {
        if(GraphLCDSetup.PluginActive) {
          CurrTime = time(NULL);

          if(timerisset(&UpdateAt)) {
            // timed Update enabled
            if (gettimeofday(&CurrTimeval, NULL) == 0) {  // get current time
              if(CurrTimeval.tv_sec > UpdateAt.tv_sec) {
                timerclear(&UpdateAt);
                update = true;
//dsyslog("##### UpdateIn 1");
              } else if((CurrTimeval.tv_sec == UpdateAt.tv_sec) &&
                        (CurrTimeval.tv_usec > UpdateAt.tv_usec)) {
                timerclear(&UpdateAt);
                update = true;
//dsyslog("##### UpdateIn 2");
              }
            }
          }
          if (GraphLCDSetup.ShowVolume && !update && iLastVolChange > 0)
          {
            if (time_ms() - iLastVolChange > 2000)
            {
              iLastVolChange = 0;
              update = true;
            }
          }

          switch(State) {
            case Normal:
              // check and update external triggered symbols
              if(GraphLCDSetup.ShowETSymbols) {
                if(CurrTime != LastTimeCheckSym) {
                  update |= CheckAndUpdateSymbols();
                  LastTimeCheckSym = CurrTime;
                }
              }

              {
                std::vector<cScroller>::iterator i = scroller.begin();
                std::vector<cScroller>::const_iterator e = scroller.end();
                for(; e != i; ++i )
                {
                  if (i->NeedsUpdate())
                    update = true;
                }
              }

              // update Display if animated Logo is present, and an update is necessary
              if (!update 
                && IsLogoActive() 
                && logo->Cnt() > 1 
                && (time_ms() - logo->LastChange() >= logo->Delay()))
                  update = true;
              
              // update Display every minute or due to an update
              if((CurrTime/60 != LastTime/60) ||
                 update) {
                if(CurrTime/60 != LastTime/60) GetProgramme();
                timerclear(&UpdateAt);
                update = false;

                pLCDDevice->Clear();
                DisplayTime();
                DisplayLogo();
                DisplayChannel();
                DisplaySymbols();
                DisplayProgramm();
                DisplayVolume();
                DisplayMessage();
                //pLCDDevice->Refresh(CurrTime/60 != LastTime/60);
                pLCDDevice->Refresh(false);
                LastTime = CurrTime;
              } else {
                usleep(100000);
              }
              break;

            case Replay:
              if(replayControl) {
                if(replayControl->GetIndex(replayCurrent, replayTotal, false)) {
                  replayTotal=(replayTotal==0)?1:replayTotal;
                  replayProgress = replayCurrent*(pLCDDevice->SizeX()-2*FRAME_SPACE_X-1)/replayTotal;

                  {
                    update = false;
                    std::vector<cScroller>::iterator i = scroller.begin();
                    std::vector<cScroller>::const_iterator e = scroller.end();
                    for(; e != i; ++i )
                    {
                      if (i->NeedsUpdate())
                        update = true;
                    }
                  }

                  // update Display if animated Logo is present, and an update is necessary
                  if (!update 
                    && IsLogoActive() 
                    && logo->Cnt() > 1 
                    && (time_ms() - logo->LastChange() >= logo->Delay()))
                      update = true;

                  // update Display every second or due to an update
                  if((CurrTime != LastTime) ||
                     update) {
                    // but only, if something has changed
                    if((replayProgress != replayProgressLast) ||
                       (replayCurrent/FRAMESPERSEC != replayCurrentLast/FRAMESPERSEC) ||
                        update) {
                      timerclear(&UpdateAt);
                      update = false;

                      pLCDDevice->Clear();
                      DisplayTime();
                      DisplayLogo();
                      DisplayReplay(replayCurrent, replayTotal);
//XXX                      DisplaySymbols();
                      DisplayVolume();
                      DisplayMessage();
                      pLCDDevice->Refresh(false);
                      replayCurrentLast = replayCurrent;
                      replayProgressLast = replayProgress;
                      LastTime = CurrTime;
                    } else {
                      usleep(100000);
                    }
                  } else {
                    usleep(100000);
                  }
                } else {
                  replayControl = NULL;
                  State = Normal;
                  Update();
                }
              } else {
                State = Normal;
                Update();
              }
              break;

            case Menu:
              if(GraphLCDSetup.ShowMenu) {
                // update Display every minute or due to an update
                if((CurrTime/60 != LastTime/60) ||
                   update) {
                  timerclear(&UpdateAt);
                  update = false;

                  pLCDDevice->Clear();
                  DisplayTime();
                  DisplayMenu();
                  DisplayTextItem();
                  DisplayVolume();
                  DisplayMessage();
                  DisplayColorButtons();
                  pLCDDevice->Refresh(false);
                  LastTime = CurrTime;
                } else {
                  usleep(100000);
                }
              } else {
                osdTitle = "";
                osdItems.clear();

                State = LastState;
                // activate delayed Update
                UpdateIn(100000);   // XXX ???
              }
              break;

            default:
              break;
          }
        } else {
          usleep(100000);
        }
      }
      pLCDDevice->DeInit();
    }
  }
  dsyslog("graphlcd plugin: Display update thread ended (pid=%d)", getpid());
}


void cGraphLCDDisplay::SetChannel(int ChannelNumber)
{
  char tmp[16];

  if (ChannelNumber > 0)
  {
    currentChannelNr = ChannelNumber;

    cChannel * ch = Channels.GetByNumber(currentChannelNr);
    currentChannelID = ch->GetChannelID();

    sprintf(tmp, "%d ", currentChannelNr);
    currentChannel = tmp;
    currentChannel += ch->Name();

    if (GraphLCDSetup.ShowLogo)
    {
      ePicType picType;

      switch (GraphLCDSetup.ShowLogo)
      {
        case 1:   // auto
          if (pLCDDevice->SizeY() <= MAXY_M)
            picType = LOGO_MEDIUM;
          else
            picType = LOGO_LARGE;
          break;
        case 2:   // medium
          picType = LOGO_MEDIUM;
          break;
        case 3:   // large
          picType = LOGO_LARGE;
          break;
        default:  // should not happen at the moment !!
          picType = LOGO_SMALL;
          break;
      }
#if VDRVERSNUM >= 10300
      char * strId = strstr(ch->GetChannelID().ToString(), "-") + 1;
      logo = logoList->GetLogo(currentChannelNr, strId, picType);
#else
	  char strId[16];
	  sprintf(strId, "%d", ch->Sid());
      logo = logoList->GetLogo(currentChannelNr, strId, picType);
#endif
      if (logo)
        logo->First(time_ms());
    }
    else
    {
      logo = NULL;
    }

    presentTime = 0;
    followingTime = 0;
    GetProgramme();
    Update();
  }
}


void cGraphLCDDisplay::SetChannelText(const char * Text)
{
  if (Text)
  {
    currentChannelTmp = compactspace(string(Text));
    Update();
  }
  else
    currentChannelTmp = "";
}


void cGraphLCDDisplay::SetProgramme(time_t PresentTime, const char * PresentTitle, const char * PresentSubtitle,
                                    time_t FollowingTime, const char * FollowingTitle, const char * FollowingSubtitle)
{
  presentTime = PresentTime;
  presentTitle = "";
  if (!isempty(PresentTitle))
    presentTitle = PresentTitle;
  presentSubtitle = "";
  if (!isempty(PresentSubtitle))
    presentSubtitle = PresentSubtitle;

  followingTime = FollowingTime;
  followingTitle = "";
  if (!isempty(FollowingTitle))
    followingTitle = FollowingTitle;
  followingSubtitle = "";
  if (!isempty(FollowingSubtitle))
    followingSubtitle = FollowingSubtitle;

  Update();
}


void cGraphLCDDisplay::SetClear()
{
  currentChannelTmp = "";

  osdTitle = "";
  osdItems.clear();

  osdMessage = "";   // clear message

  osdTextItemComplete = "";
  osdTextItemLines.clear();

  for (int i = 0; i < 4; i++)
    osdColorButton[i] = "";

  iAnzTabs = 0;
  for (int i = 0; i < MAX_TABS; i++)
    iTab[i] = 0;

  if (State == Menu)
  {
    State = LastState;
    // activate delayed Update
    UpdateIn(100000);
  }
  else
  {
    Update();
  }
}


void cGraphLCDDisplay::SetOsdTitle(const char * Title)
{
  osdMessage = "";   // clear message
  osdTitle = "";
  if (Title)
  {
    UpdateIn(0);  // stop delayed Update
    osdTitle = Title;
    // remove the time
    int pos = osdTitle.find('\t');
    if (pos != -1)
      osdTitle.resize(pos);
    osdTitle = compactspace(osdTitle);

    if (State != Menu)
    {
      menuTop = 0;
      LastState = State;
      State = Menu;
    }
    // activate delayed Update
    UpdateIn(100000);
  }
}


void cGraphLCDDisplay::SetOsdItem(const char * Text, int Index)
{
  int iAT, t;
  string str;
  int pos1, pos2;

  osdMessage = "";   // clear message

  if (Text)
  {
    UpdateIn(0);  // stop delayed Update
    str = trim(string(Text));
    osdItems.push_back(str);
    pos1 = 0;
    pos2 = str.find('\t');
    iAT = 0;
    while (pos1 < (int) str.length() && pos2 != -1)
    {
      iAT++;
      t = min(normalFont->WidthS(str.c_str() + pos1, pos2 - pos1), (iMaxTab[iAT] * pLCDDevice->SizeX()) / 100);
      iTab[iAT] = max(iTab[iAT], t);
      iAnzTabs = max(iAnzTabs, iAT);
      pos1 = pos2 + 1;
      pos2 = str.find('\t', pos1);
    }
  }
}


void cGraphLCDDisplay::SetOsdCurrentItem(const char * Text)
{
  int tabs;
  int pos;

  osdMessage = "";   // clear message

  osdCurrentItem = "";
  if (Text)
  {
    UpdateIn(0);  // stop delayed Update
    osdCurrentItem = trim(string(Text));

    // count nr of tabs in text
    tabs = 0;
    for (unsigned int i = 0; i < osdCurrentItem.length(); i++)
    {
      if (osdCurrentItem[i] == '\t')
        tabs++;
    }

    if (tabs == 1)
    {
      // only one tab => prob. Setup Menu
      pos = osdCurrentItem.find('\t');
      menuCurrent = 0;
      if (pos != -1)
      {
        for (unsigned int i = 0; i < osdItems.size(); i++)
        {
          if (osdItems[i].find(osdCurrentItem.c_str(), 0, pos) == 0)
          {
            menuCurrent = i;
            osdItems[i] = osdCurrentItem;
            break;
          }
        }
      }
    }
    else
    {
      menuCurrent = 0;
      for (unsigned int i = 0; i < osdItems.size(); i++)
      {
        if (osdItems[i].compare(osdCurrentItem) == 0)
        {
          menuCurrent = i;
          break;
        }
      }
    }

    UpdateIn(100000); //XXX
  }
}


void cGraphLCDDisplay::SetColorButtons(const char * Red, const char * Green, const char * Yellow, const char * Blue)
{

  for (int i = 0; i < 4; i++)
    osdColorButton[i] = "";

  if (Red)
    osdColorButton[0] = compactspace(string(Red));
  if (Green)
    osdColorButton[1] = compactspace(string(Green));
  if (Yellow)
    osdColorButton[2] = compactspace(string(Yellow));
  if (Blue)
    osdColorButton[3] = compactspace(string(Blue));
}


void cGraphLCDDisplay::Recording(const cDevice * Device, const char * Name)
{
  int cardIndex = Device->CardIndex();
  if (cardIndex > LCDMAXCARDS - 1 )
    cardIndex = LCDMAXCARDS-1;

  if (cardIndex < LCDMAXCARDS)
  {
    if (Name)
    {
      cardIsRecording[cardIndex]++;
      cardNameRecording[cardIndex] = Name;
    }
    else
    {
      if (cardIsRecording[cardIndex] > 0)
        cardIsRecording[cardIndex]--;
      cardNameRecording[cardIndex] = "";
    }
  }
  else
  {
    esyslog("graphlcd plugin: only up to %d devices are supported by this plugin", LCDMAXCARDS);
  }
  Update();
}


void cGraphLCDDisplay::Replaying(const cControl * Control, const char * Name)
{
  if (Name)
  {
    replayControl = (cControl *) Control;
    if (State != Menu)
    {
      State = Replay;
    }
    else
    {
      LastState = Replay;
    }
    replayMode = eReplayNormal;
    replayName = "";
    if (!isempty(Name))
    {
      if (GraphLCDSetup.IdentifyReplayType)
      {
        bool bFound = false;
        ///////////////////////////////////////////////////////////////////////
        //Looking for mp3-Plugin Replay : [LS] (444/666) title
        //
        if(strlen(Name)>6
          && *(Name+0)=='['
          && *(Name+3)==']'
          && *(Name+5)=='(')
        {
          unsigned int i;
          for(i=6;*(Name+i) != '\0';++i) //search for [xx] (xxxx) title
          {
            if(*(Name+i)==' ' && *(Name+i-1)==')')
            {
              bFound = true;
              break;
            }
          }
          if(bFound) //found MP3-Plugin replaymessage
          {
            unsigned int j;
            for(j=0;*(Name+i+j) != '\0';++j) //trim name
            {
              if(*(Name+i+j)!=' ')
                break;
            }
            
            if(strlen(Name+i+j) > 0)
            { //if name isn't empty, then copy
              replayName = Name + i + j;
            }
            else
            { //if Name empty, set fallback title
              replayName = tr("Unknown title");
            }
            replayMode = eReplayMusic;
          }
        }
        ///////////////////////////////////////////////////////////////////////
        //Looking for DVD-Plugin Replay : 1/8 4/28,  de 2/5 ac3, no 0/7,  16:9, VOLUMENAME
        //cDvdPlayerControl::GerDisplayHeaderLine
        //                         titleinfo, audiolang,  spulang, aspect, title
        if(!bFound) 
        {
          if(strlen(Name)>7)
          {
            unsigned int i,n;
            for(n=0,i=0;*(Name+i) != '\0';++i) 
            { //search volumelabel after 4*", " => xxx, xxx, xxx, xxx, title
              if(*(Name+i)==' ' && *(Name+i-1)==',') 
              {
                if(++n == 4)
                {
                  bFound = true;
                  break;
                }
              }
            }    
            if(bFound) //found DVD replaymessage
            {
              unsigned int j;bool b; 
              for(j=0;*(Name+i+j) != '\0';++j) //trim name
              {	
                if(*(Name+i+j)!=' ') 
                  break;
              }
              
              if (strlen(Name+i+j) > 0) 
              { // if name isn't empty, then copy
                replayName = Name + i + j;
                // replace all '_' with ' '
                replace(replayName.begin(), replayName.end(), '_', ' ');
                for (j = 0, b = true; j < replayName.length(); ++j) 
                {
                  // KAPITALIZE -> Kaptialize
                  if (replayName[j] == ' ')
                    b = true;
                  else if (b)
                    b = false;
                  else replayName[j] = tolower(replayName[j]);
                }
              } 
              else 
              { //if Name empty, set fallback title
                replayName = tr("Unknown title");
              }
              replayMode = eReplayDVD;
            }
          }
        }
        if (!bFound)
        {
          int i;
          for(i=strlen(Name)-1;i>0;--i) 
          { //Reversesearch last Subtitle 
            // - filename contains '~' => subdirectory
            // or filename contains '/' => subdirectory
            switch(*(Name+i))
            {
              case '/':
                replayMode = eReplayFile;
              case '~':
              {
                replayName = Name + i + 1;
                bFound = true;
                i = 0;
              }
              default:
                break;
            }
          }
        }

        if(0 == strncmp(Name,"[image] ",8)) {
          if(replayMode != eReplayFile) //if'nt already Name stripped-down as filename
            replayName = Name + 8;
          replayMode = eReplayImage;
          bFound = true;
        }
        else if(0 == strncmp(Name,"[audiocd] ",10)) {
          replayName = Name + 10;
          replayMode = eReplayAudioCD;
          bFound = true;
        }
        if (!bFound || !GraphLCDSetup.ModifyReplayString)
        {
          replayName = Name;
        }
      }
      else
      {
        replayName = Name;
      }
    }
    replayCurrentLast = FRAMESPERSEC;
    replayProgressLast = 1;

    if (GraphLCDSetup.ReplayLogo)
    {
      ePicType picType;

      switch (GraphLCDSetup.ReplayLogo)
      {
        case 1:   // auto
          if (pLCDDevice->SizeY() <= MAXY_M)
            picType = LOGO_MEDIUM;
          else
            picType = LOGO_LARGE;
          break;
        case 2:   // medium
          picType = LOGO_MEDIUM;
          break;
        case 3:   // large
          picType = LOGO_LARGE;
          break;
        default:  // should not happen at the moment !!
          picType = LOGO_SMALL;
          break;
      }
      switch(replayMode)
      {
        default:        
        case eReplayNormal : logo = logoList->GetLogo(-1, "REPLAY-VDR",   picType);break;
        case eReplayMusic  : logo = logoList->GetLogo(-2, "REPLAY-MUSIC", picType);break;
        case eReplayDVD    : logo = logoList->GetLogo(-3, "REPLAY-DVD",   picType);break;
        case eReplayFile   : logo = logoList->GetLogo(-4, "REPLAY-FILE",  picType);break;
        case eReplayImage  : logo = logoList->GetLogo(-5, "REPLAY-IMAGE", picType);break;
        case eReplayAudioCD: logo = logoList->GetLogo(-6, "REPLAY-AUDIOCD",picType);break;
      }
      if (logo)
        logo->First(time_ms());
    }
    else
    {
      logo = NULL;
    }
  }
  else
  {
    replayControl = NULL;
    if(State != Menu)
    {
      State = Normal;
      // reset Logo
      SetChannel(currentChannelNr); 
    }
    else
    {
      LastState = Normal;
    }
  }
  Update();
}


void cGraphLCDDisplay::SetStatusMessage(const char * Msg)
{
  if (Msg)
  {
    osdMessage = Msg;
    osdMessage = compactspace(osdMessage);
  }
  else
    osdMessage = "";
  Update();
}


void cGraphLCDDisplay::SetOsdTextItem(const char * Text, bool Scroll)
{
  static const char * lastText = NULL;
  if (Text)
  {
    if (osdTextItemComplete.length() == 0)
      lastText = NULL;
    if (lastText != Text)
    {
      if (osdTextItemComplete.length() > 0)
        osdTextItemComplete += " \n\n";
      osdTextItemComplete += trim(string(Text));
      string text = osdTextItemComplete;
      // replace '\n' with ' '
      for (unsigned int i = 0; i < text.length(); i++)
        if (text[i] == '\n' && (i + 1) < text.length() && text[i + 1] != '\n')
          text[i] = ' ';
      if (pLCDDevice && normalFont)
      {
        int maxTextLen = pLCDDevice->SizeX() - 2 * FRAME_SPACE_X - 2 * TEXT_OFFSET_X;
        WrapText(text, osdTextItemLines, normalFont, maxTextLen);
        osdTextItemLines.push_back("");
      }
      lastText = Text;
      osdTextItemTop = 0;
    }
  }
  else
  {
    if (Scroll)
    {
      if (osdTextItemTop > 0)
        osdTextItemTop--;
    }
    else
    {
      if (osdTextItemTop < (int) osdTextItemLines.size() - 2)
        osdTextItemTop++;
    }
  }
  UpdateIn(100000);
}


void cGraphLCDDisplay::SetVolume(int Volume, bool Absolute)
{

/* XXX
  it seems, that VDR does not send the 'Absolute' flag correctly
  (itseems to be a mute flag)
  iVolume = min(max(Absolute ? Volume : iVolume + Volume, 0), MAXVOLUME);
*/
  iVolume = Volume;

  if (iLastVolChange != -1)
  {
    iLastVolChange = time_ms();
    Update();
  }
  else
  {  // first time
    iLastVolChange = 0;
  }
}


void cGraphLCDDisplay::Update()
{
  update = true;
}


void cGraphLCDDisplay::GetProgramme()
{
#if VDRVERSNUM < 10300
  const cEventInfo * present = NULL, * following = NULL;
  cMutexLock mutexLock;
  const cSchedules * schedules = cSIProcessor::Schedules(mutexLock);
  if (currentChannelID.Valid())
  {
    if (schedules)
    {
      const cSchedule * schedule = schedules->GetSchedule(currentChannelID);
      if (schedule)
      {
        if ((present = schedule->GetPresentEvent()) != NULL)
        {
          presentTime = present->GetTime();
          presentTitle = "";
          if (!isempty(present->GetTitle()))
            presentTitle = present->GetTitle();
          presentSubtitle = "";
          if (!isempty(present->GetSubtitle()))
            presentSubtitle = present->GetSubtitle();
        }
        if ((following = schedule->GetFollowingEvent()) != NULL)
        {
          followingTime = following->GetTime();
          followingTitle = "";
          if (!isempty(following->GetTitle()))
            followingTitle = following->GetTitle();
          followingSubtitle = "";
          if (!isempty(following->GetSubtitle()))
            followingSubtitle = following->GetSubtitle();
        }
      }
    }
  }
#else
  const cEvent * present = NULL, * following = NULL;
  cSchedulesLock schedulesLock;
  const cSchedules * schedules = cSchedules::Schedules(schedulesLock);
  if (currentChannelID.Valid())
  {
    if (schedules)
    {
      const cSchedule * schedule = schedules->GetSchedule(currentChannelID);
      if (schedule)
      {
        if ((present = schedule->GetPresentEvent()) != NULL)
        {
          presentTime = present->StartTime();
          presentTitle = "";
          if (!isempty(present->Title()))
            presentTitle = present->Title();
          presentSubtitle = "";
          if (!isempty(present->ShortText()))
            presentSubtitle = present->ShortText();
        }
        if ((following = schedule->GetFollowingEvent()) != NULL)
        {
          followingTime = following->StartTime();
          followingTitle = "";
          if (!isempty(following->Title()))
            followingTitle = following->Title();
          followingSubtitle = "";
          if (!isempty(following->ShortText()))
            followingSubtitle = following->ShortText();
        }
      }
    }
  }
#endif
}


void cGraphLCDDisplay::DisplayTime()
{
  static char buffer[32];
  static char month [5];
  int FrameWidth, TextLen, yPos;
  struct tm tm_r;

  if((GraphLCDSetup.ShowDateTime==1) ||
     ((GraphLCDSetup.ShowDateTime==2)&&(State!=Menu))) {
    if(pLCDDevice && pGraLib && normalFont) {

      FrameWidth = max(pLCDDevice->SizeX()-2*FRAME_SPACE_X,1);
      if(State==Normal || State==Replay) {
        if (IsLogoActive()) { // Logo enabled & available
          FrameWidth = max(FrameWidth-FRAME_SPACE_XB-logo->Width()-2,1);
        }
        if(pLCDDevice->SizeY()<=MAXY_M) {
          // tiny, small & medium display
          if(IsSymbolsActive()) { // Symbols enabled
            FrameWidth = max(FrameWidth-FRAME_SPACE_XB-symbols->Width(),1);
          }
        }
      }
      yPos = FRAME_SPACE_Y;

      // draw Rectangle
      pGraLib->RoundRectangle(FRAME_SPACE_X, yPos,
          FRAME_SPACE_X+FrameWidth-1, yPos + normalFont->Height2() + 2*TEXT_OFFSET_Y_TIME -1,
          true, colorBlack, (TEXT_OFFSET_Y_TIME>=4) ? 4 : 1);

      if(CurrTime == 0) time(&CurrTime);
      tm *tm = localtime_r(&CurrTime, &tm_r);

      strncpy(month, (char*)(tr("JanFebMarAprMayJunJulAugSepOctNovDec") + 3*tm->tm_mon), 3);
      month[3]=0;
      snprintf(buffer, sizeof(buffer), "%s %2d.%s  %d:%02d", WeekDayName(tm->tm_wday), tm->tm_mday, month, tm->tm_hour, tm->tm_min);
      TextLen = normalFont->WidthS(buffer);

      if(TextLen>max(FrameWidth-2*TEXT_OFFSET_X,1)) {
        snprintf(buffer, sizeof(buffer), "%d.%s  %d:%02d", tm->tm_mday, month, tm->tm_hour, tm->tm_min);
        TextLen = normalFont->WidthS(buffer);
      }

      if(TextLen>max(FrameWidth-2*TEXT_OFFSET_X,1)) {
        snprintf(buffer, sizeof(buffer), "%d.%d. %d:%02d", tm->tm_mday, tm->tm_mon+1, tm->tm_hour, tm->tm_min);
        TextLen = normalFont->WidthS(buffer);
      }

      if(TextLen>max(FrameWidth-2*TEXT_OFFSET_X,1)) {
        snprintf(buffer, sizeof(buffer), "%d:%02d", tm->tm_hour, tm->tm_min);
        TextLen = normalFont->WidthS(buffer);
      }

      if(TextLen<max(FrameWidth-2*TEXT_OFFSET_X,1)) {
        pGraLib->Text(FRAME_SPACE_X+FrameWidth-TextLen-TEXT_OFFSET_X,
                   yPos+TEXT_OFFSET_Y_TIME,
                   FRAME_SPACE_X+FrameWidth-1,
                   buffer, normalFont,
                   true, colorWhite, colorBlack);
      } else {
        pGraLib->Text(FRAME_SPACE_X+TEXT_OFFSET_X,
                   yPos+TEXT_OFFSET_Y_TIME,
                   FRAME_SPACE_X+FrameWidth-1,
                   buffer, normalFont,
                   true, colorWhite, colorBlack);
      }
    }
  }
}


void cGraphLCDDisplay::DisplayChannel()
{
  int FrameWidth, yPos;

  if(GraphLCDSetup.ShowChannel) {
    if(pLCDDevice && pGraLib && normalFont) {

      FrameWidth = max(pLCDDevice->SizeX()-2*FRAME_SPACE_X,1);
      if(State==Normal) {
        if(IsLogoActive()) { // Logo enabled & available
          FrameWidth = max(FrameWidth-FRAME_SPACE_XB-logo->Width()-2,1);
        }
        if(pLCDDevice->SizeY()<=MAXY_M) {
          // tiny, small & medium display
          if(IsSymbolsActive()) { // Symbols enabled
            FrameWidth = max(FrameWidth-FRAME_SPACE_XB-symbols->Width(),1);
          }
        }
      }

      if((GraphLCDSetup.ShowDateTime==1) ||
         ((GraphLCDSetup.ShowDateTime==2)&&(State!=Menu))) {
        yPos = FRAME_SPACE_Y + normalFont->Height2() + 2*TEXT_OFFSET_Y_TIME + FRAME_SPACE_YB;

        if(pLCDDevice->SizeY()>=MINY_L) {
          // align bottom border with logo
          if(IsLogoActive()) { // Logo enabled & available
            yPos += max(0, FRAME_SPACE_Y+logo->Height()+2 - yPos -
                        (normalFont->Height2() + 2*TEXT_OFFSET_Y_CHANNEL));
          }
        }

      } else {
        yPos = FRAME_SPACE_Y;
      }

      // draw Rectangle
      pGraLib->RoundRectangle(FRAME_SPACE_X, yPos,
          FRAME_SPACE_X+FrameWidth-1, yPos + normalFont->Height2() + 2*TEXT_OFFSET_Y_CHANNEL -1,
          true, colorBlack, (TEXT_OFFSET_Y_CHANNEL>=4) ? 4 : 1);

      if (currentChannelTmp.length() > 0)
      {
        pGraLib->Text(FRAME_SPACE_X+TEXT_OFFSET_X,
            yPos+TEXT_OFFSET_Y_CHANNEL,
            FRAME_SPACE_X+FrameWidth-1,
            currentChannelTmp, normalFont,
            true, colorWhite, colorBlack);
      }
      else if (currentChannel.length() > 0)
      {
        pGraLib->Text(FRAME_SPACE_X+TEXT_OFFSET_X,
            yPos+TEXT_OFFSET_Y_CHANNEL,
            FRAME_SPACE_X+FrameWidth-1,
            currentChannel, normalFont,
            true, colorWhite, colorBlack);
      }
    }
  }
}

/** Returns true if Logo loaded and active*/
bool cGraphLCDDisplay::IsLogoActive() const
{
  if ((State==Normal && GraphLCDSetup.ShowLogo)
    || (State==Replay && GraphLCDSetup.IdentifyReplayType && GraphLCDSetup.ReplayLogo))
  {
    return logo != NULL;
  }
  return false;
}
 
void cGraphLCDDisplay::DisplayLogo()
{
  int x;
  int y;

  if (IsLogoActive() && pLCDDevice && pGraLib)
  {
    if(logo->Cnt() > 1)
    {
      int t = time_ms();
      if(t - logo->LastChange() >= logo->Delay())
      {
        if(!logo->Next(t))
          logo->First(t);
      }
    }
    
    x = max(pLCDDevice->SizeX() - FRAME_SPACE_X - logo->Width() - 2, 0);
    y = FRAME_SPACE_Y;

    pGraLib->RoundRectangle(x, y, x + logo->Width() + 1, y + logo->Height() + 1, false, colorBlack, 1);
    pGraLib->Bitmap(x + 1, y + 1, logo, colorBlack);
  }
}

/** Returns true if Symbols loaded and active*/
bool cGraphLCDDisplay::IsSymbolsActive() const
{
  return GraphLCDSetup.ShowSymbols 
      && symbols != NULL;
}

void cGraphLCDDisplay::DisplaySymbols()
{
  int yPos = 0;
  int xPos = 0;
  int i;

  if (IsSymbolsActive())
  {
    if (pLCDDevice && pGraLib)
    {
      cChannel * ch = Channels.GetByNumber(currentChannelNr);
      if (ch)
      {

#if 0
        if(pLCDDevice->GetSizeY()<=MAXY_S) {
          // tiny and small displays

          if(((GraphLCDSetup.ShowDateTime==1) ||
              ((GraphLCDSetup.ShowDateTime==2)&&(State!=Menu))) &&
              GraphLCDSetup.ShowChannel && GraphLCDSetup.ShowProgram) {
            // compressed
            yPos = FRAME_SPACE_Y + normalFont->Height2() + 2*TEXT_OFFSET_Y_CHANNEL +FRAME_SPACE_YB;
          } else {
            yPos = FRAME_SPACE_Y;
            if((GraphLCDSetup.ShowDateTime==1) ||
               ((GraphLCDSetup.ShowDateTime==2)&&(State!=Menu))) {
              yPos += normalFont->Height2() + 2*TEXT_OFFSET_Y_TIME    +FRAME_SPACE_YB;
            }
            if(GraphLCDSetup.ShowChannel) {
              yPos += normalFont->Height2() + 2*TEXT_OFFSET_Y_CHANNEL +FRAME_SPACE_YB;
            }
          }
          if(IsLogoActive()) {
            yPos = max(yPos, FRAME_SPACE_Y + pcLogo->GetHeight() + 2 + FRAME_SPACE_YB);
          }

          xPos = pLCDDevice->GetSizeX()-FRAME_SPACE_X;

          if(ch->Ca()) {
            xPos -= symbols->WidthC('C')+1;
            pGraLib->Char(xPos, yPos, pLCDDevice->GetSizeX()-1, 'C', symbols, colorBlack, colorWhite);
          }
          if(ch->Dpid1()) {
            xPos -= symbols->WidthC('D')+1;
            pGraLib->Char(xPos, yPos, pLCDDevice->GetSizeX()-1, 'D', symbols, colorBlack, colorWhite);
          }
          if(ch->Apid2()) {
            xPos -= symbols->WidthC('A')+1;
            pGraLib->Char(xPos, yPos, pLCDDevice->GetSizeX()-1, 'A', symbols, colorBlack, colorWhite);
          }
          if(ch->Tpid()) {
            xPos -= symbols->WidthC('T')+1;
            pGraLib->Char(xPos, yPos, pLCDDevice->GetSizeX()-1, 'T', symbols, colorBlack, colorWhite);
          }
          for(i=LCDMAXCARDS-1; i>=0; i--) {
            if(iCardIsRecording[i]>0) {
              xPos -= symbols->WidthC((unsigned char)(49+i))+1;
              pGraLib->Char(xPos, yPos, pLCDDevice->GetSizeX()-1, (unsigned char)(49+i), symbols, colorBlack, colorWhite);
            }
          }

          if (iVolume==0) { // mute
            xPos -= symbols->WidthC('S')+1;
            pGraLib->Char(xPos, yPos, pLCDDevice->GetSizeX()-1, 'M', symbols, colorBlack, colorWhite);
          }
    
        } else
#endif
          if(pLCDDevice->SizeY()<=MAXY_M) {
          // medium display

          yPos = FRAME_SPACE_Y;
          xPos = pLCDDevice->SizeX() - FRAME_SPACE_X - symbols->Width();

          if (IsLogoActive())
          {
            xPos -= FRAME_SPACE_XB + logo->Width() + 2;
          }

          if(GraphLCDSetup.ShowSymbols == 1) {  // normal/fixed symbols
            // new layout:
            //    displays rec symbols for every card and
            //    2chan + dolby have their own symbols
            //    user triggered symbols

            int yPos2 = 0;

            // blank or 2chan or mute
            if (iVolume == 0) {
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'M', symbols, colorBlack, colorWhite);
              yPos += symbols->HeightC('S')+SYMBOL_SPACE;
            } else if (ch->Apid2()) {
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'A', symbols, colorBlack, colorWhite);
              yPos += symbols->HeightC('A')+SYMBOL_SPACE;
            } else {
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, ' ', symbols, colorBlack, colorWhite);
              yPos += symbols->HeightC(' ')+SYMBOL_SPACE;
            }
        
            
            // blank or dolby
            pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, ch->Dpid1() ?'D':' ', symbols, colorBlack, colorWhite);
            yPos += symbols->HeightC(ch->Dpid1() ?'D':' ')+SYMBOL_SPACE;

            // blank or teletext
            pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, ch->Tpid()  ?'T':' ', symbols, colorBlack, colorWhite);
            yPos += symbols->HeightC(ch->Tpid()  ?'T':' ')+SYMBOL_SPACE;

            // blank or crypt
            pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, ch->Ca()    ?'C':' ', symbols, colorBlack, colorWhite);
            yPos += symbols->HeightC(ch->Tpid()  ?'C':' ')+SYMBOL_SPACE;

            // show REC symbols at the right border below the logo
            yPos2 = yPos;
            yPos = FRAME_SPACE_Y;
            if((GraphLCDSetup.ShowDateTime==1) ||
               ((GraphLCDSetup.ShowDateTime==2)&&(State!=Menu))) {
              yPos += normalFont->Height2() + 2*TEXT_OFFSET_Y_TIME    +FRAME_SPACE_YB;
            }
            if(GraphLCDSetup.ShowChannel) {
              yPos += normalFont->Height2() + 2*TEXT_OFFSET_Y_CHANNEL +FRAME_SPACE_YB;
            }
            if (IsLogoActive())
            {
              yPos = max(yPos, FRAME_SPACE_Y + logo->Height() + 2 + FRAME_SPACE_YB);
            }

            yPos = max(yPos, yPos2);
            xPos = pLCDDevice->SizeX()-FRAME_SPACE_X+SYMBOL_SPACE;
            for (i = LCDMAXCARDS - 1; i >= 0; i--)
            {
              if (cardIsRecording[i] > 0)
              {
                xPos -= symbols->WidthC((unsigned char)(49+i))+SYMBOL_SPACE;
                pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, (unsigned char)(49+i), symbols, colorBlack, colorWhite);
              }
            }

            // show external triggered symbols
            if(GraphLCDSetup.ShowETSymbols && (strlen(szETSymbols)>0)) {
              for(i=strlen(szETSymbols)-1; i>=0; i--) {
                xPos -= symbols->WidthC((unsigned char)szETSymbols[i])+SYMBOL_SPACE;
                pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, (unsigned char)szETSymbols[i], symbols, colorBlack, colorWhite);
              }
            }
          } else {  // compressed symbols
            // old layout:
            //    displays only 1 rec symbol and
            //    a combined 2chan + dolby symbol

            // blank or teletext
            pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, ch->Tpid() ?'T':' ', symbols, colorBlack, colorWhite);
            yPos += symbols->HeightC(ch->Tpid()  ?'T':' ')+SYMBOL_SPACE;

            // blank or crypt
            pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, ch->Ca()   ?'C':' ', symbols, colorBlack, colorWhite);
            yPos += symbols->HeightC(ch->Tpid()  ?'C':' ')+SYMBOL_SPACE;

            // blank, 2chan, dolby or combined symbol
            if(iVolume == 0) {
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'M', symbols, colorBlack, colorWhite);
              yPos += symbols->HeightC('S')+SYMBOL_SPACE;
            } else if(ch->Apid2() && ch->Dpid1()) {
              // if Apid2 and Dpid1 are set then use combined symbol
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'B', symbols, colorBlack, colorWhite);
              yPos += symbols->HeightC('B')+SYMBOL_SPACE;
            } else if(ch->Apid2()) {
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'A', symbols, colorBlack, colorWhite);
              yPos += symbols->HeightC('A')+SYMBOL_SPACE;
            } else if(ch->Dpid1()) {
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'D', symbols, colorBlack, colorWhite);
              yPos += symbols->HeightC('D')+SYMBOL_SPACE;
            } else {
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1,' ', symbols, colorBlack, colorWhite);
              yPos += symbols->HeightC(' ')+SYMBOL_SPACE;
            }

            // blank or rec
            if(cRecordControls::Active()) {
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1,'1' , symbols, colorBlack, colorWhite);
            } else {
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1,' ', symbols, colorBlack, colorWhite);
            }
          }



        } else {
          // large display

          yPos = FRAME_SPACE_Y;
          if((GraphLCDSetup.ShowDateTime==1) ||
             ((GraphLCDSetup.ShowDateTime==2)&&(State!=Menu))) {
            yPos += normalFont->Height2() + 2*TEXT_OFFSET_Y_TIME    +FRAME_SPACE_YB;
          }
          if(GraphLCDSetup.ShowChannel) {
            yPos += normalFont->Height2() + 2*TEXT_OFFSET_Y_CHANNEL +FRAME_SPACE_YB;
          }
          if (IsLogoActive())
          {
            yPos = max(yPos, FRAME_SPACE_Y + logo->Height() + 2 + FRAME_SPACE_YB);
          }

          xPos = pLCDDevice->SizeX()-FRAME_SPACE_X - symbols->WidthC(' ');

          if(GraphLCDSetup.ShowSymbols == 1) {  // normal/fixed symbols

            // blank or teletext
            pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, ch->Tpid() ?'T':' ', symbols, colorBlack, colorWhite);
            xPos -= symbols->WidthC(ch->Tpid() ?'T':' ')+SYMBOL_SPACE;

            // blank or dolby
            pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, ch->Dpid1()?'D':' ', symbols, colorBlack, colorWhite);
            xPos -= symbols->WidthC(ch->Dpid1()?'D':' ')+SYMBOL_SPACE;

            if(pLCDDevice->SizeY() > MAXY_M) {  // with 128 pixel width only 3 symbols...
              // blank or crypt
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, ch->Ca()?'C':' ', symbols, colorBlack, colorWhite);
              xPos -= symbols->WidthC(ch->Ca()?'C':' ')+SYMBOL_SPACE;
            }

            // blank or 2chan or mute
            if (iVolume == 0) {
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'M', symbols, colorBlack, colorWhite);
              xPos -= symbols->WidthC('S')+SYMBOL_SPACE;
            } else if (ch->Apid2()) {
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'A', symbols, colorBlack, colorWhite);
              xPos -= symbols->WidthC('A')+SYMBOL_SPACE;
            } else {
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, ' ', symbols, colorBlack, colorWhite);
              xPos -= symbols->WidthC(' ')+SYMBOL_SPACE;
            }

          } else {  // compressed symbols

            // crypt
            if(ch->Ca()) {
              xPos -= symbols->WidthC('C')+SYMBOL_SPACE;
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'C', symbols, colorBlack, colorWhite);
            }

            // teletext
            if(ch->Tpid()) {
              xPos -= symbols->WidthC('T')+SYMBOL_SPACE;
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'T', symbols, colorBlack, colorWhite);
            }

            // dolby
            if(ch->Dpid1()) {
              xPos -= symbols->WidthC('D')+SYMBOL_SPACE;
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'D', symbols, colorBlack, colorWhite);
            }

            // 2chan
            if(ch->Apid2()) {
              xPos -= symbols->WidthC('A')+SYMBOL_SPACE;
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'A', symbols, colorBlack, colorWhite);
            }

            // mute
            if(iVolume == 0) {
              xPos -= symbols->WidthC('S')+SYMBOL_SPACE;
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, 'M', symbols, colorBlack, colorWhite);
            }
          }

          // show REC symbols at the right border of the 'next line'
          xPos = pLCDDevice->SizeX()-FRAME_SPACE_X+SYMBOL_SPACE;
          yPos += symbols->Height() + FRAME_SPACE_YB;
          for (i = cDevice::NumDevices() - 1; i >= 0; i--)
          {
            // Just display present devices
            xPos -= symbols->WidthC((unsigned char)(49+i))+SYMBOL_SPACE;
            if (cardIsRecording[i] > 0)
            { // Show a recording Symbol
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, (unsigned char)(49+i), symbols, colorBlack, colorWhite);
            }
            else
            {
              if (GraphLCDSetup.ShowNotRecording == 1)
              { // Do we want an empty frame around not recording card's icons?
                // Show an empty frame instead of the recording symbol
                pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, ' ', symbols, colorBlack, colorWhite);
              }
            }
          }

          // show external triggered symbols
          if(GraphLCDSetup.ShowETSymbols && (strlen(szETSymbols)>0)) {
            for(i=strlen(szETSymbols)-1; i>=0; i--) {
              xPos -= symbols->WidthC((unsigned char)szETSymbols[i])+SYMBOL_SPACE;
              pGraLib->Char(xPos, yPos, pLCDDevice->SizeX()-1, (unsigned char)szETSymbols[i], symbols, colorBlack, colorWhite);
            }
          }
        }
      }
    }
  }
}


void cGraphLCDDisplay::DisplayProgramm()
{
  struct tm tm_r;
  char buffer[25];
  string str;
  bool showTimeBar = false;
  int timeBarWidth = 0;
  int timeBarValue = 0;

  if (GraphLCDSetup.ShowProgram)
  {
    strftime(buffer, sizeof(buffer), "%R", localtime_r(&presentTime, &tm_r));
    if (followingTime && followingTime != presentTime)
    {
      str = buffer;
      if ((pLCDDevice->SizeX() >= MINY_L || !IsSymbolsActive()))
      {
        str += " - ";
      }
      else
      {
        str += "-";
      }
      strftime(buffer, sizeof(buffer), "%R", localtime_r(&followingTime, &tm_r));
      str += buffer;
      showTimeBar = true;
      timeBarWidth = normalFont->WidthS(str) - 1;
      timeBarValue = (timeBarValue < 0) ? 0 : (time(NULL) - presentTime) * timeBarWidth / (followingTime - presentTime);
    }
    else
    {
      str = buffer;
    }

    if (!presentTime)
    {
      std::vector<cScroller>::iterator i = scroller.begin();
      std::vector<cScroller>::const_iterator e = scroller.end();
      for(;e != i;++i)
      {
        i->Reset();
      }
    }

    if (pLCDDevice && pGraLib && normalFont && largeFont && presentTime)
    {
      if (scroller.size() < 1 
        || presentTitle != scroller[0].Text()
        || (scroller.size() > 1 && presentSubtitle != scroller[1].Text()))
      {
        if (pLCDDevice->SizeY() <= MAXY_S)
        {
          scroller.resize(1);

          int nTopY = pLCDDevice->SizeY() - (TEXT_OFFSET_Y_TITLE - 1) - largeFont->Height();
          int nMaxX = max(1,pLCDDevice->SizeX()-1-(2*FRAME_SPACE_X));
          // Logo enabled & available, and text with Logo is overlapped
          if (IsLogoActive() && nTopY < logo->Height()) {
            nMaxX -= FRAME_SPACE_XB;
            nMaxX -= logo->Width();
            nMaxX = max(nMaxX-2,1);//Frame around Logo
          }
          // If symbols used, and text with symbols is overlapped
          if (IsSymbolsActive()) {
            nMaxX -= FRAME_SPACE_XB;
            nMaxX -= symbols->Width();
            nMaxX = max(nMaxX,1);
          }

          scroller[0].Init(FRAME_SPACE_X + TEXT_OFFSET_X,
              nTopY,
              nMaxX,
              largeFont, presentTitle);
        }
        else
        {
          scroller.resize(2);

          scroller[0].Init(FRAME_SPACE_X + TEXT_OFFSET_X,
              pLCDDevice->SizeY() - 2 * (TEXT_OFFSET_Y_TITLE - 1) - largeFont->Height() - normalFont->Height(),
              pLCDDevice->SizeX() - 1,
              largeFont, presentTitle);
          scroller[1].Init(FRAME_SPACE_X + TEXT_OFFSET_X,
              pLCDDevice->SizeY()- (TEXT_OFFSET_Y_TITLE-1) - normalFont->Height(),
              pLCDDevice->SizeX() - 1,
              normalFont, presentSubtitle);
        }
      }
      if ((pLCDDevice->SizeY()<=MAXY_S))
      {
        // tiny and small LCDs
        pGraLib->Text(FRAME_SPACE_X,
            pLCDDevice->SizeY()- 2*(TEXT_OFFSET_Y_TITLE-1) - largeFont->Height() - normalFont->Height(),
            pLCDDevice->SizeX()-1,
            str,  normalFont,
            true, colorBlack, colorWhite);
      }
      else
      {
        // medium and large LCDs
        pGraLib->Text(FRAME_SPACE_X,
            pLCDDevice->SizeY()- 3*(TEXT_OFFSET_Y_TITLE-1) - largeFont->Height() - 2*normalFont->Height() - (showTimeBar && GraphLCDSetup.ShowTimebar ? TIMEBAR_HEIGHT+1 : 0),
            pLCDDevice->SizeX()-1,
            str,  normalFont,
            true, colorBlack, colorWhite);
        if (showTimeBar && GraphLCDSetup.ShowTimebar)
        {
          pGraLib->Rectangle(FRAME_SPACE_X,
              pLCDDevice->SizeY()- 3*(TEXT_OFFSET_Y_TITLE-1) - largeFont->Height() - normalFont->Height() - TIMEBAR_HEIGHT-1,
              FRAME_SPACE_X + timeBarWidth,
              pLCDDevice->SizeY()- 3*(TEXT_OFFSET_Y_TITLE-1) - largeFont->Height() - normalFont->Height() - 2,
              false,
              colorBlack);
          pGraLib->Rectangle(FRAME_SPACE_X,
              pLCDDevice->SizeY()- 3*(TEXT_OFFSET_Y_TITLE-1) - largeFont->Height() - normalFont->Height() - TIMEBAR_HEIGHT-1,
              FRAME_SPACE_X + timeBarValue,
              pLCDDevice->SizeY()- 3*(TEXT_OFFSET_Y_TITLE-1) - largeFont->Height() - normalFont->Height() - 2,
              true,
              colorBlack);
        }
      }
      /** Draw Programmtext*/
      {
        std::vector<cScroller>::iterator i = scroller.begin();
        std::vector<cScroller>::const_iterator e = scroller.end();
        for(;e != i;++i)
        {
          i->Draw(pGraLib);
        }
      }
    }
  }
}


/** Check if replay index bigger as one hour */
bool cGraphLCDDisplay::IndexIsGreaterAsOneHour(int Index) const
{
  int h = (Index / FRAMESPERSEC) / 3600;
  return h > 0;
}


/** Translate replay index to string with minute and second MM:SS */
const char *cGraphLCDDisplay::IndexToMS(int Index) const
{
  static char buffer[16];
  int s = (Index / FRAMESPERSEC);
  int m = s / 60;
  s %= 60;
  snprintf(buffer, sizeof(buffer), "%02d:%02d", m, s);
  return buffer;
}

/** Compare Scroller with new Textbuffer*/
bool cGraphLCDDisplay::IsScrollerTextChanged(const std::vector<cScroller> & scrollers, const std::vector <std::string> & lines) const
{
  if(lines.size() == 0)
    return true; //Different size found
  if(scrollers.size() == 0)
    return true; //Different size found

  std::vector<cScroller>::const_iterator i = scrollers.begin();
  std::vector<cScroller>::const_iterator e = scrollers.end();
  std::vector<string>::const_iterator li = lines.begin();
  std::vector<string>::const_iterator le = lines.end();

  for (; e != i && le != li; ++li,++i)
  {
    if (i->Text() != (*li))
      return true; //Different text found
  }  
  return false; //Text seem equal
}

void cGraphLCDDisplay::DisplayReplay(int Current, int Total)
{
  int nMaxX, nProgressbarHeight, nTopY;
  int nWidthPreMsg = 0, nWidthCurrent = 0, nWidthTotal = 0,nWidthOffset = 0;
  std::string szPreMsg,szCurrent,szTotal;

  if (pLCDDevice && pGraLib && largeFont && normalFont)
  {
    if (pLCDDevice->SizeY() >= MINY_L)
      nProgressbarHeight = 15;
    else if (pLCDDevice->SizeY() >= MINY_M)
      nProgressbarHeight = 9;
    else if (pLCDDevice->SizeY() >= MINY_S)
      nProgressbarHeight = 5;
    else
      nProgressbarHeight = 3;

    if (IsLogoActive())
      nTopY = FRAME_SPACE_Y + logo->Height() + 2;
    else if (GraphLCDSetup.ShowDateTime)
      nTopY = FRAME_SPACE_Y + normalFont->Height2() + 2 * TEXT_OFFSET_Y_TIME + FRAME_SPACE_YB;
    else
      nTopY = FRAME_SPACE_Y;
    if (replayName.length() > 0)
    {
      int lineHeight, maxLines;
      vector <string> lines;

      nMaxX = max(1,pLCDDevice->SizeX()-(2*FRAME_SPACE_X) - 2*TEXT_OFFSET_X);
      lineHeight = FRAME_SPACE_Y + largeFont->Height();
      maxLines = max(0, (pLCDDevice->SizeY() - normalFont->Height() - FRAME_SPACE_Y - nProgressbarHeight - 2 - nTopY) / lineHeight);

      if (maxLines == 0)
      {
        if (IsLogoActive())
        {
          // draw replayname next to logo
          nMaxX = max(1, nMaxX - FRAME_SPACE_X - logo->Width() - 2);
          if (GraphLCDSetup.ShowDateTime)
            nTopY = FRAME_SPACE_Y + normalFont->Height2() + 2 * TEXT_OFFSET_Y_TIME + FRAME_SPACE_YB;
          else
            nTopY = FRAME_SPACE_Y;
          maxLines = (pLCDDevice->SizeY() - normalFont->Height() - FRAME_SPACE_Y - nProgressbarHeight - 2 - nTopY) / lineHeight;
        }
        if (maxLines <= 1)
        {
          // use singleline mode
          lines.push_back(replayName);
        }
        else
          WrapText(replayName, lines, largeFont, nMaxX, maxLines, false);
      }
      else if (maxLines == 1) //singleline mode
        lines.push_back(replayName);
      else {
        WrapText(replayName, lines, largeFont, nMaxX, maxLines, false);
      }

      if (scroller.size() != lines.size() // if any text is changed
        || IsScrollerTextChanged(scroller,lines))
      {
        // Same size for Scroller and Textbuffer
        scroller.resize(lines.size());

        std::vector<cScroller>::iterator i = scroller.begin();
        std::vector<cScroller>::const_iterator e = scroller.end();
        std::vector<string>::const_iterator li = lines.begin();
        std::vector<string>::const_iterator le = lines.end();

        for (int n = lines.size(); e != i && le != li; ++li,++i,--n)
        {
          nTopY = pLCDDevice->SizeY() - normalFont->Height() - FRAME_SPACE_Y - nProgressbarHeight - n * lineHeight - 2;
          i->Init(FRAME_SPACE_X + TEXT_OFFSET_X, nTopY, nMaxX + FRAME_SPACE_X + TEXT_OFFSET_X, largeFont, *li);
        }
      }
    }

    /** Draw Replaytext*/
    {
      std::vector<cScroller>::iterator i = scroller.begin();
      std::vector<cScroller>::const_iterator e = scroller.end();
      for(;e != i;++i)
      {
        i->Draw(pGraLib);
      }
    }

    /** Draw Progressbar with current and total replay time*/
    nTopY = pLCDDevice->SizeY() - normalFont->Height() - FRAME_SPACE_Y - nProgressbarHeight - 2;
    nMaxX = max(1, pLCDDevice->SizeX() - 1 - 2*FRAME_SPACE_X);
    // Logo enabled & available, and text with Logo is overlapped
    if (IsLogoActive() && nTopY < logo->Height()) {  
      nMaxX -= FRAME_SPACE_X;
      nMaxX -= logo->Width();
      nMaxX = max(nMaxX-2,1);//Frame around Logo
    }

    pGraLib->Rectangle(FRAME_SPACE_X,
        nTopY,
        FRAME_SPACE_X + nMaxX,
        nTopY + nProgressbarHeight,
        false, colorBlack);

    if((1 < Total) && (1 < Current)) // Don't show full progressbar for endless streams 
    {
      pGraLib->Rectangle(FRAME_SPACE_X,
          nTopY,
          FRAME_SPACE_X + (min(Total,Current)*(nMaxX)/Total),
          nTopY + nProgressbarHeight,
          true, colorBlack);
    }

    /** Draw Strings with current and total replay time*/
    nTopY = pLCDDevice->SizeY() - normalFont->Height() - FRAME_SPACE_Y;
    // use same width like Progressbar
//    if(!IsLogoActive() || nTopY > logo->Height()) 
//      nMaxX = max(1,pLCDDevice->SizeX()-1-(2*FRAME_SPACE_X));

    switch (replayMode)
    {
      case eReplayDVD:    szPreMsg = tr("DVD");    break;
      case eReplayMusic:  szPreMsg = tr("Music");  break;
      case eReplayFile:   szPreMsg = tr("File");   break;
      case eReplayImage:  szPreMsg = tr("Image");  break;
      case eReplayAudioCD:szPreMsg = tr("CD");     break;
      default:            szPreMsg = tr("Replay"); break;
    }

    if(pLCDDevice->SizeX() >= MINY_M)
      szPreMsg += " : ";
    else
      szPreMsg += ":";

    if( replayMode == eReplayImage) {//Image-Plugin hasn't Frames per Seconds
      char buffer[8];
      snprintf(buffer, sizeof(buffer),"%d",Current); 
      szCurrent = buffer;
      snprintf(buffer, sizeof(buffer),"%d",Total); 
      szTotal = buffer;
    }
    else
    {
      if((Total>1 && IndexIsGreaterAsOneHour(Total)) // Check if any index bigger as one hour
        || IndexIsGreaterAsOneHour(Current))
      {
        szCurrent = IndexToHMSF(Current);
        if( Total>1 ) // Don't draw totaltime for endless streams  
          szTotal   = IndexToHMSF(Total);
      }
      else
      { // Show only minutes and seconds on short replays
        szCurrent = IndexToMS(Current);
        if( Total>1 ) // Don't draw totaltime for endless streams  
          szTotal   = IndexToMS(Total);
      }
    }
    // Get width of drawable strings
    nWidthPreMsg  = normalFont->WidthS(szPreMsg);
    nWidthCurrent = normalFont->WidthS(szCurrent);
    if(szTotal.length()) // Don't draw empty string 
      nWidthTotal = normalFont->WidthS(szTotal);

    // Draw depends on display width, any placeable informations
    if (nWidthTotal && nWidthPreMsg && (nWidthPreMsg + nWidthCurrent + nWidthTotal + 5 < nMaxX)) {
      // Show prefix and all position
      nWidthOffset = pGraLib->Text(FRAME_SPACE_X, nTopY, nMaxX, szPreMsg, normalFont, true, colorBlack, colorWhite);
      pGraLib->Text(FRAME_SPACE_X + nWidthOffset + 1, nTopY, nMaxX, szCurrent, normalFont, true, colorBlack, colorWhite);
      pGraLib->Text(nMaxX - nWidthTotal, nTopY, nMaxX, szTotal, normalFont, true, colorBlack, colorWhite);
    }
    else if (nWidthTotal && (nWidthCurrent + nWidthTotal + 5 < nMaxX)) {
      //Show current and total position
      pGraLib->Text(FRAME_SPACE_X, nTopY, nMaxX, szCurrent, normalFont, true, colorBlack, colorWhite);
      pGraLib->Text(nMaxX - nWidthTotal, nTopY, nMaxX, szTotal, normalFont, true, colorBlack, colorWhite);
    }
    else if(!nWidthTotal && nWidthPreMsg && (nWidthPreMsg + nWidthCurrent + 1 < nMaxX)) {
      // Show prefix and current position
      nWidthOffset = pGraLib->Text(FRAME_SPACE_X, nTopY, nMaxX, szPreMsg, normalFont, true, colorBlack, colorWhite);
      pGraLib->Text(FRAME_SPACE_X + nWidthOffset + 1, nTopY, nMaxX, szCurrent, normalFont, true, colorBlack, colorWhite);
    }        
    else { //if(nWidthCurrent < nMaxX) {                                              //Show only current position
      pGraLib->Text(FRAME_SPACE_X, nTopY, nMaxX, szCurrent, normalFont, true, colorBlack, colorWhite);
    }
  }
}


void
cGraphLCDDisplay::DisplayMenu(void) {
  char  buffer2[255];
  char *pszTmp1,
       *pszTmp2;
  int   iAT, t;
  int   FrameWidth, yPos, iEntryHeight;

  if (pLCDDevice && pGraLib && normalFont) {
    FrameWidth = max(pLCDDevice->SizeX()-2*FRAME_SPACE_X,1);

    if((GraphLCDSetup.ShowDateTime==1) ||
       ((GraphLCDSetup.ShowDateTime==2)&&(State!=Menu))) {
      yPos = FRAME_SPACE_Y + normalFont->Height2() + 2*TEXT_OFFSET_Y_TIME + FRAME_SPACE_YB;
    } else {
      yPos = FRAME_SPACE_Y;
    }

    // draw Menu Title
    if (osdTitle.length() > 0)
    {
      pGraLib->RoundRectangle(FRAME_SPACE_X,
          yPos,
          FRAME_SPACE_X+FrameWidth-1,
          yPos + normalFont->Height2() + 2*TEXT_OFFSET_Y_CHANNEL -1,
          true, colorBlack, (TEXT_OFFSET_Y_CHANNEL>=4) ? 4 : 1);
      pGraLib->Text(FRAME_SPACE_X+TEXT_OFFSET_X,
                       yPos+TEXT_OFFSET_Y_CHANNEL,
                       FRAME_SPACE_X+FrameWidth-1,
                       osdTitle,  normalFont,
                       true, colorWhite, colorBlack);
    }

    if (!(osdTextItemLines.size() > 0))
    {
      // draw Menu Entries
      iEntryHeight = 2*(normalFont->Height()-normalFont->Height2())+normalFont->Height2();
      yPos = yPos + normalFont->Height2() + 2*TEXT_OFFSET_Y_CHANNEL + FRAME_SPACE_YB;
      if (GraphLCDSetup.ShowColorButtons && smallFont &&
         (osdColorButton[0].length() > 0 || osdColorButton[1].length() > 0 ||
          osdColorButton[2].length() > 0 || osdColorButton[3].length() > 0))
      {
        menuCount = (pLCDDevice->SizeY()-yPos- smallFont->Height() - 4 - FRAME_SPACE_Y/3) / iEntryHeight;
      } else {
        menuCount = (pLCDDevice->SizeY()-yPos) / iEntryHeight;
      }


      if (menuCurrent < menuTop)
        menuTop = menuCurrent;
      if (menuCurrent > menuTop + menuCount - 1)
        menuTop = max(0, menuCurrent + 1 - menuCount);

      pGraLib->Rectangle(0,yPos,pLCDDevice->SizeX()-1,pLCDDevice->SizeY()-1, true, colorWhite);


      for (int i = menuTop; i < min((int) osdItems.size(), menuTop + menuCount); i++)
      {
        if (i == menuCurrent)
        {
          pGraLib->RoundRectangle(FRAME_SPACE_X,yPos + (i - menuTop)*iEntryHeight,
              pLCDDevice->SizeX()-1-FRAME_SPACE_X,
              yPos + (i - menuTop+1)*iEntryHeight - 1,
              true, colorBlack, (TEXT_OFFSET_Y_CHANNEL>=4) ? 3 : 1);
        }
        strncopy(buffer2, osdItems[i].c_str(), sizeof(buffer2));
        pszTmp1 = buffer2;
        pszTmp2 = strchr(pszTmp1, '\t');
        iAT = 0; t = 0;

        while(pszTmp1 && pszTmp2) {
          *pszTmp2 = '\0';
          pGraLib->Text(FRAME_SPACE_X+TEXT_OFFSET_X + t,
                     yPos + (i - menuTop)*iEntryHeight + (normalFont->Height()-normalFont->Height2()),
                     min(FRAME_SPACE_X+TEXT_OFFSET_X + t+iTab[iAT+1], pLCDDevice->SizeX()-1-FRAME_SPACE_X),
                     pszTmp1, normalFont,
                     true,
                     (i == menuCurrent) ? colorWhite:colorBlack,
                     (i == menuCurrent) ? colorBlack:colorWhite);
          pszTmp1 = pszTmp2+1;
          pszTmp2 = strchr(pszTmp1, '\t');
          t = t + iTab[iAT+1]+TEXT_OFFSET_X;
          iAT++;
        }

        pGraLib->Text(FRAME_SPACE_X+TEXT_OFFSET_X + t,
                   yPos + (i - menuTop)*iEntryHeight + (normalFont->Height()-normalFont->Height2()),
                   pLCDDevice->SizeX()-1-FRAME_SPACE_X,
                   pszTmp1, normalFont,
                   true,
                   (i == menuCurrent) ? colorWhite:colorBlack,
                   (i == menuCurrent) ? colorBlack:colorWhite);
      }
    }
  }
}


void cGraphLCDDisplay::DisplayMessage()
{
  vector <string> lines;
  int lineCount;
  int maxTextLen, recW, recH;
  int entryHeight;
  

  if (GraphLCDSetup.ShowMessages && osdMessage.length() > 0)
  {
    if (pLCDDevice && pGraLib && normalFont)
    {
      maxTextLen = pLCDDevice->SizeX() - 2 * FRAME_SPACE_X - 2 * FRAME_SPACE_XB - 2 * TEXT_OFFSET_X - 10;
      recW = WrapText(osdMessage, lines, normalFont, maxTextLen, MAXLINES_MSG);
      lineCount = lines.size();
      
      // display text
      entryHeight = 2 * (normalFont->Height() - normalFont->Height2()) + normalFont->Height2();
      recH = lineCount * entryHeight + 2 * TEXT_OFFSET_Y_CHANNEL + 2 * FRAME_SPACE_YB;
      recW = recW + 2 * TEXT_OFFSET_X + 2 * FRAME_SPACE_XB + 2 * FRAME_SPACE_X;
      recW += (recW % 2);

      pGraLib->Rectangle((pLCDDevice->SizeX()-recW)/2,
                      (pLCDDevice->SizeY()-recH)/2,
                      pLCDDevice->SizeX()-1-(pLCDDevice->SizeX()-recW)/2,
                      pLCDDevice->SizeY()-1-(pLCDDevice->SizeY()-recH)/2,
                      true, colorWhite);
      recH = recH - 2 * FRAME_SPACE_YB;
      recW = recW - 2 * FRAME_SPACE_XB;
      pGraLib->Rectangle((pLCDDevice->SizeX()-recW)/2,
                      (pLCDDevice->SizeY()-recH)/2,
                      pLCDDevice->SizeX()-1-(pLCDDevice->SizeX()-recW)/2,
                      pLCDDevice->SizeY()-1-(pLCDDevice->SizeY()-recH)/2,
                      false, colorBlack);
      recH = recH - 2 * TEXT_OFFSET_Y_CHANNEL;
      recW = recW - 2 * TEXT_OFFSET_X;
      for (int i = 0; i < lineCount; i++)
      {
        pGraLib->Text((pLCDDevice->SizeX()-normalFont->WidthS(lines[i])) / 2,
                   (pLCDDevice->SizeY()-recH)/2 + i * entryHeight + (normalFont->Height()-normalFont->Height2()),
                   pLCDDevice->SizeX()-(pLCDDevice->SizeX()-recW)/2,
                   lines[i], normalFont, true,
                   colorBlack, colorWhite);
      }
    }
  }
}


void cGraphLCDDisplay::DisplayTextItem()
{
  int lineCount;
  int iEntryHeight, iLineAnz;
  int yPos;

  if (osdTextItemLines.size() > 0)
  {
    if (pLCDDevice && pGraLib && normalFont)
    {
      lineCount = osdTextItemLines.size();

      if (GraphLCDSetup.ShowDateTime == 1 ||
          (GraphLCDSetup.ShowDateTime == 2 && State!=Menu))
      {
        yPos = FRAME_SPACE_Y + normalFont->Height2() + 2*TEXT_OFFSET_Y_TIME + FRAME_SPACE_YB;
      }
      else
      {
        yPos = FRAME_SPACE_Y;
      }

      // draw Text
      iEntryHeight = 2 * (normalFont->Height() - normalFont->Height2()) + normalFont->Height2();
      yPos = yPos + normalFont->Height2() + 2 * TEXT_OFFSET_Y_CHANNEL + FRAME_SPACE_YB;
      if (GraphLCDSetup.ShowColorButtons && smallFont &&
          (osdColorButton[0].length() > 0 || osdColorButton[1].length() > 0 ||
           osdColorButton[2].length() > 0 || osdColorButton[3].length() > 0))
      {
        iLineAnz = (pLCDDevice->SizeY() - yPos - smallFont->Height() - 4 - FRAME_SPACE_Y / 3) / iEntryHeight;
      }
      else
      {
        iLineAnz = (pLCDDevice->SizeY() - yPos) / iEntryHeight;
      }

      int startLine = osdTextItemTop;
      for (int i = 0; i < min(lineCount, iLineAnz); i++)
      {
        if (i + startLine < lineCount)
          pGraLib->Text(FRAME_SPACE_X + TEXT_OFFSET_X,
              yPos + i * iEntryHeight + (normalFont->Height() - normalFont->Height2()),
              pLCDDevice->SizeX() - 1 - FRAME_SPACE_X,
              osdTextItemLines[i + startLine], normalFont,
              true, colorBlack, colorWhite);
      }
    }
  }
}


void cGraphLCDDisplay::DisplayColorButtons()
{
  int i, buttonWidth, textLen;

  if (GraphLCDSetup.ShowColorButtons)
  {
    if (pLCDDevice && pGraLib && smallFont)
    {
      buttonWidth = (pLCDDevice->SizeX() / 4) - (FRAME_SPACE_X ? 2 * FRAME_SPACE_X : 1);

      for (i = 0; i < 4; i++)
      {
        if (osdColorButton[i].length() > 0)
        {
          pGraLib->RoundRectangle(i * (pLCDDevice->SizeX()/4) + FRAME_SPACE_X,
              pLCDDevice->SizeY()- smallFont->Height() - 4 - FRAME_SPACE_Y/3,
              i    *(pLCDDevice->SizeX()/4) + FRAME_SPACE_X + buttonWidth - 1,
              pLCDDevice->SizeY()                 - 1 - FRAME_SPACE_Y/3,
              true, colorBlack, max(1,(smallFont->Height()+4)/5));
          textLen = smallFont->WidthS(osdColorButton[i]);
          if (textLen <= buttonWidth - 2)
          {
            pGraLib->Text(i*(pLCDDevice->SizeX()/4) + (pLCDDevice->SizeX()/8) - (textLen+1)/2,
                       pLCDDevice->SizeY()- smallFont->Height() - 2 - FRAME_SPACE_Y/3,
                       i*(pLCDDevice->SizeX()/4) + FRAME_SPACE_X + buttonWidth - 1,
                       osdColorButton[i], smallFont,
                       true, colorWhite, colorBlack);
          }
          else
          {
            pGraLib->Text(i*(pLCDDevice->SizeX()/4) + FRAME_SPACE_X + 1,
                       pLCDDevice->SizeY()- smallFont->Height() - 2 - FRAME_SPACE_Y/3,
                       i*(pLCDDevice->SizeX()/4) + FRAME_SPACE_X + buttonWidth - 1,
                       osdColorButton[i], smallFont,
                       true, colorWhite, colorBlack);
          }
        }
      }
    }
  }
}


void cGraphLCDDisplay::DisplayVolume()
{
  int RecW, RecH;

  if (GraphLCDSetup.ShowVolume) {
    if (iLastVolChange > 0) {
      if (time_ms() - iLastVolChange < 2000) {
        RecH = (pLCDDevice->SizeY() / 5) + 2 * FRAME_SPACE_YB + 4 * FRAME_SPACE_YB;
        RecW = pLCDDevice->SizeX() / 2;
        pGraLib->RoundRectangle((pLCDDevice->SizeX()-RecW)/2,       // draw frame
            (pLCDDevice->SizeY()-RecH)/2,
            pLCDDevice->SizeX()-(pLCDDevice->SizeX()-RecW)/2-1,
            pLCDDevice->SizeY()-(pLCDDevice->SizeY()-RecH)/2-1,
            true, colorWhite, 1);
        RecH = RecH - 2 * FRAME_SPACE_YB;
        RecW = RecW - 2 * FRAME_SPACE_XB;
        pGraLib->RoundRectangle((pLCDDevice->SizeX()-RecW)/2,       // draw box
            (pLCDDevice->SizeY()-RecH)/2,
            pLCDDevice->SizeX()-1-(pLCDDevice->SizeX()-RecW)/2,
            pLCDDevice->SizeY()-1-(pLCDDevice->SizeY()-RecH)/2,
            false, colorBlack, 1);
        RecH = RecH - 2;
        RecW = RecW - 2;
        pGraLib->Rectangle((pLCDDevice->SizeX()-RecW)/2,       // draw bar
            (pLCDDevice->SizeY()-RecH)/2,
            (pLCDDevice->SizeX()-RecW)/2 + (iVolume * RecW) / 255,
            pLCDDevice->SizeY()-1-(pLCDDevice->SizeY()-RecH)/2,
            true, colorBlack);
        if (iVolume == 0) {            // display big mute symbol
          pGraLib->Char(pLCDDevice->SizeX()/2-symbols->WidthC('5'), pLCDDevice->SizeY()/2-symbols->HeightC('5'),
          pLCDDevice->SizeX()-1, '5', symbols, colorBlack, colorWhite);

          pGraLib->Char(pLCDDevice->SizeX()/2, pLCDDevice->SizeY()/2-symbols->HeightC('6'),
          pLCDDevice->SizeX()-1, '6', symbols, colorBlack, colorWhite);

          pGraLib->Char(pLCDDevice->SizeX()/2-symbols->WidthC('7'), pLCDDevice->SizeY()/2,
          pLCDDevice->SizeX()-1, '7', symbols, colorBlack, colorWhite);

          pGraLib->Char(pLCDDevice->SizeX()/2, pLCDDevice->SizeY()/2,
          pLCDDevice->SizeX()-1, '8', symbols, colorBlack, colorWhite);

        }
      } else {
        iLastVolChange = 0;
      }
    }
  }
}


void cGraphLCDDisplay::UpdateIn(long usec)
{
  if (usec == 0)
  {
    timerclear(&UpdateAt);
  }
  else
  {
    if (gettimeofday(&CurrTimeval, NULL) == 0)
    {
      // get current time
      UpdateAt.tv_sec = CurrTimeval.tv_sec;
      UpdateAt.tv_usec = CurrTimeval.tv_usec + usec;
      while (UpdateAt.tv_usec >= 1000000)
      {
        // take care of an overflow
        UpdateAt.tv_sec++;
        UpdateAt.tv_usec -= 1000000;
      }
    }
  }
}


bool cGraphLCDDisplay::CheckAndUpdateSymbols()
{
  bool                bRet = false;
  static struct stat  filestat;
  FILE*               InFile = NULL;
  static char         szLine[8];

  if(stat(FILENAME_EXTERNAL_TRIGGERED_SYMBOLS, &filestat)==0) {
    if(LastTimeModSym != filestat.st_mtime) {
      InFile = fopen(FILENAME_EXTERNAL_TRIGGERED_SYMBOLS, "r");
      if(InFile) {
        strcpy(szETSymbols, "");
        while(!feof(InFile) && (strlen(szETSymbols)+1<sizeof(szLine))) {
          strcpy(szLine, "");
          fgets(szLine, sizeof(szLine), InFile);
          compactspace(szLine);
          if((strlen(szLine)==2) && (szLine[1]=='1')) {
            strcat(szETSymbols, ".");
            szETSymbols[strlen(szETSymbols)-1] = szLine[0];
          }
        }
        fclose(InFile);
        LastTimeModSym = filestat.st_mtime;
        bRet = true;
      }
    }
  } else {
    if((errno == ENOENT) && (strlen(szETSymbols)>0)) {
      strcpy(szETSymbols, "");
      bRet = true;
    }
  }
  return bRet;
}


int cGraphLCDDisplay::WrapText(string & text, vector <string> & lines, cGraphLCDFont * font, int maxTextWidth, int maxLines, bool cutTooLong)
{
  int lineCount;
  int textWidth;
  int pos;
  int posLast;

  lines.clear();
  lines.push_back(text);
  lineCount = 1;

  pos = 0;
  posLast = 0;
  while (pos < (int) lines[lineCount - 1].length() && (lineCount < maxLines))
  {
    if (lines[lineCount - 1][pos] == '\n')
    {
      lines.push_back(trim(lines[lineCount - 1].substr(pos + 1)));
      lines[lineCount - 1].resize(pos);
      pos = 0;
      lineCount++;
    }
    else if (lines[lineCount - 1][pos] == ' ')
    {
      textWidth = font->WidthS(lines[lineCount - 1], pos);

      if (textWidth <= maxTextWidth)
      {
        posLast = pos;
        pos++;
      }
      else
      {
        if (posLast > 0)
        {
          lines.push_back(trim(lines[lineCount - 1].substr(posLast + 1)));
          lines[lineCount - 1].resize(posLast);
        }
        else
        {
          lines.push_back(trim(lines[lineCount - 1].substr(pos + 1)));
          lines[lineCount - 1].resize(pos);
        }
        pos = 0;
        posLast = 0;
        lineCount++;
      }
    }
    else
    {
      pos++;
    }
  }

  if (lineCount < maxLines)
  {
    textWidth = font->WidthS(lines[lineCount - 1], pos);
    if (textWidth > maxTextWidth)
    {
      if (posLast > 0)
      {
        lines.push_back(trim(lines[lineCount - 1].substr(posLast + 1)));
        lines[lineCount - 1].resize(posLast);
      }
      lineCount++;
    }
  }
  else if (cutTooLong)
  {
    pos = lines[lineCount - 1].length();
    while (font->WidthS(lines[lineCount - 1], pos) > maxTextWidth)
    {
      pos = lines[lineCount - 1].rfind(' ', pos - 1);
      if (pos == -1)
        break;
    }
    if (pos != -1)
      lines[lineCount - 1].resize(pos);
  }
  else
    return maxTextWidth;

  textWidth = 0;
  for (int i = 0; i < lineCount; i++)
    textWidth = max(textWidth, font->WidthS(lines[i]));
  textWidth = min(textWidth, maxTextWidth);

  return textWidth;
}

