/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  framebuffer.c   -  framebuffer device
 *                     Output goes to a framebuffer device
 *
 *  (c) 2004 Stephan Skrodzki
 *           based on simlcd.c by 
 *  (c) 2001-2004 Carsten Siebholz <c.siebholz AT t-online de>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <vdr/tools.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include "framebuffer.h"
#include "fct.h"



cGraphLCDDriverFramebuffer::cGraphLCDDriverFramebuffer(cGraphLCDSetup * setup)
  : setup(setup)
  , offbuff(0)
  , fbfd(-1)
{
  oldSetup = new cGraphLCDSetup();
  oldSetup->CopyFrom(setup);

  sizeX = setup->SizeX;
  if (sizeX < 0)
    sizeX = 0;
  sizeY = setup->SizeY;
  if (sizeY < 0)
    sizeY = 0;
  zoom = setup->Zoom;
  if (zoom != 1)
    zoom = 0;
}

cGraphLCDDriverFramebuffer::~cGraphLCDDriverFramebuffer()
{
  if(offbuff);
    free(offbuff);
  munmap(fbp, screensize);
  if (-1 != fbfd)
    close(fbfd);
  delete oldSetup;
}


int cGraphLCDDriverFramebuffer::Init()
{
  oldSetup->CopyFrom(setup);

  /* Open the file for reading and writing */
  fbfd = open("/dev/fb0", O_RDWR);
  if (1 == fbfd) {
    esyslog("Error: cannot open framebuffer device.\n");
    return -1;
  }
  isyslog("The framebuffer device was opened successfully.\n");

  /* Get fixed screen information */
  if (ioctl(fbfd, FBIOGET_FSCREENINFO, &finfo)) {
    esyslog("Error reading fixed information.\n");
    return -1;
  }

  /* Get variable screen information */
  if (ioctl(fbfd, FBIOGET_VSCREENINFO, &vinfo)) {
    esyslog("Error reading variable information.\n");
    return -1;
  }

  /* Figure out the size of the screen in bytes */
  screensize = vinfo.xres * vinfo.yres * vinfo.bits_per_pixel / 8;

  isyslog("V01: xres: %d, yres %d, vyres: %d, bpp: %d, linelenght: %d\n",vinfo.xres,vinfo.yres,vinfo.yres_virtual,vinfo.bits_per_pixel,finfo.line_length);

  /* reserve another memory to draw into */
  offbuff = (char *)malloc(screensize);
  if (!offbuff) {
    esyslog("Error: failed to alloc memory for framebuffer device.\n"); return -1;
  }
 
  /* Map the device to memory */
  fbp = (char *)mmap(0, screensize, PROT_READ | PROT_WRITE, MAP_SHARED,
                     fbfd, 0);
  if ((int)fbp == -1) {
    esyslog("Error: failed to map framebuffer device to memory.\n"); return -1;
  }
  isyslog("The framebuffer device was mapped to memory successfully.\n");
  // clear display
  Clear();
  Refresh(true);
                                
  isyslog("graphlcd plugin: Framebuffer initialized.\n");
  return 0;
}


int cGraphLCDDriverFramebuffer::DeInit()
{
  return 0;
}


void cGraphLCDDriverFramebuffer::CheckSetup()
{
  if (setup->DiffersImportantlyWith(oldSetup))
  {
    DeInit();
    Init();
  }
}


void cGraphLCDDriverFramebuffer::Clear()
{
  memset(offbuff, 0, screensize);
}

void cGraphLCDDriverFramebuffer::SetPixel(int x, int y, eColor color)
{
  int location;
  int outcol;

  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  if (setup->UpsideDown)
  {
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
  }

  /* Figure out where in memory to put the pixel */
  location = (x*(1+zoom)+vinfo.xoffset) * (vinfo.bits_per_pixel/8) +
             (y*(1+zoom)+vinfo.yoffset) * finfo.line_length;

  if (color == colorBlack)
  {
    if (vinfo.bits_per_pixel <= 8)
    {
      outcol = 15;
    }
    else
    {
      outcol = 255; 
    }
  }
  else
  {
    outcol = 0;
  }

  if (vinfo.bits_per_pixel <= 8) 
  {
    *(offbuff + location) = outcol;
    if (zoom == 1)
    {
      *(offbuff + location + 1) = outcol;
      *(offbuff + location + finfo.line_length) = outcol;
      *(offbuff + location + finfo.line_length + 1) = outcol;
    }
  }
  else
  {
    if (vinfo.bits_per_pixel <= 16)
    {
      *(offbuff + location) = outcol;
      *(offbuff + location + 1) = outcol;
      if (zoom == 1)
      {
        *(offbuff + location + 2) = outcol;
        *(offbuff + location + 3) = outcol;
        *(offbuff + location + finfo.line_length) = outcol;
        *(offbuff + location + finfo.line_length + 1) = outcol;
        *(offbuff + location + finfo.line_length + 2) = outcol;
        *(offbuff + location + finfo.line_length + 3) = outcol;
      }
    } 
    else 
    {
      *(offbuff + location) = outcol;
      *(offbuff + location + 1) = outcol;
      *(offbuff + location + 2) = outcol;
      *(offbuff + location + 3) = 0;     /* should be transparency */
      if (zoom == 1)
      {
        *(offbuff + location + 4) = outcol;
        *(offbuff + location + 5) = outcol;
        *(offbuff + location + 6) = outcol;
        *(offbuff + location + 7) = 0;
        *(offbuff + location + finfo.line_length) = outcol;
        *(offbuff + location + finfo.line_length + 1) = outcol;
        *(offbuff + location + finfo.line_length + 2) = outcol;
        *(offbuff + location + finfo.line_length + 3) = 0;
        *(offbuff + location + finfo.line_length + 4) = outcol;
        *(offbuff + location + finfo.line_length + 5) = outcol;
        *(offbuff + location + finfo.line_length + 6) = outcol;
        *(offbuff + location + finfo.line_length + 7) = 0;
      }
    }
  }
}


void cGraphLCDDriverFramebuffer::Set8Pixels(int x, int y, byte data, eColor color)
{
//  clip(x, 0, sizeX - 1);
//  clip(y, 0, sizeY - 1);
  int n;

  x &= 0xFFF8;

  for (n = 0; n < 8; ++n)
  {
    if (data & (0x80 >> n))      // if bit is set
      SetPixel(x + n, y, color);
  }
}


void cGraphLCDDriverFramebuffer::Refresh(bool refreshAll)
{
  memcpy(fbp, offbuff, screensize);
}
