/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 * gu140x32f.c - 8-bit driver module for Noritake GU140x32-F7806
 *               VFD displays. The VFD is operated in it's 8 bit-mode to be
 *               connected to a single PC parallel port.
 *
 * (c) 2003-2004 Andreas Brachold <vdr04 AT deltab de>
 *  based on: HD61830 device
 *              (c) 2001-2003 by Carsten Siebholz <c.siebholz AT t-online de>
 *            lcdproc 0.4 driver hd44780-ext8bit
 *              (c) 1999, 1995 Benjamin Tse <blt AT Comports com>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <unistd.h>
#include <sys/time.h>
#include <vdr/tools.h>

#include "gu140x32f.h"
#include "fct.h"

// Defines for hd44780 Displays
#define RS_DAT      0x00
#define RS_CMD      0x01

#define CLEAR       0x01

#define HOMECURSOR  0x02

#define ENTRYMODE   0x04
#define E_MOVERIGHT 0x02
#define E_MOVELEFT  0x00
#define EDGESCROLL  0x01
#define NOSCROLL    0x00

#define ONOFFCTRL   0x08
#define DISPON      0x04
#define DISPOFF     0x00
#define CURSORON    0x02
#define CURSOROFF   0x00
#define CURSORBLINK 0x01
#define CURSORNOBLINK 0x00

#define CURSORSHIFT 0x10
#define SCROLLDISP  0x08
#define MOVECURSOR  0x00
#define MOVERIGHT   0x04
#define MOVELEFT    0x00

#define FUNCSET     0x20
#define IF_8BIT     0x10
#define IF_4BIT     0x00

// Control output lines
// Write to baseaddress+2
#define nSTRB       0x01    // pin 1; negative logic
#define nLF         0x02    // pin 14
#define nINIT       0x04    // pin 16; the only positive logic output line
#define nSEL        0x08    // pin 17
#define ENIRQ       0x10    // Enable IRQ via ACK line (don't enable this withouT
                            // setting up interrupt stuff too)
#define ENBI        0x20    // Enable bi-directional port (is nice to play with!
                            // I first didn't know a SPP could do this)

#define OUTMASK     0x0B    // SEL, LF and STRB are hardware inverted
                            // Use this mask only for the control output lines
                            // XOR with this mask ( ^ OUTMASK )

static const int kWiringStandard = 0;
// #define RS        nSTRB
// #define RW        nLF
// #define EN1       nINIT
// #define BL        nSEL
static const int kWiringWindows  = 1;
// #define RS        nINIT
// #define RW        nLF
// #define EN1       nSTRB
// #define BL        nSEL


cGraphLCDDriverGU140X32F::cGraphLCDDriverGU140X32F(cGraphLCDSetup * setup)
: m_WiringRS(nSTRB)
, m_WiringEN1(nINIT)
//, m_WiringRW(nLF)
//, m_WiringBL(nSEL)
, setup(setup)
{
  oldSetup = new cGraphLCDSetup();
  oldSetup->CopyFrom(setup);

  port = new cParallelPort();

  sizeX = setup->SizeX;         // 140
  sizeY = setup->SizeY;         // 32
  m_iSizeYb = ((sizeY + 7) / 8); // 4
  m_bSleepIsInit = false;

  // setup the memory array for the drawing array gu140x32f
  m_pDrawMem = new byte[sizeX * m_iSizeYb];
  Clear();

  // setup the memory array for the display array gu140x32f
  m_pVFDMem = new byte[sizeX * m_iSizeYb];
  ClearVFDMem();

  m_nTimingAdjustCmd = 0;
  m_nRefreshCounter = 0;
}


void cGraphLCDDriverGU140X32F::ClearVFDMem()
{
  for (int n = 0; m_pVFDMem && n < (sizeX * m_iSizeYb); n++)
    m_pVFDMem[n] = 0x00;
}


void cGraphLCDDriverGU140X32F::Clear()
{
  for (int n = 0; m_pDrawMem && n < (sizeX * m_iSizeYb); n++)
    m_pDrawMem[n] = 0x00;
}


cGraphLCDDriverGU140X32F::~cGraphLCDDriverGU140X32F()
{
  if (m_pVFDMem)
    delete[] m_pVFDMem;
  if(m_pDrawMem)
    delete[] m_pDrawMem;
  delete oldSetup;
  delete port;
}


int cGraphLCDDriverGU140X32F::Init()
{
  int x;
  struct timeval tv1, tv2;

  if (setup->Wiring == kWiringWindows) { // 0 == Standard, 1 == Windows
    m_WiringRS = nINIT;
    m_WiringEN1 = nSTRB;
    //m_WiringRW = nLF;
    //m_WiringBL = nSEL;
  }
  else {
    m_WiringRS = nSTRB;
    m_WiringEN1 = nINIT;
    //m_WiringRW = nLF;
    //m_WiringBL = nSEL;
  }

  if (!setup->PPDevice)
  {
    // use DirectIO
    if (port->Open(setup->PPPort) != 0)
      return -1;
    uSleep(10);
  }
  else
  {
    // use ppdev
    if (port->Open(setup->PPDevice) != 0)
      return -1;
  }

  if (nSleepInit() != 0)
  {
    esyslog("graphlcd plugin: INFO: cannot change wait parameters  Err:%s (cGraphLCDDriver::Init)\n",strerror(errno));
    m_bSleepIsInit = false;
  }
  else
  {
    m_bSleepIsInit = true;
  }

  dsyslog("graphlcd plugin: benchmark started.\n");
  gettimeofday(&tv1, 0);
  for (x = 0; x < 10000; x++)
  {
    port->WriteData(x % 0x100);
  }
  gettimeofday(&tv2, 0);
  nSleepDeInit();
  m_nTimingAdjustCmd = ((tv2.tv_sec - tv1.tv_sec) * 10000 + (tv2.tv_usec - tv1.tv_usec)) / 1000;
  dsyslog("graphlcd plugin: benchmark stopped. Time for Port Command: %ldns\n",
      m_nTimingAdjustCmd);


  // setup the lcd in 8 bit mode
  Write(RS_CMD, FUNCSET | IF_8BIT, 4100);
  Write(RS_CMD, FUNCSET | IF_8BIT, 100);
  Write(RS_CMD, FUNCSET | IF_8BIT, 40);

  Write(RS_CMD, ONOFFCTRL | DISPON | CURSOROFF | CURSORNOBLINK, 40);
  Write(RS_CMD, CLEAR, 1600);
  Write(RS_CMD, HOMECURSOR, 1600);

  port->Release();

  oldSetup->CopyFrom(setup);

  // Set Display SetBrightness
  SetBrightness(setup->Brightness);
  // clear display
  ClearVFDMem();
  Refresh(true);

  isyslog("graphlcd plugin: gu140x32f initialized.\n");
  return 0;
}


void cGraphLCDDriverGU140X32F::SetBrightness(unsigned int percent)
{
  port->Claim();

  byte level;
  if (percent > 100)
    percent = 100;
  level = percent / 25;
  if (level < 1)
    level = 1;
  level = (4 - level) & 0x03;
  /* Set Brightness
      00 - 100%
      01 -  75%
      02 -  50%
      03 -  25%  */
  Write(RS_CMD, FUNCSET | IF_8BIT, 40);
  Write(RS_DAT, level, 40);

  port->Release();
}


int cGraphLCDDriverGU140X32F::DeInit()
{
  if (port->Close() != 0)
    return -1;
  return 0;
}


void cGraphLCDDriverGU140X32F::CheckSetup()
{
  bool bBrightnessOnly = (setup->Brightness != oldSetup->Brightness);

  if (setup->DiffersImportantlyWith(oldSetup))
  {
    DeInit();
    Init();
  }
  else if (bBrightnessOnly)
  { //Change throw Setup, if isn't update with DiffersImportantlyWith
    SetBrightness(setup->Brightness);
    oldSetup->Brightness = setup->Brightness;
  }
}


void cGraphLCDDriverGU140X32F::Write(byte nFlags, byte bData, unsigned int nMicroSecBusyTime)
{
  if (m_bSleepIsInit)
    nSleepInit();

  byte enableLines = 0, portControl;

  // Only one controller is supported
  enableLines = m_WiringEN1;

  if (nFlags == RS_CMD)
    portControl = 0;
  else                          //if (nFlags == RS_DAT)
    portControl = m_WiringRS;

  // portControl |= m_WiringBL;

  port->WriteControl(portControl ^ OUTMASK);                  //Reset controlbits
  port->WriteData(bData);                                     //Set data
  port->WriteControl((enableLines | portControl) ^ OUTMASK);  //Set controlbits

  // How long hold the data active
  if (m_bSleepIsInit && (25 + (100 * setup->AdjustTiming) - m_nTimingAdjustCmd > 0))
  { //Wait 50ns
    nSleep(max(25L, 50 + (100 * setup->AdjustTiming) - m_nTimingAdjustCmd));
  }

  port->WriteControl(portControl ^ OUTMASK);                  //Reset controlbits

  nSleep((nMicroSecBusyTime * 1000) + (100 * setup->AdjustTiming) -
      m_nTimingAdjustCmd);

  if (m_bSleepIsInit)
    nSleepDeInit();
}


void cGraphLCDDriverGU140X32F::SetPixel(int x, int y, eColor color)
{
  byte c;
  int n;

  if (!m_pDrawMem)
    return;

  if(x >= sizeX || x < 0)
    return;
  if(y >= sizeY || y < 0)
    return;

  if (setup->UpsideDown)
  {
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
  }

  n = x + ((y / 8) * sizeX);
  c = 0x80 >> (y % 8);

  if (color == colorBlack)
  {
    m_pDrawMem[n] |= c;
  }
  else
  {
    m_pDrawMem[n] &= ~c;
  }
}


void cGraphLCDDriverGU140X32F::Set8Pixels(int x, int y, byte data, eColor color)
{
  int n;

  // x - pos is'nt mayby align to 8
  x &= 0xFFF8;

  for (n = 0; n < 8; ++n)
  {
    if (data & (0x80 >> n))      // if bit is set
      SetPixel(x + n, y, color);
  }
}


void cGraphLCDDriverGU140X32F::Refresh(bool refreshAll)
{
  int n, x, yb;

  CheckSetup();

  if (!m_pVFDMem || !m_pDrawMem)
    return;

  bool doRefresh = false;
  int minX = sizeX;
  int maxX = 0;
  int minYb = m_iSizeYb;
  int maxYb = 0;

  for (yb = 0; yb < m_iSizeYb; ++yb)
    for (x = 0; x < sizeX; ++x)
    {
      n = x + (yb * sizeX);
      if (m_pVFDMem[n] != m_pDrawMem[n])
      {
        m_pVFDMem[n] = m_pDrawMem[n];
        minX = min(minX, x);
        maxX = max(maxX, x);
        minYb = min(minYb, yb);
        maxYb = max(maxYb, yb + 1);
        doRefresh = true;
      }
    }

  m_nRefreshCounter = (m_nRefreshCounter + 1) % setup->RefreshDisplay;
  if (!refreshAll && !m_nRefreshCounter)
    refreshAll = true;

  if (refreshAll || doRefresh)
  {
    if (refreshAll) {
      minX = 0;
      maxX = sizeX;
      minYb = 0;
      maxYb = m_iSizeYb;
      // and reset RefreshCounter
      m_nRefreshCounter = 0;
    }

    minX = max(minX, 0);
    maxX = min(maxX, sizeX - 1);
    minYb = max(minYb, 0);
    maxYb = min(maxYb, m_iSizeYb);

/*  FILE *f = fopen("/tmp/gu.log","a+");
    fprintf(f,"Refresh %d,iMinX %d iMaxX %d, iMinYb %d, iMaxYb %d\n",bRefreshAll?1:0,iMinX,iMaxX,iMinYb,iMaxYb);
    fclose(f); */

    port->Claim();
    // send lcd data to display, controller
    Write(RS_CMD, 0xF1, 40);
    Write(RS_DAT, minX, 40);
    Write(RS_DAT, (minYb * 8) & 0xFFF8, 40);
    Write(RS_DAT, maxX, 40);
    Write(RS_DAT, (maxYb * 8), 40);

    Write(RS_DAT, 'v', 500);

    for (yb = minYb; yb <= maxYb; ++yb)
      for (x = minX; x <= maxX; ++x)
      {
        n = x + (yb * sizeX);

        if (n >= (sizeX * m_iSizeYb))
          break;
        Write(RS_DAT, (m_pVFDMem[n]) ^ ((setup->Invert != 0) ? 0xff : 0x00), 40);
      }
    port->Release();
  }
}
