/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 * gu256x64-372.c - 8-bit driver module for Noritake GU256X64-372
 *               VFD displays. The VFD is operated in it's 8 bit-mode to be
 *               connected to a single PC parallel port.
 *
 * (c) 20040411 andreas 'randy' weinberger (randy AT smue org)
 *
 *  based on:
 *            gu256x32f driver module for graphlcd
 *              (c) 2003 Andreas Brachold <vdr04 AT deltab de>
 *            HD61830 device
 *              (c) 2001-2003 by Carsten Siebholz <c.siebholz AT t-online de>
 *            lcdproc 0.4 driver hd44780-ext8bit
 *              (c) 1999, 1995 Benjamin Tse <blt AT Comports com>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <unistd.h>
#include <sys/time.h>
#include <vdr/tools.h>

#include "gu256x64-372.h"
#include "fct.h"

#define SCREENSOFF	0x00	// both screens are off
#define SCREEN1ON	0x01	// only screen #1 is on (graphic screen)
#define SCREEN2ON	0x02	// only screen #2 is on (graphic/character screen)
#define SCREENSON	0x03	// both screens are on

#define CURS_AUTOINC	0x04	// cursor increments automatically
#define CURS_HOLD	0x05	// cursor holds

#define SCREEN2CHAR	0x06	// screen #2 sets to "character" display
#define SCREEN2GRAPH	0x07	// screen #2 sets to "graphic" display

#define DATA_WRITE	0x08	// data write mode
#define DATA_READ	0x09	// data read mode

#define DISP_LOSTA1	0x0A	// lower addr. of display start of screen #1
#define DISP_HISTA1	0x0B	// upper addr. of display start of screen #1
#define DISP_LOSTA2	0x0C	// lower addr. of display start of screen #2
#define DISP_HISTA2	0x0D	// upper addr. of display start of screen #2
#define CURS_LOADDR	0x0E	// lower addr. of cursor of screen #1 & #2
#define CURS_HIADDR	0x0F	// upper addr. of cursor start of screen #1 & #2

#define DISP_OR		0x10	// or display of screen #1 & #2
#define DISP_EXOR	0x11	// ex-or display of screen #1 & #2
#define DISP_AND	0x12	// and display of screen #1 & #2

#define BRIGHT_1	0x18	// luminance level 1 100.0%
#define BRIGHT_2	0x19	// luminance level 2  87.5%
#define BRIGHT_3	0x1A	// luminance level 3  75.0%
#define BRIGHT_4	0x1B	// luminance level 4  62.5%

#define WRHI	0x04 	// INIT
#define WRLO	0x00
#define CDHI	0x00	// nSEL
#define CDLO	0x08

cGraphLCDDriverGU256X64_372::cGraphLCDDriverGU256X64_372(cGraphLCDSetup * setup)
: setup(setup)
{
  int x=0;

  oldSetup = new cGraphLCDSetup();
  oldSetup->CopyFrom(setup);

  port = new cParallelPort();

  sizeX = setup->SizeX;         // 256
  sizeY = setup->SizeY;         // 64
  m_iSizeYb = ((sizeY + 7) / 8); // 8
  m_bSleepIsInit = false;

  // setup linear lcd array
  m_pDrawMem = new byte*[sizeX];
  if (m_pDrawMem)
  {
    for (x = 0; x < sizeX; x++)
    {
      m_pDrawMem[x] = new byte[(sizeY + 7) / 8];
      memset(m_pDrawMem[x], 0, (sizeY + 7) / 8);
    }
  }
  Clear();

  // setup the lcd array for the "vertikal" mem
  m_pVFDMem = new byte*[sizeX];
  if (m_pVFDMem)
  {
    for (x = 0; x < sizeX; x++)
    {
      m_pVFDMem[x] = new byte[(sizeY + 7) / 8];
      memset(m_pVFDMem[x], 0, (sizeY + 7) / 8);
    }
  }
  ClearVFDMem();

  m_nTimingAdjustCmd = 0;
  m_nRefreshCounter = 0;
}


void cGraphLCDDriverGU256X64_372::ClearVFDMem()
{
  for (int x = 0; x < sizeX; x++)
    memset(m_pVFDMem[x], 0, (sizeY + 7)/8);
}



void cGraphLCDDriverGU256X64_372::Clear()
{
  for (int x = 0; x < sizeX; x++)
    memset(m_pDrawMem[x], 0, (sizeY + 7)/8);
}


cGraphLCDDriverGU256X64_372::~cGraphLCDDriverGU256X64_372()
{
  int x;

  if (m_pVFDMem)
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      delete[] m_pVFDMem[x];
    }
    delete[] m_pVFDMem;
  if(m_pDrawMem)
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      delete[] m_pDrawMem[x];
    }
    delete[] m_pDrawMem;
  delete oldSetup;
  delete port;
}


int cGraphLCDDriverGU256X64_372::Init()
{
  int x;
  struct timeval tv1, tv2;

  if (!setup->PPDevice)
  {
    // use DirectIO
    if (port->Open(setup->PPPort) != 0)
      return -1;
    uSleep(10);
  }
  else
  {
    // use ppdev
    if (port->Open(setup->PPDevice) != 0)
      return -1;
  }

  if (nSleepInit() != 0)
  {
    esyslog("graphlcd plugin: INFO: cannot change wait parameters  Err:%s (cGraphLCDDriver::Init)\n",strerror(errno));
    m_bSleepIsInit = false;
  }
  else
  {
    m_bSleepIsInit = true;
  }

  dsyslog("graphlcd plugin: benchmark started.\n");
  gettimeofday(&tv1, 0);
  for (x = 0; x < 10000; x++)
  {
    port->WriteData(x % 0x100);
  }
  gettimeofday(&tv2, 0);
  nSleepDeInit();
  m_nTimingAdjustCmd = ((tv2.tv_sec - tv1.tv_sec) * 10000 + (tv2.tv_usec - tv1.tv_usec)) / 1000;
  dsyslog("graphlcd plugin: benchmark stopped. Time for Port Command: %ldns\n",
      m_nTimingAdjustCmd);

  GU256X64Cmd(SCREEN1ON);
  GU256X64Cmd(CURS_AUTOINC);
  GU256X64Cmd(SCREEN2CHAR);

  GU256X64Cmd(DISP_LOSTA1); GU256X64Data(0x00);
  GU256X64Cmd(DISP_HISTA1); GU256X64Data(0x00);
  GU256X64Cmd(DISP_LOSTA2); GU256X64Data(0x00);
  GU256X64Cmd(DISP_HISTA2); GU256X64Data(0x10);
  GU256X64Cmd(CURS_LOADDR); GU256X64Data(0x00);
  GU256X64Cmd(CURS_HIADDR); GU256X64Data(0x00);
 
  GU256X64Cmd(DISP_OR);

  port->Release();

  oldSetup->CopyFrom(setup);

  // Set Display SetBrightness
  SetBrightness(setup->Brightness);
  // clear display
  ClearVFDMem();
  Refresh(true);

  isyslog("graphlcd plugin: gu256x64-372 initialized.\n");
  return 0;
}


void cGraphLCDDriverGU256X64_372::SetBrightness(unsigned int percent)
{
  port->Claim();

  if (percent > 88) {
    GU256X64Cmd(BRIGHT_1);
  } else if (percent > 75) {
    GU256X64Cmd(BRIGHT_2);
  } else if (percent > 66) {
    GU256X64Cmd(BRIGHT_3);
  } else if (percent > 0 ) {
    GU256X64Cmd(BRIGHT_4);
  } else {
    GU256X64Cmd(SCREENSOFF);
  }
  port->Release();
}


int cGraphLCDDriverGU256X64_372::DeInit()
{
  if (port->Close() != 0)
    return -1;
  return 0;
}


void cGraphLCDDriverGU256X64_372::CheckSetup()
{
  bool bBrightnessOnly = (setup->Brightness != oldSetup->Brightness);

  if (setup->DiffersImportantlyWith(oldSetup))
  {
    DeInit();
    Init();
  }
  else if (bBrightnessOnly)
  { //Change throw Setup, if isn't update with DiffersImportantlyWith
    SetBrightness(setup->Brightness);
    oldSetup->Brightness = setup->Brightness;
  }
}

void cGraphLCDDriverGU256X64_372::GU256X64Cmd(byte data)
{
  if (m_bSleepIsInit)
    nSleepInit();

  port->WriteControl(CDHI | WRHI);
  port->WriteData(data);
  nSleep(100 + (100 * setup->AdjustTiming) - m_nTimingAdjustCmd);
  port->WriteControl(CDHI | WRLO);
  nSleep(100 + (100 * setup->AdjustTiming) - m_nTimingAdjustCmd);
  port->WriteControl(CDHI | WRHI);
  nSleep(100 + (100 * setup->AdjustTiming) - m_nTimingAdjustCmd);
} 

void cGraphLCDDriverGU256X64_372::GU256X64Data(byte data)
{
  if (m_bSleepIsInit)
    nSleepInit();

  port->WriteControl(CDLO | WRHI);
  port->WriteData(data);
  nSleep(100 + (100 * setup->AdjustTiming) - m_nTimingAdjustCmd);
  port->WriteControl(CDLO | WRLO);
  nSleep(100 + (100 * setup->AdjustTiming) - m_nTimingAdjustCmd);
  port->WriteControl(CDLO | WRHI);
  nSleep(100 + (100 * setup->AdjustTiming) - m_nTimingAdjustCmd);
} 

void cGraphLCDDriverGU256X64_372::SetPixel(int x, int y, eColor color)
{
  byte c;

  if (!m_pDrawMem)
    return;

  if(x >= sizeX || x < 0)
    return;
  if(y >= sizeY || y < 0)
    return;

  if (setup->UpsideDown)
  {
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
  }

  c = 0x80 >> (y % 8);

  if (color == colorBlack)
  {
    m_pDrawMem[x][y/8] = m_pDrawMem[x][y/8] | c;
  }
  else
  {
    m_pDrawMem[x][y/8] = m_pDrawMem[x][y/8] & ~c;
  }
}


void cGraphLCDDriverGU256X64_372::Set8Pixels(int x, int y, byte data, eColor color)
{
  int n;

  // x - pos is'nt mayby align to 8
  x &= 0xFFF8;

  for (n = 0; n < 8; ++n)
  {
    if (data & (0x80 >> n))      // if bit is set
      SetPixel(x + n, y, color);
  }
}


void cGraphLCDDriverGU256X64_372::Refresh(bool refreshAll)
{
  int xb, yb;
  
  CheckSetup();

  if (!m_pVFDMem || !m_pDrawMem)
    return;

  bool doRefresh = false;
  int minX = sizeX;
  int maxX = 0;
  int minYb = m_iSizeYb;
  int maxYb = 0;

  for (xb = 0; xb < sizeX; ++xb)
  {
    for (yb = 0; yb < (sizeY+7)/8; ++yb)
    {
      if (m_pVFDMem[xb][yb] != m_pDrawMem[xb][yb])
      {
        m_pVFDMem[xb][yb] = m_pDrawMem[xb][yb];
        minX = min(minX, xb);
        maxX = max(maxX, xb);
        minYb = min(minYb, yb);
        maxYb = max(maxYb, yb + 1);
        doRefresh = true;
      }
    }
  }
  
  m_nRefreshCounter = (m_nRefreshCounter + 1) % setup->RefreshDisplay;
  if (!refreshAll && !m_nRefreshCounter)
    refreshAll = true;

  if (refreshAll || doRefresh)
  {
    if (refreshAll) {
      minX = 0;
      maxX = sizeX;
      minYb = 0;
      maxYb = m_iSizeYb;
      // and reset RefreshCounter
      m_nRefreshCounter = 0;
    }

    minX = max(minX, 0);
    maxX = min(maxX, sizeX - 1);
    minYb = max(minYb, 0);
    maxYb = min(maxYb, m_iSizeYb);

    port->Claim();

    GU256X64Cmd(CURS_LOADDR); GU256X64Data(0x00);
    GU256X64Cmd(CURS_HIADDR); GU256X64Data(0x00);
    GU256X64Cmd(DATA_WRITE);

    for (xb = 0; xb < sizeX; xb++)
    {
      for (yb = 0; yb < (sizeY+7)/8; yb++ )
      {
        GU256X64Data( (m_pVFDMem[xb][yb]) ^ ((setup->Invert != 0) ? 0xff : 0x00) );
      }
    }
    port->Release();
  }
}
