/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 * gu256x64-3900.c - 8-bit driver module for Noritake GU256X64X-3XX0
 *               VFD displays. The VFD is operated in it's 8 bit-mode to be
 *               connected to a single PC parallel port.
 *
 * (c) 20040810 Ralf Mueller (ralf AT bj-ig de)
 *
 *  based on:
 *            gu256x64-372 driver module for graphlcd
 *              (c) 2004 Andreas 'randy' Weinberger <randy AT smue org>
 *            gu256x32f driver module for graphlcd
 *              (c) 2003 Andreas Brachold <vdr04 AT deltab de>
 *            HD61830 device
 *              (c) 2001-2003 by Carsten Siebholz <c.siebholz AT t-online de>
 *            lcdproc 0.4 driver hd44780-ext8bit
 *              (c) 1999, 1995 Benjamin Tse <blt AT Comports com>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <unistd.h>
#include <termios.h>
#include <sys/time.h>
#include <vdr/tools.h>

#include "gu256x64-3900.h"
#include "fct.h"

static const byte dSETSTART0      = 0x02;    // Set display start address (CMD prefix)
static const byte dSETSTART1      = 0x44;    // Set display start address (CMD)
static const byte dSETSTART2      = 0x00;    // Set display start address
static const byte dSETSTART3      = 0x53;    // Set display start address

static const byte nINITDISPLAY0   = 0x1b;    // Initialize display (CMD prefix)
static const byte nINITDISPLAY1   = 0x40;    // Initialize display (CMD)

static const byte nPOWER0         = 0x1f;    // Power on/off
static const byte nPOWER1         = 0x28;    // Power on/off
static const byte nPOWER2         = 0x61;    // Power on/off
static const byte nPOWER3         = 0x40;    // Power on/off

static const byte nSETBRIGHT0     = 0x1f;    // Set brightness (CMD prefix)
static const byte nSETBRIGHT1     = 0x58;    // Set brightness (CMD)

static const byte dSETBRIGHT0     = 0x02;    // Set brightness (CMD prefix)
static const byte dSETBRIGHT1     = 0x44;    // Set brightness (CMD)
static const byte dSETBRIGHT2     = 0x00;    // Set brightness
static const byte dSETBRIGHT3     = 0x58;    // Set brightness

static const byte BRIGHT_000      = 0x10;    // Brightness 0%
static const byte BRIGHT_012      = 0x11;    // Brightness 12.5%
static const byte BRIGHT_025      = 0x12;    // Brightness 25%
static const byte BRIGHT_037      = 0x13;    // Brightness 37.5%
static const byte BRIGHT_050      = 0x14;    // Brightness 50%
static const byte BRIGHT_062      = 0x15;    // Brightness 62.5%
static const byte BRIGHT_075      = 0x16;    // Brightness 75%
static const byte BRIGHT_087      = 0x17;    // Brightness 87.5%
static const byte BRIGHT_100      = 0x18;    // Brightness 100%

static const byte nSETCURSOR0     = 0x1f;    // Set cursor (CMD prefix)
static const byte nSETCURSOR1     = 0x24;    // Set cursor (CMD)

static const byte nDISPLAYCURSOR0 = 0x1f;    // display cursor
static const byte nDISPLAYCURSOR1 = 0x43;    // display cursor

static const byte nBLITIMAGE0     = 0x1f;    // Display image
static const byte nBLITIMAGE1     = 0x28;    // Display image
static const byte nBLITIMAGE2     = 0x66;    // Display image
static const byte nBLITIMAGE3     = 0x11;    // Display image

static const byte dBLITIMAGE0     = 0x02;    // Image write
static const byte dBLITIMAGE1     = 0x44;    // Image write
static const byte dBLITIMAGE2     = 0x00;    // Image write
static const byte dBLITIMAGE3     = 0x46;    // Image write


static const byte WRHI            = 0x0f;    // any control pin to high
static const byte WRLO            = 0x00;
static const byte RDYHI           = 0x40;    // RDY
static const byte RDYHIALT        = 0x80;    // RDY satyr wiring
static const byte RDYLO           = 0x00;

static const int kDMA             = 0x00;      // parallel DMA mode
static const int kNormal          = 0x01;      // normal parallel mode
static const int kSerial          = 0x02;      // normal serial mode

static const int kDefault         = 0x00;      // wiring mode
static const int kSatyr           = 0x01;      // different wiring mode by satyr


cGraphLCDDriverGU256X64_3900::cGraphLCDDriverGU256X64_3900(cGraphLCDSetup * setup)
: setup(setup)
{
  int x=0;

  oldSetup = new cGraphLCDSetup();
  oldSetup->CopyFrom(setup);

  // lower 4 bit for switching parallel/seriell/DMA mode
  // higher 4 bit for switching different wiring setups
  wiring=(0x03&setup->Wiring);

  portFd = -1;
  if(wiring!=kSerial) {
    port = new cParallelPort();
  } else {
    port = 0;
  }

  sizeX = setup->SizeX;         // 256
  sizeY = setup->SizeY;         // 64
  m_iSizeYb = ((sizeY + 7) / 8); // 8
  m_bSleepIsInit = false;

  // setup linear lcd array
  m_pDrawMem = new byte*[sizeX];
  if(m_pDrawMem) {
    for(x = 0; x < sizeX; x++) {
      m_pDrawMem[x] = new byte[m_iSizeYb];
      memset(m_pDrawMem[x], 0, m_iSizeYb);
    }
  }
  Clear();

  // setup the lcd array for the "vertikal" mem
  m_pVFDMem = new byte*[sizeX];
  if(m_pVFDMem) {
    for(x = 0; x < sizeX; x++) {
      m_pVFDMem[x] = new byte[m_iSizeYb];
      memset(m_pVFDMem[x], 0, m_iSizeYb);
    }
  }
  ClearVFDMem();

  m_nTimingAdjustCmd = 0;
  m_nRefreshCounter = 0;
}


void cGraphLCDDriverGU256X64_3900::ClearVFDMem()
{
  for(int x = 0; x < sizeX; x++) {
    memset(m_pVFDMem[x], 0, m_iSizeYb);
  }
}



void cGraphLCDDriverGU256X64_3900::Clear()
{
  for(int x = 0; x < sizeX; x++) {
    memset(m_pDrawMem[x], 0, m_iSizeYb);
  }
}


cGraphLCDDriverGU256X64_3900::~cGraphLCDDriverGU256X64_3900()
{
  int x;

  if(m_pVFDMem) {
    for(x = 0; x < (sizeX + 7) / 8; x++) {
      delete[] m_pVFDMem[x];
    }
    delete[] m_pVFDMem;
  }
  if(m_pDrawMem) {
    for(x = 0; x < (sizeX + 7) / 8; x++) {
      delete[] m_pDrawMem[x];
    }
    delete[] m_pDrawMem;
  }
  delete oldSetup;
  delete port;

  if(portFd>=0) {
    close(portFd);
    portFd=-1;
  }
}


int cGraphLCDDriverGU256X64_3900::InitSerialPort() {
  if(!setup->PPDevice) {
    isyslog("graphlcd plugin: unable to initialize gu256x64-3900!\n");
    return -1;
  }
  portFd = open(setup->PPDevice, O_RDWR | O_NOCTTY);
  if(portFd>=0) {
    struct termios options;
    tcgetattr(portFd, &options);
    cfsetispeed(&options, B38400);
    cfsetospeed(&options, B38400);
    options.c_cflag &= ~CSIZE;
    options.c_cflag &= ~PARENB;
    options.c_cflag &= ~CSTOPB;
    options.c_cflag |= CS8;
    tcsetattr(portFd, TCSANOW, &options);
  } else {
    isyslog("graphlcd plugin: unable to initialize gu256x64-3900!\n");
    return -1;
  }
  return 0;
}

int cGraphLCDDriverGU256X64_3900::InitParallelPort() {
  if(!setup->PPDevice) {
    // use DirectIO
    if(port->Open(setup->PPPort) != 0) {
      isyslog("graphlcd plugin: unable to initialize gu256x64-3900!\n");
      return -1;
    }
    uSleep(10);
  } else {
    // use ppdev
    if(port->Open(setup->PPDevice) != 0) {
      isyslog("graphlcd plugin: unable to initialize gu256x64-3900!\n");
      return -1;
    }
  }

  if(nSleepInit() != 0) {
    esyslog("graphlcd plugin: INFO: cannot change wait parameters  Err:%s (cGraphLCDDriver::Init)\n", strerror(errno));
    m_bSleepIsInit = false;
  } else {
    m_bSleepIsInit = true;
  }

  port->Claim();

  dsyslog("graphlcd plugin: benchmark started.\n");
  struct timeval tv1, tv2;
  gettimeofday(&tv1, 0);
  int iterations=10000;
  for (int x = 0; x < iterations; x++) {
    port->WriteData((x % 0x100));
  }
  gettimeofday(&tv2, 0);
  nSleepDeInit();
  m_nTimingAdjustCmd = ((tv2.tv_sec - tv1.tv_sec)*1000000 + (tv2.tv_usec - tv1.tv_usec)) / (iterations/1000);
  dsyslog("graphlcd plugin: benchmark stopped. Time for Port Command: %ldns\n",
      m_nTimingAdjustCmd);
  return 0;
}

void cGraphLCDDriverGU256X64_3900::InitNormalDisplay() {
  Write(nPOWER0);
  Write(nPOWER1);
  Write(nPOWER2);
  Write(nPOWER3);
  Write(1); // power on

  Write(nINITDISPLAY0);
  Write(nINITDISPLAY1);

  Write(nDISPLAYCURSOR0);
  Write(nDISPLAYCURSOR1);
  Write(0); // off

  Write(nSETCURSOR0);
  Write(nSETCURSOR1);
  Write(0); // low byte x
  Write(0); // high byte x
  Write(0); // low byte y
  Write(0); // high byte y
}

void cGraphLCDDriverGU256X64_3900::InitDMADisplay() {
  Write(dSETSTART0);
  Write(dSETSTART1);
  Write(dSETSTART2);
  Write(dSETSTART3);
  Write(0);
  Write(0);
}

int cGraphLCDDriverGU256X64_3900::Init()
{
  if(wiring>kSerial) {
    esyslog("graphlcd plugin: ERROR: illegal wiring parameter: %d\n", setup->Wiring);
    return -1;
  }
  switch(setup->Wiring>>2) {
    case kDefault:
      readyMask=RDYHI;
      readyHi=RDYHI;
      break;
    case kSatyr:
      readyMask=RDYHIALT;
      readyHi=RDYLO;
      break;
    default:
      esyslog("graphlcd plugin: ERROR: illegal wiring parameter: %d\n", setup->Wiring);
      return -1;
  }

  if(wiring==kSerial) {
    if(InitSerialPort()<0) {
      return -1;
    }
  } else {
    if(InitParallelPort()<0) {
      return -1;
    }
  }

  if(wiring==kDMA) {
    InitDMADisplay();
  } else {
    InitNormalDisplay();
  }

  if(wiring!=kSerial) {
    // claim is in InitParallelPort
    port->Release();
  }

  oldSetup->CopyFrom(setup);

  // Set Display SetBrightness
  SetBrightness(setup->Brightness);
  // clear display
  ClearVFDMem();
  Refresh(true);

  isyslog("graphlcd plugin: gu256x64-3900 initialized.\n");
  return 0;
}


void cGraphLCDDriverGU256X64_3900::SetBrightness(unsigned int percent)
{
  if(wiring!=kSerial) {
    port->Claim();
  }

  if(wiring==kDMA) {
    Write(dSETBRIGHT0);
    Write(dSETBRIGHT1);
    Write(dSETBRIGHT2);
    Write(dSETBRIGHT3);
  } else {
    Write(nSETBRIGHT0);
    Write(nSETBRIGHT1);
  }
  if(percent > 87) {
    Write(BRIGHT_100);
  } else if(percent > 75) {
    Write(BRIGHT_087);
  } else if(percent > 62) {
    Write(BRIGHT_075);
  } else if(percent > 50 ) {
    Write(BRIGHT_062);
  } else if(percent > 37 ) {
    Write(BRIGHT_050);
  } else if(percent > 25 ) {
    Write(BRIGHT_037);
  } else if(percent > 12 ) {
    Write(BRIGHT_025);
  } else if(percent > 1 ) {
    Write(BRIGHT_012);
  } else {
    Write(BRIGHT_000);
  }
  if(wiring!=kSerial) {
    port->Release();
  }
}


int cGraphLCDDriverGU256X64_3900::DeInit()
{
  if(wiring==kSerial) {
    if(portFd>=0) {
      close(portFd);
      portFd=-1;
    }
  }
  if(port && (port->Close()!=0)) {
    return -1;
  }
  return 0;
}


void cGraphLCDDriverGU256X64_3900::CheckSetup()
{
  bool bBrightnessOnly = (setup->Brightness != oldSetup->Brightness);

  if(setup->DiffersImportantlyWith(oldSetup)) {
    DeInit();
    Init();
  } else if(bBrightnessOnly) {
    // Change throw Setup, if isn't update with DiffersImportantlyWith
    SetBrightness(setup->Brightness);
    oldSetup->Brightness = setup->Brightness;
  }
}

void cGraphLCDDriverGU256X64_3900::WriteParallel(byte data) {
  if(m_bSleepIsInit) {
    nSleepInit();
  }
  if((port->ReadStatus()&readyMask)!=readyHi) {
    int i=0;
    int status=port->ReadStatus();
    for(; ((status&readyMask)!=readyHi) && i<1000; i++) {
      // wait until display ack's write but not forever
      status=port->ReadStatus();
    }
  }

  port->WriteControl(WRLO);
  port->WriteData(data);
  nSleep(100 + (100 * setup->AdjustTiming) - m_nTimingAdjustCmd);
  port->WriteControl(WRHI);
  nSleep(500 + (100 * setup->AdjustTiming) - m_nTimingAdjustCmd);
}

void cGraphLCDDriverGU256X64_3900::WriteSerial(byte data) {
  write(portFd, &data, 1);
}

void cGraphLCDDriverGU256X64_3900::Write(byte data) {
  if(wiring==kSerial) {
    WriteSerial(data);
  } else {
    WriteParallel(data);
  }
}

void cGraphLCDDriverGU256X64_3900::SetPixel(int x, int y, eColor color)
{
  byte c;

  if(!m_pDrawMem) {
    return;
  }

  if(x >= sizeX || x < 0) {
    return;
  }
  if(y >= sizeY || y < 0) {
    return;
  }

  if(setup->UpsideDown) {
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
  }

  c = 0x80 >> (y % 8);

  if(color == colorBlack) {
    m_pDrawMem[x][y/8] = m_pDrawMem[x][y/8] | c;
  } else {
    m_pDrawMem[x][y/8] = m_pDrawMem[x][y/8] & ~c;
  }
}


void cGraphLCDDriverGU256X64_3900::Set8Pixels(int x, int y, byte data, eColor color)
{
  int n;

  // x - pos is'nt mayby align to 8
  x &= 0xFFF8;

  for(n = 0; n < 8; ++n) {
    if(data & (0x80 >> n)) {
      // if bit is set
      SetPixel(x + n, y, color);
    }
  }
}


void cGraphLCDDriverGU256X64_3900::Refresh(bool refreshAll)
{
  int xb, yb;
  
  CheckSetup();

  if(!m_pVFDMem || !m_pDrawMem) {
    return;
  }

  bool doRefresh = false;
  int minX = sizeX;
  int maxX = 0;
  int minYb = m_iSizeYb;
  int maxYb = 0;

  for(xb = 0; xb < sizeX; ++xb) {
    for(yb = 0; yb < (sizeY+7)/8; ++yb) {
      if (m_pVFDMem[xb][yb] != m_pDrawMem[xb][yb]) {
        m_pVFDMem[xb][yb] = m_pDrawMem[xb][yb];
        minX = min(minX, xb);
        maxX = max(maxX, xb);
        minYb = min(minYb, yb);
        maxYb = max(maxYb, yb + 1);
        doRefresh = true;
      }
    }
  }
  
  m_nRefreshCounter = (m_nRefreshCounter + 1) % setup->RefreshDisplay;
  if(!refreshAll && !m_nRefreshCounter) {
    refreshAll = true;
  }

  if(refreshAll || doRefresh) {
    if(refreshAll) {
      minX = 0;
      maxX = sizeX;
      minYb = 0;
      maxYb = m_iSizeYb;
      // and reset RefreshCounter
      m_nRefreshCounter = 0;
    }

    minX = max(minX, 0);
    maxX = min(maxX, sizeX - 1);
    minYb = max(minYb, 0);
    maxYb = min(maxYb, m_iSizeYb);

    if(wiring!=kSerial) {
      port->Claim();
    }

    if(wiring==kDMA) {
      Write(dBLITIMAGE0);
      Write(dBLITIMAGE1);
      Write(dBLITIMAGE2);
      Write(dBLITIMAGE3);
      Write(0); // low byte address
      Write(0); // high byte address
      Write(0); // low byte size
      Write(8); // high byte size
    } else {
      Write(nSETCURSOR0);
      Write(nSETCURSOR1);
      Write(0); // low byte x
      Write(0); // high byte x
      Write(0); // low byte y
      Write(0); // high byte y

      Write(nBLITIMAGE0);
      Write(nBLITIMAGE1);
      Write(nBLITIMAGE2);
      Write(nBLITIMAGE3);
      Write(0); // low byte width
      Write(1); // high byte width
      Write(8); // low byte height
      Write(0); // high byte height
      Write(1); // end header
    }

    for(xb = 0; xb < sizeX; xb++) {
      for(yb = 0; yb < m_iSizeYb; yb++) {
        Write( (m_pVFDMem[xb][yb]) ^ ((setup->Invert != 0) ? 0xff : 0x00) );
      }
      // parallel port writing is busy waiting - with realtime priority you
      // can lock the system - so don't be so greedy ;) 
      if((xb%32)==31) {
        uSleep(1000);
      }
    }

    if(wiring!=kSerial) {
      port->Release();
    }
  }
}
