/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 * gu256x64-3900.h - 8-bit driver module for Noritake GU256X64x-3900
 *               VFD displays. The VFD is operated in it's 8 bit-mode to be
 *               connected to a single PC parallel port.
 *
 * (c) 20040810 Ralf Mueller (ralf AT bj-ig de)
 *
 *  based on: 
 *            gu256x64-372 driver module for graphlcd
 *              (c) 2004 Andreas 'randy' Weinberger <randy AT smue org>
 *            gu256x32f driver module for graphlcd
 *              (c) 2003 Andreas Brachold <vdr04 AT deltab de>
 *            HD61830 device
 *              (c) 2001-2003 by Carsten Siebholz <c.siebholz AT t-online de>
 *            lcdproc 0.4 driver hd44780-ext8bit
 *              (c) 1999, 1995 Benjamin Tse <blt AT Comports com>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

/**
 * The VFD should be connected to a parallel or serial port of the PC.
 * For DMA Mode it has to be connected to the parallel port and switch
 * SW6 on the display has to be ON.
 *
 * The connections are:
 * (Note: This pin number might differ with other displays)
 *
 *   printerport                                    LCD          other
 *   default     satyr      inoff satyr         
 *   wiring      wiring     wiring                  
 *  -----------  ---------- -----------   -->       ----------   ----------
 *               STRB (1)                           nWR (10)                 (Write)
 *   D0    (2)   D0   (2)   D0    (2)               DB0 (08)                 (DataBit 0)
 *   D1    (3)   D1   (3)   D1    (3)               DB1 (07)                 (DataBit 1)
 *   D2    (4)   D2   (4)   D2    (4)               DB2 (06)                 (DataBit 2)
 *   D3    (5)   D3   (5)   D3    (5)               DB3 (05)                 (DataBit 3)
 *   D4    (6)   D4   (6)   D4    (6)               DB4 (04)                 (DataBit 4)
 *   D5    (7)   D5   (7)   D5    (7)               DB5 (03)                 (DataBit 5)
 *   D6    (8)   D6   (8)   D6    (8)               DB6 (02)                 (DataBit 6)
 *   D7    (9)   D7   (9)   D7    (9)               DB7 (01)                 (DataBit 7)
 *   ACK   (10)                                     RDY (12)                 (Display Ready)
 *               Busy (11)  Busy (11)               RDY (12)                 (Display Ready)
 *   INIT  (16)                                     nWR (10)                 (Write)
 *                          SS   (17)               nWR (10)                 (Write)
 *   GND   (18)                                     GND (09)     GND         (Ground)
 *   GND   (19)                                     GND (11)     GND         (Ground)
 *   GND   (20)                                     GND (13)     GND         (Ground)
 *   GND   (21)                                     GND (14)     GND         (Ground)
 *                                                  VCC (15)     VCC         (Power supply 5V)
 *                                                  VCC (16)     VCC         (Power supply 5V)
 *
 *   serialport    LCD          other
 *  -----------    ----------   ----------
 *   DCD   (1)      DTR (2)                  
 *   RxD   (2)      TxD (4)                 
 *   TxD   (3)      RxD (1)                
 *   DTR   (4)      DSR (3)               
 *   GND   (5)      GND (7)     GND         (Ground)
 *   DSR   (6)      DTR (2) 
 *                  VCC (6)     VCC         (Power supply 5V)
 *
 * On Display Switch:
 * SW No.    Function                 Default
 *  1        Display address select   Off
 *  2        Display address select   Off
 *  3        Display address select   Off
 *  4        Display address select   Off
 *  5        Baud rate select         Off
 *  6        Command mode select      Off      switch DMA On/Off
 *  7        Operating mode select    Off
 *  8        Protocol Select          Off
 **/

#ifndef _LIBGRAPHLCD_GU256X64_3900_H_
#define _LIBGRAPHLCD_GU256X64_3900_H_

#include "device.h"
#include "port.h"
#include "../setup.h"

class cGraphLCDDriverGU256X64_3900 : public cGraphLCDDevice
{
  int wiring;
  byte readyMask;
  byte readyHi;

  cParallelPort* port;
  int portFd;

  cGraphLCDSetup* setup;
  cGraphLCDSetup* oldSetup;

  int m_iSizeYb;
  int m_nRefreshCounter;
  
  byte** m_pDrawMem;             /* the draw "memory" */
  byte** m_pVFDMem;              /* the double buffed display "memory" */

  long m_nTimingAdjustCmd;
  bool m_bSleepIsInit;

protected:
  virtual void ClearVFDMem();
  virtual void CheckSetup();

  virtual int InitSerialPort();
  virtual int InitParallelPort();
  virtual void InitNormalDisplay();
  virtual void InitDMADisplay();

  virtual void WriteSerial(byte data);
  virtual void WriteParallel(byte data);
  virtual void Write(byte data);
public:
  cGraphLCDDriverGU256X64_3900(cGraphLCDSetup * setup);
  virtual ~cGraphLCDDriverGU256X64_3900();

  virtual int Init();
  virtual int DeInit();

  virtual void Clear();
  virtual void SetPixel(int x, int y, eColor color);
  virtual void Set8Pixels(int x, int y, byte data, eColor color);
  virtual void Refresh(bool refreshAll = false);

  virtual void SetBrightness(unsigned int percent);
};

#endif

