/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  hd61830.c  -  HD61830 device
 *
 *  (c) 2001-2004 Carsten Siebholz <c.siebholz AT t-online de>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <sys/time.h>
#include <vdr/tools.h>

#include "hd61830.h"
#include "fct.h"


// commands
#define MCNT    0x00
#define CPIT    0x01
#define NOCH    0x02
#define NOTD    0x03
#define CPOS    0x04

#define DSAL    0x08
#define DSAH    0x09
#define CACL    0x0A
#define CACH    0x0B

#define WDDI    0x0C
#define RDDI    0x0D

#define CBIT    0x0E
#define SBIT    0x0F

// control bits for DirectIO
#define EN      0x01
#define ENHI    0x00
#define ENLO    0x01

#define RW      0x02
#define RWHI    0x00
#define RWLO    0x02

#define RS      0x04
#define RSHI    0x04
#define RSLO    0x00



cGraphLCDDriverHD61830::cGraphLCDDriverHD61830(cGraphLCDSetup * setup)
: setup(setup)
{
  int x;

  oldSetup = new cGraphLCDSetup();
  oldSetup->CopyFrom(setup);

  port = new cParallelPort();

  sizeX = setup->SizeX;
  if (sizeX < 0)
    sizeX = 0;
  sizeY = setup->SizeY;
  if (sizeY < 0)
    sizeY = 0;

  useSleepInit = false;

  // setup lcd array (wanted state)
  newLCD = new byte*[(sizeX + 7) / 8];
  if (newLCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      newLCD[x] = new byte[sizeY];
      memset(newLCD[x], 0, sizeY);
    }
  }
  // setup lcd array (current state)
  oldLCD = new byte*[(sizeX + 7) / 8];
  if (oldLCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      oldLCD[x] = new byte[sizeY];
      memset(oldLCD[x], 0, sizeY);
    }
  }
  refreshCounter = 0;
  timeForPortCmdInNs = 0;
}

cGraphLCDDriverHD61830::~cGraphLCDDriverHD61830()
{
  int x;

  delete port;

  // free lcd array (wanted state)
  if (newLCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      delete[] newLCD[x];
    }
    delete[] newLCD;
  }
  // free lcd array (current state)
  if(oldLCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      delete[] oldLCD[x];
    }
    delete[] oldLCD;
  }
  delete oldSetup;
}


int cGraphLCDDriverHD61830::InitGraphic()
{
  Write(MCNT, 0x32);   /* set Mode Control Register */
                      /* DISP ON, MASTER ON, BLINK OFF, CURSOR OFF, GRAPHIC-Mode, int.Clock */
  Write(CPIT, 0x07);   /* set Character Pitch Register */
                      /* 8 pixels per byte */
  Write(NOCH, max(1, (sizeX + 7) / 8 - 1)); /* set Number-Of-Characters Register */
                      /* (iSizeX-1)/8 bytes per line horizontally */
  Write(NOTD, max(1, sizeY - 1));       /* set Number-Of-Time-Divisions Register */
                      /* iSizeY-1 */
  Write(CPOS, 0x00);   /* set Cursor Position Register */
                      /* optional, because we havn't enabled a cursor */
  Write(DSAL, 0x00);   /* set Display Start Address Register (Low Order Byte) */
  Write(DSAH, 0x00);   /* set Display Start Address Register (High Order Byte) */
  Write(CACL, 0x00);   /* set Cursor Address Counter Register (Low Order Byte) */
  Write(CACH, 0x00);   /* set Cursor Address Counter Register (High Order Byte) */

  return 0;
}


int cGraphLCDDriverHD61830::Init()
{
  int i;
  struct timeval tv1, tv2;


  if (!setup->PPDevice)
  {
    // use DirectIO
    if (port->Open(setup->PPPort) != 0)
      return -1;
    uSleep(10);
  }
  else
  {
    // use ppdev
    if (port->Open(setup->PPDevice) == -1)
      return -1;
  }

  if (nSleepInit() != 0)
  {
    dsyslog("graphlcd plugin: INFO: cannot change wait parameters (cGraphLCDDriver::Init)\n");
    useSleepInit = false;
  }
  else
  {
    useSleepInit = true;
  }

  dsyslog("graphlcd plugin: benchmark started.\n");
  gettimeofday(&tv1, 0);
  for (i = 0; i < 1000; i++)
  {
    port->WriteData(1 % 0x100);
  }
  gettimeofday(&tv2, 0);
  if (useSleepInit)
    nSleepDeInit();
  timeForPortCmdInNs = (tv2.tv_sec-tv1.tv_sec) * 1000000 + (tv2.tv_usec-tv1.tv_usec);
  dsyslog("graphlcd plugin: benchmark stopped. Time for Port Command: %ldns\n", timeForPortCmdInNs);

  // initialize graphic mode
  InitGraphic();

  port->Release();
  
  oldSetup->CopyFrom(setup);

  // clear display
  Clear();
  Refresh(true);

  isyslog("graphlcd plugin: HD61830 initialized.\n");
  return 0;
}


int cGraphLCDDriverHD61830::DeInit()
{
  if (port->Close() != 0)
    return -1;
  return 0;
}


void cGraphLCDDriverHD61830::Write(byte cmd, byte data)
{
  if (useSleepInit)
    nSleepInit();

  // set RS high (instruction), RW low (write) and E low
  port->WriteControl(RSHI | RWLO | ENLO);
  nSleep(140 - timeForPortCmdInNs + 100 * setup->AdjustTiming);

  // Output the actual command
  port->WriteData(cmd);

  // set E high
  port->WriteControl(RSHI | RWLO | ENHI);
  nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);

  // set E low
  port->WriteControl(RSHI | RWLO | ENLO);
  nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);


  // set RS low (data), RW low (write) and E low
  port->WriteControl(RSLO | RWLO | ENLO);
  nSleep(140 - timeForPortCmdInNs + 100 * setup->AdjustTiming);

  // Output the actual data
  port->WriteData(data);

  // set E high
  port->WriteControl(RSLO | RWLO | ENHI);
  nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);

  // set E low
  port->WriteControl(RSLO | RWLO | ENLO);
  nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);

  switch (cmd)
  {
    case MCNT:
    case CPIT:
    case NOCH:
    case NOTD:
    case CPOS:
    case DSAL:
    case DSAH:
    case CACL:
    case CACH:
      nSleep(4000 - max(450l, timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      break;
    case WDDI:
    case RDDI:
      nSleep(6000 - max(450l, timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      break;
    case CBIT:
    case SBIT:
      nSleep(36000 - max(450l, timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      break;
  }
  if (useSleepInit)
    nSleepDeInit();
}


void cGraphLCDDriverHD61830::CheckSetup()
{
  if (setup->DiffersImportantlyWith(oldSetup))
  {
    DeInit();
    Init();
  }
}


void cGraphLCDDriverHD61830::Clear()
{
  for (int x = 0; x < (sizeX + 7) / 8; x++)
    memset(newLCD[x], 0, sizeY);
}


void cGraphLCDDriverHD61830::SetPixel(int x, int y, eColor color)
{
  byte c;

  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  c = 0x01;
  if (setup->UpsideDown)
  {
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
  }

  c = c << (x % 8);
  if (color == colorBlack)
  {
    newLCD[x / 8][y] = newLCD[x / 8][y] | c;
  }
  else
  {
    newLCD[x / 8][y] = newLCD[x / 8][y] & ~c;
  }
}


void cGraphLCDDriverHD61830::Set8Pixels(int x, int y, byte data, eColor color)
{
  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  if (!setup->UpsideDown)
  {
    // normal orientation
    if (color == colorBlack)
    {
      newLCD[x / 8][y] = newLCD[x / 8][y] | ReverseBits(data);
    }
    else
    {
      newLCD[x / 8][y] = newLCD[x / 8][y] & ReverseBits(~data);
    }
  }
  else
  {
    // upside down orientation
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
    if (color == colorBlack)
    {
      newLCD[x / 8][y] = newLCD[x / 8][y] | data;
    }
    else
    {
      newLCD[x / 8][y] = newLCD[x / 8][y] & ~data;
    }
  }
}


void cGraphLCDDriverHD61830::Refresh(bool refreshAll)
{
  int x;
  int y;
  int pos = 0;

  CheckSetup();

  if (setup->RefreshDisplay > 0)
  {
    refreshCounter = (refreshCounter + 1) % setup->RefreshDisplay;
    if (!refreshAll && !refreshCounter)
      refreshAll = true;
  }

  port->Claim();

  if (refreshAll)
  {
    // draw all

    for (y = 0; y < sizeY; y++)
    {
      for (x = 0; x < (sizeX + 7) / 8; x++)
      {
        // (re-setting the cursor position
        //  might be removed, when the graphic glitches are solved)
        Write(CACL, (pos % 0x100));
        Write(CACH, (pos / 0x100));
        Write(WDDI, (newLCD[x][y]) ^ ((setup->Invert == 0) ? 0xff : 0x00));
        oldLCD[x][y] = newLCD[x][y];
        pos++;
      }
    }
    // and reset RefreshCounter
    refreshCounter = 0;

  }
  else
  {
    // draw only the changed bytes

    bool cs = false;
    for (y = 0; y < sizeY; y++)
    {
      for (x = 0; x < (sizeX + 7) / 8; x++)
      {
        if (newLCD[x][y] != oldLCD[x][y])
        {
          if (!cs)
          {
            Write(CACL, (pos % 0x100));
            Write(CACH, (pos / 0x100));
            cs = true;
          }
          Write(WDDI, (newLCD[x][y]) ^ ((setup->Invert == 0) ? 0xff : 0x00));
          oldLCD[x][y] = newLCD[x][y];
        }
        else
        {
          cs = false;
        }
        pos++;
      }
    }
  }
  port->Release();
}

