/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  image.h  -  Image output device
 *              Output goes to a image file instead of lcd.
 *
 *  (c) 2004 Andreas Regel <andreas.regel AT gmx de>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <vdr/tools.h>

#include "image.h"
#include "fct.h"


cGraphLCDDriverImage::cGraphLCDDriverImage(cGraphLCDSetup * setup)
: setup(setup)
{
  oldSetup = new cGraphLCDSetup();
  oldSetup->CopyFrom(setup);

  sizeX = setup->SizeX;
  if (sizeX < 0)
    sizeX = 0;
  sizeY = setup->SizeY;
  if (sizeY < 0)
    sizeY = 0;
  lineSize = (sizeX + 7) / 8;
}

cGraphLCDDriverImage::~cGraphLCDDriverImage()
{
  delete oldSetup;
}


int cGraphLCDDriverImage::Init()
{
  oldSetup->CopyFrom(setup);

  newLCD = new byte[lineSize * sizeY];
  if (newLCD)
    memset(newLCD, 0, lineSize * sizeY);
  oldLCD = new byte[lineSize * sizeY];
  if (oldLCD)
    memset(oldLCD, 0, lineSize * sizeY);

  counter = 0;

  // clear display
  Clear();
  Refresh(true);

  isyslog("graphlcd plugin: image driver initialized.\n");
  return 0;
}


int cGraphLCDDriverImage::DeInit()
{
  if (newLCD)
    delete[] newLCD;
  if (oldLCD)
    delete[] oldLCD;
  return 0;
}


void cGraphLCDDriverImage::CheckSetup()
{
  if (setup->DiffersImportantlyWith(oldSetup))
  {
    DeInit();
    Init();
  }
}


void cGraphLCDDriverImage::Clear()
{
  memset(newLCD, 0, lineSize * sizeY);
}


void cGraphLCDDriverImage::SetPixel(int x, int y, eColor color)
{
  byte c;

  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  c = 0x80;
  if (setup->UpsideDown)
  {
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
  }

  c = c >> (x % 8);
  if (color == colorBlack)
    newLCD[lineSize * y + x / 8] |= c;
  else
    newLCD[lineSize * y + x / 8] &= ~c;
}


void cGraphLCDDriverImage::Set8Pixels(int x, int y, byte data, eColor color)
{
  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  if (!setup->UpsideDown)
  {
    // normal orientation
    if (color == colorBlack)
      newLCD[lineSize * y + x / 8] |= data;
    else
      newLCD[lineSize * y + x / 8] &= ~data;
  }
  else
  {
    // upside down orientation
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
    if (color == colorBlack)
      newLCD[lineSize * y + x / 8] |= ReverseBits(data);
    else
      newLCD[lineSize * y + x / 8] &= ReverseBits(~data);
  }
}


void cGraphLCDDriverImage::Refresh(bool refreshAll)
{
  int i;
  bool refresh;
  char fileName[256];
  char str[32];
  FILE * fp;
  unsigned char c;

  CheckSetup();

  refresh = false;
  for (i = 0; i < lineSize * sizeY; i++)
  {
    if (newLCD[i] != oldLCD[i])
    {
      refresh = true;
      break;
    }
  }

  if (refresh)
  {
    sprintf(fileName, "%s/%s%05d.%s", "/tmp", "lcd", counter, "pbm");
    fp = fopen(fileName, "wb");
    if (fp)
    {
      sprintf(str, "P4\n%d %d\n", sizeX, sizeY);
      fwrite(str, strlen(str), 1, fp);
      for (i = 0; i < lineSize * sizeY; i++)
      {
        c = newLCD[i] ^ (setup->Invert ? 0xff : 0x00);
        fwrite(&c, 1, 1, fp);
        oldLCD[i] = newLCD[i];
      }
      fclose(fp);
    }
    counter++;
    if (counter > 99999)
      counter = 0;
  }
}
