/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  ks0108.c   -  KS0108 device
 *
 *  (c) 2003 andreas 'randy' weinberger <vdr AT smue org>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/


#include <sys/time.h>
#include <vdr/tools.h>

#include "ks0108.h"
#include "fct.h"


// commands
const byte kSEAD = 0x40;	// Set (X) Address
const byte kSEPA = 0xb8;	// Set (Y) Page
const byte kSEDS = 0xc0;	// Set Display Start Line
const byte kDIOF = 0x3e;	// Display off
const byte kDION = 0x3f;	// Display on

const byte kCEHI = 0x01;	// Chip Enable on
const byte kCELO = 0x00;	
const byte kCDHI = 0x08;	// Command/Data Register Select
const byte kCDLO = 0x00;

const byte kCS1HI = 0x02;	// ChipSelect 1
const byte kCS1LO = 0x00;
const byte kCS2HI = 0x00;	// ChipSelect 2
const byte kCS2LO = 0x04;


extern byte bitmask[];

cGraphLCDDriverKS0108::cGraphLCDDriverKS0108(cGraphLCDSetup * setup)
: setup(setup)
{
  int x;

  oldSetup = new cGraphLCDSetup();
  oldSetup->CopyFrom(setup);

  port = new cParallelPort();

  if (setup->SizeX <= 128) {
    sizeX = 128;
  } else if (setup->SizeX > 192) {
    sizeX = 256;
  } else if (setup->SizeX > 128) {
    sizeX = 192;
  }

  if (setup->SizeY <= 64) {
    sizeY = 64;
  } else if (setup->SizeY > 64) {
    sizeY = 128;
    sizeX = 128; 	// force 2* 128x64 display
  }

  if (sizeX == 128 && sizeY == 64) {
    CS1 = kCS2HI | kCS1LO;
    CS2 = kCS2LO | kCS1HI;
    CS3 = -1;			// invalid
    CS4 = -1;
  } else {			// multiplexed via 74LS42
    CS1 = kCS2HI | kCS1HI;
    CS2 = kCS2HI | kCS1LO;
    CS3 = kCS2LO | kCS1HI;
    CS4 = kCS2LO | kCS1LO;
  }   

  SEAD = kSEAD;
  SEPA = kSEPA;
  SEDS = kSEDS; 
  DIOF = kDIOF;
  DION = kDION;


  CEHI = kCEHI;
  CELO = kCELO;
  CDHI = kCDHI;
  CDLO = kCDLO;
  
  useSleepInit = false;
  // setup linear lcd array
  LCD = new byte*[(sizeX + 7) / 8];
  if (LCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      LCD[x] = new byte[sizeY];
      memset(LCD[x], 0, sizeY);
    }
  }
  // setup the lcd array for the paged ks0108
  LCD_page = new byte*[sizeX];
  if (LCD_page)
  {
    for (x = 0; x < sizeX; x++)
    {
      LCD_page[x] = new byte[(sizeY + 7) / 8];
      memset(LCD_page[x], 0, (sizeY + 7) / 8);
    }
  }
  refreshCounter = 0;
  timeForPortCmdInNs = 0;
  timeForLCDInNs = 50;
}


cGraphLCDDriverKS0108::~cGraphLCDDriverKS0108()
{
  int x;

  // free linear lcd array
  if (LCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      delete[] LCD[x];
    }
    delete[] LCD;
  }
  // free paged lcd array
  if (LCD_page)
  {
    for (x = 0; x < sizeX; x++)
    {
      delete[] LCD_page[x];
    }
    delete[] LCD_page;
  }
  delete oldSetup;
  delete port;
}


int cGraphLCDDriverKS0108::InitGraphic()
{
  // init controllers
  if (CS1 > -1) {
    KS0108Cmd(SEDS, 1);
    KS0108Cmd(SEPA, 1);
    KS0108Cmd(SEAD, 1);
    KS0108Cmd(DION, 1);
  }
  if (CS2 > -1) {
    KS0108Cmd(SEDS, 2);
    KS0108Cmd(SEPA, 2);
    KS0108Cmd(SEAD, 2);
    KS0108Cmd(DION, 2);
  }
  if (CS3 > -1) {
    KS0108Cmd(SEDS, 3);
    KS0108Cmd(SEPA, 3);
    KS0108Cmd(SEAD, 3);
    KS0108Cmd(DION, 3);
  }
  if (CS4 > -1) {
    KS0108Cmd(SEDS, 4);
    KS0108Cmd(SEPA, 4);
    KS0108Cmd(SEAD, 4);
    KS0108Cmd(DION, 4);
  }
  return 0;
}


int cGraphLCDDriverKS0108::Init()
{
  int i;
  struct timeval tv1, tv2;

  if (!setup->PPDevice)
  {
    // use DirectIO
    if (port->Open(setup->PPPort) != 0)
      return -1;
    uSleep(10);
  }
  else
  {
    // use ppdev
    if (port->Open(setup->PPDevice) != 0)
      return -1;
  }

  if (nSleepInit() != 0)
  {
    dsyslog("graphlcd plugin: INFO: cannot change wait parameters (cGraphLCDDriver::Init)\n");
    useSleepInit = false;
  }
  else
  {
    useSleepInit = true;
  }

  dsyslog("graphlcd plugin: benchmark started.\n");
  gettimeofday(&tv1, 0);
  for (i = 0; i < 1000; i++)
  {
    port->WriteData(i % 0x100);
  }
  gettimeofday(&tv2, 0);
  if (useSleepInit)
    nSleepDeInit();
  timeForPortCmdInNs = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);

  dsyslog("graphlcd plugin: benchmark stopped. Time for Command: %ldns\n", timeForPortCmdInNs);

  // initialize graphic mode
  InitGraphic();

  port->Release();

  oldSetup->CopyFrom(setup);

  // clear display
  Clear();
  Refresh(true);

  dsyslog("graphlcd plugin: KS0108 initialized.\n");
  return 0;
}


int cGraphLCDDriverKS0108::DeInit()
{
  if (port->Close() != 0)
    return -1;
  return 0;
}

void cGraphLCDDriverKS0108::KS0108Cmd(byte data, int cs)
{
  if (useSleepInit)
    nSleepInit();
  switch (cs) {
    case 1:
      // set Command Register on, CS2 on, CS1 off, Chip Enable on;
      port->WriteControl(CDHI | CS1 | CEHI);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteData(data);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteControl(CDHI | CS1 | CELO);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      break;
    case 2:
      port->WriteControl(CDHI | CS2 | CEHI);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteData(data);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteControl(CDHI | CS2 | CELO);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      break;
    case 3:
      port->WriteControl(CDHI | CS3 | CEHI);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteData(data);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteControl(CDHI | CS3 | CELO);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      break;
    case 4:
      port->WriteControl(CDHI | CS4 | CEHI);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteData(data);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteControl(CDHI | CS4 | CELO);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      break;
  }
  if (useSleepInit)
    nSleepDeInit();
}

void cGraphLCDDriverKS0108::KS0108Data(byte data, int cs)
{
  if (useSleepInit)
    nSleepInit();
  switch (cs) {
    case 1:
      port->WriteControl(CDLO | CS1 | CEHI);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteData(data);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteControl(CDLO | CS1 | CELO);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      break;
    case 2:
      port->WriteControl(CDLO | CS2 | CEHI);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteData(data);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteControl(CDLO | CS2 | CELO);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      break;
    case 3:
      port->WriteControl(CDLO | CS3 | CEHI);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteData(data);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteControl(CDLO | CS3 | CELO);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      break;
    case 4:
      port->WriteControl(CDLO | CS4 | CEHI);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteData(data);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      port->WriteControl(CDLO | CS4 | CELO);
      nSleep((timeForLCDInNs + timeForPortCmdInNs) + 100 * setup->AdjustTiming);
      break;
  }
  if (useSleepInit)
    nSleepDeInit();
}



void cGraphLCDDriverKS0108::CheckSetup()
{
  if (setup->DiffersImportantlyWith(oldSetup))
  {
    DeInit();
    Init();
  }
}


void cGraphLCDDriverKS0108::Clear()
{
  for (int x = 0; x < (sizeX + 7) / 8; x++)
    memset(LCD[x], 0, sizeY);
}


void cGraphLCDDriverKS0108::SetPixel(int x, int y, eColor color)
{
  byte c;

  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  if (setup->UpsideDown)
  {
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
  }

  c = 0x80;
  c = c >> (x % 8);
  if (color == colorBlack)
  {
    LCD[x / 8][y] = LCD[x / 8][y] | c;
  }
  else
  {
    LCD[x / 8][y] = LCD[x / 8][y] & ~c;
  }
}


void cGraphLCDDriverKS0108::Set8Pixels(int x, int y, byte data, eColor color)
{
  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  if (!setup->UpsideDown)
  {
    // normal orientation
    if (color == colorBlack)
    {
      LCD[x / 8][y] = LCD[x / 8][y] | data;
    }
    else
    {
      LCD[x / 8][y] = LCD[x / 8][y] & ~data;
    }
  }
  else
  {
    // upside down orientation
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
    if (color == colorBlack)
    {
      LCD[x / 8][y] = LCD[x / 8][y] | ReverseBits(data);
    }
    else
    {
      LCD[x / 8][y] = LCD[x / 8][y] & ReverseBits(~data);
    }
  }
}


void cGraphLCDDriverKS0108::Refresh(bool refreshAll)
{
  int   x,y;
  int   xx,yy;
  byte  dByte, oneBlock[8];

  CheckSetup();

  if (setup->RefreshDisplay > 0)
  {
    refreshCounter = (refreshCounter + 1) % setup->RefreshDisplay;
    if (!refreshAll && !refreshCounter)
      refreshAll=true;
  }

  refreshAll = true; // differential update is not yet supported

  if (refreshAll)
  {
    // draw all

    // convert the linear lcd array to the paged array for the display
    for (y = 0; y < (sizeY + 7) / 8; y++)
    {
      for (x = 0; x < (sizeX + 7) / 8; x++)
      {
        for (yy = 0; yy < 8; yy++)
        {
          oneBlock[yy] = LCD[x][yy + (y * 8)] ^ ((setup->Invert != 0) ? 0xff : 0x00);
        }
        for (xx = 0; xx < 8; xx++)
        {
          dByte = 0;
          for (yy = 0; yy < 8; yy++)
          {
            if(oneBlock[yy] & bitmask[xx])
            {
              dByte += (1 << yy);
            }
          }
          LCD_page[x * 8 + xx][y] = dByte;
        }
      }
    }

    port->Claim();

    if (sizeX == 128 && sizeY == 64) {
      for (y = 0; y < 64/8; y++) {
        KS0108Cmd(SEPA + y, 1);
        KS0108Cmd(SEAD, 1);

        for (x = 0; x < 64; x++) {
          KS0108Data(LCD_page[x][y], 1);
        }
      }

      for (y = 0; y < 64/8; y++) {
        KS0108Cmd(SEPA + y, 2);
        KS0108Cmd(SEAD, 2);

        for (x = 64; x < 128; x++) {
          KS0108Data(LCD_page[x][y], 2);
        }
      }
    }

    if (sizeX > 128 && sizeY == 64) {
      for (y = 0; y < 64/8; y++) {
        KS0108Cmd(SEPA + y, 1);
        KS0108Cmd(SEAD, 1);

        for (x = 0; x < 64; x++) {
          KS0108Data(LCD_page[x][y], 1);
        }
      }

      for (y = 0; y < 64/8; y++) {
        KS0108Cmd(SEPA + y, 2);
        KS0108Cmd(SEAD, 2);
        for (x = 64; x < 128; x++) {
          KS0108Data(LCD_page[x][y], 2);
        }
      }

      for (y = 0; y < 64/8; y++) {
        KS0108Cmd(SEPA + y, 3);
        KS0108Cmd(SEAD, 3);

        for (x = 128; x < 192; x++) {
          KS0108Data(LCD_page[x][y], 3);
        }
      }

      for (y = 0; y < 64/8; y++) {
        if (sizeX > 192) {
          KS0108Cmd(SEPA + y, 4);
          KS0108Cmd(SEAD, 4);
          for (x = 192; x < 256; x++) {
            KS0108Data(LCD_page[x][y], 4);
          }
        }
      }
    }

    if (sizeX == 128 && sizeY == 128) {
      for (y = 0; y < 64/8; y++) {
        KS0108Cmd(SEPA + y, 1);
        KS0108Cmd(SEAD, 1);

        for (x = 0; x < 64; x++) {
          KS0108Data(LCD_page[x][y], 1);
        }
      }

      for (y = 0; y < 64/8; y++) {
        KS0108Cmd(SEPA + y, 2);
        KS0108Cmd(SEAD, 2);

        for (x = 64; x < 128; x++) {
          KS0108Data(LCD_page[x][y], 2);
        }
      }

      for (y = 0; y < 64/8; y++) {
        KS0108Cmd(SEPA + y, 3);
        KS0108Cmd(SEAD, 3);

        for (x = 0; x < 64; x++) {
          KS0108Data(LCD_page[x][y+8], 3);
        }
      }

      for (y = 0; y < 64/8; y++) {
        KS0108Cmd(SEPA + y, 4);
        KS0108Cmd(SEAD, 4);

        for (x = 64; x < 128; x++) {
          KS0108Data(LCD_page[x][y+8], 4);
        }
      }
    }
    port->Release();
  }
  else
  {
    // draw only the changed bytes
  }
}
