/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  ks0108.h   -  KS0108 device
 *
 *  (c) 2003 andreas 'randy' weinberger <vdr AT smue org>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

/**
 * The LCD should be connected to the parallel printer port of the PC.
 *
 *
 * The connections are (like LCD_INFO and LiquidMP3):
 * (Note: This pin number might differ with other displays)
 *                       vv
 *  printerport   LCD          other
 *  -----------   ----------   ----------
 *  GND   (18)    GND   (01)   GND          (Ground)
 *                Vdd   (02)   +5V
 *                Vo    (03)                (LCD Contrast In)
 *  nSEL  (17)    R/S   (04)                (Register Select)
 *                R/W   (05)   GND          (Read/Write)
 *  nSTRB (01)    EN    (06)                (Enable)
 *  D0    (02)    D0    (07)                (DataBit0)
 *  D1    (03)    D1    (08)                (DataBit1)
 *  D2    (04)    D2    (09)                (DataBit2)
 *  D3    (05)    D3    (10)                (DataBit3)
 *  D4    (06)    D4    (11)                (DataBit4)
 *  D5    (07)    D5    (12)                (DataBit5)
 *  D6    (08)    D6    (13)                (DataBit6)
 *  D7    (09)    D7    (14)                (DataBit7)
 *  nLF   (14)    CS1   (15)                (ChipSelect Controller 1)
 *  INIT  (16)    CE2   (16)                (ChipSelect Controller 2)
 *                RESET (17)   +5V          (Controller Reset)
 *                Vout  (18)                (Contrast Out)
 *                LED+  (19)   +4V          (Backlight)
 *                LED-  (20)   GND          (Backlight GND)
 *
 *  if you want to connect a 192x64 or 256x64 displays (with 3/4 ChipSelects)
 *  or two 128x64 displays, you must add a binarydecoder for the
 *  ChipSelect lines. the mode ist autoselecte from the x/y options.
 *
 *        nLF (14)
 *        |   INIT (16)
 *    Vdd |   |   GND GND
 *    |   |   |   |   |
 *    16  15  14  13  12  11  10  09
 *   --------------------------------
 *  |                                |
 *   >                               |
 *  |                                |
 *   --------------------------------
 *    01  02  03  04  05  06  07  08
 *    |   |   |   |               |
 *    CS1 |   |   |               GND
 *        CS2 |   CS4 (CS2 second LCD)
 *            CS3 (CS1 second LCD)
 *
 *
 **/

#ifndef _LIBGRAPHLCD_KS0108_H_
#define _LIBGRAPHLCD_KS0108_H_

#include "device.h"
#include "port.h"
#include "../setup.h"

class cGraphLCDDriverKS0108 : public cGraphLCDDevice
{
private:
  cParallelPort * port;
  byte ** LCD;      // linear lcd display "memory"
  byte ** LCD_page; // paged lcd display "memory"
  int refreshCounter;
  long timeForPortCmdInNs;
  long timeForLCDInNs;
  cGraphLCDSetup * setup;
  cGraphLCDSetup * oldSetup;
  bool useSleepInit;

  int InitGraphic();
  void KS0108Cmd(byte data, int cs);
  void KS0108Data(byte data, int cs);
  void CheckSetup();

  int SEAD;
  int SEPA;
  int SEDS;
  int DIOF;
  int DION;

  int CEHI;
  int CELO;
  int CDHI;
  int CDLO;
  int CS1HI;
  int CS1LO;
  int CS2HI;
  int CS2LO;

  int CS1;
  int CS2;
  int CS3;
  int CS4;

public:
  cGraphLCDDriverKS0108(cGraphLCDSetup * setup);
  virtual ~cGraphLCDDriverKS0108();

  virtual int Init();
  virtual int DeInit();

  virtual void Clear();
  virtual void SetPixel(int x, int y, eColor color);
  virtual void Set8Pixels(int x, int y, byte data, eColor color);
  virtual void Refresh(bool refreshAll = false);
};

#endif

