/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  ks0713.c   -  KS0713 device
 *
 *  (c) 2001-2004 Carsten Siebholz <c.siebholz AT t-online de>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <sys/time.h>
#include <vdr/tools.h>

#include "ks0713.h"
#include "fct.h"


// commands
#define REST    0xe2  // Reset

#define VOLT    0x28  // Power Control
#define VCON    0x04  // Voltage Converter On
#define VCOF    0x00  // Voltage Converter Off
#define VRON    0x02  // Voltage Regulator On
#define VROF    0x00  // Voltage Regulator Off
#define VFON    0x01  // Voltage Follower  On
#define VFOF    0x00  // Voltage Follower  Off

#define ADCS    0xa0  // ADC Select (horizontal)
#define ANOR    0x00  // Normal direction
#define AREV    0x01  // Reverse direction

#define SHLS    0xa0  // SHL Select (vertical)
#define SNOR    0x00  // Normal direction
#define SREV    0x01  // Reverse direction

#define BIAS    0xa2  // LCD Bias Select
#define BIA7    0x00  // 1/7 Bias
#define BIA9    0x01  // 1/9 Bias

#define REFV    0x81  // Set Reference Voltage

#define RRRS    0x20  // Regulator Resistor Ratio Select */

#define INVD    0xa6  // Normal/Invert Display
#define IVON    0x01  // Inverted Display
#define IVOF    0x00  // Normal Display

#define INDL    0x40  // Initial Display Line */

#define SPAD    0xb0  // Set Page Address */

#define SCAM    0x10  // Set Column Address MSB */
#define SCAL    0x00  // Set Column Address LSB */

#define DION    0xaf  // Display on
#define DIOF    0xae  // Display off



// control bits for DirectIO
#define CE      0x01
#define CEHI    0x00  // Chip Enable on
#define CELO    0x01  // Chip Enable off

#define RS      0x08
#define RSHI    0x00  // Command Register Select
#define RSLO    0x08  // Data Register Select


extern byte bitmask[];


cGraphLCDDriverKS0713::cGraphLCDDriverKS0713(cGraphLCDSetup * setup)
: setup(setup)
{
  int x;

  oldSetup = new cGraphLCDSetup();
  oldSetup->CopyFrom(setup);

  port = new cParallelPort();

  sizeX = setup->SizeX;
  if (sizeX < 0)
    sizeX = 0;
  sizeY = setup->SizeY;
  if (sizeY < 0)
    sizeY = 0;

  useSleepInit = false;
  // setup linear lcd array
  LCD = new byte*[(sizeX + 7) / 8];
  if (LCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      LCD[x] = new byte[sizeY];
      memset(LCD[x], 0, sizeY);
    }
  }
  // setup the lcd array for the paged ks0713
  LCD_page = new byte*[sizeX];
  if (LCD_page)
  {
    for (x = 0; x < sizeX; x++)
    {
      LCD_page[x] = new byte[(sizeY + 7) / 8];
      memset(LCD_page[x], 0, (sizeY + 7) / 8);
    }
  }
  refreshCounter = 0;
  timeForPortCmdInNs = 0;
}


cGraphLCDDriverKS0713::~cGraphLCDDriverKS0713()
{
  int x;

  // free linear lcd array
  if (LCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      delete[] LCD[x];
    }
    delete[] LCD;
  }
  // free paged lcd array
  if (LCD_page)
  {
    for (x = 0; x < sizeX; x++)
    {
      delete[] LCD_page[x];
    }
    delete[] LCD_page;
  }
  delete oldSetup;
  delete port;
}


int cGraphLCDDriverKS0713::InitGraphic()
{
  Write(REST, true);        // Reset
#if 1
  Write(VOLT | VCOF | VROF | VFOF, true); // disable Voltage Converter, Regulator & Follower
  uSleep(20000);
  Write(ADCS | ANOR, true); // set ADC to normal direction
  Write(SHLS | SNOR, true); // set SHL to normal direction
  Write(BIAS | BIA9, true); // set 1/9 Bias
  Write(VOLT | VCON | VROF | VFOF, true); // enable Voltage Converter
  uSleep(2000);
  Write(VOLT | VCON | VRON | VFOF, true); // enable Voltage Regulator
  uSleep(2000);
  Write(VOLT | VCON | VRON | VFON, true); // enable Voltage Follower
  uSleep(2000);
  Write(REFV, true);        // Set Reference Voltage
  Write(0x00, false);       // to 0x00 
  Write(RRRS | (4&7), true);// Set Regulator Resistor ratio => 5,7V
  // Reference Voltage + Regulator Resistor ratio determine the V0
  // RRRS    REFV=0x00  ... REFV=0x3f
  // 0,0,0     3,00V    ...   3,80V
  // 0,0,1     3,46V    ...   4,38V
  // 0,1,0     4,03V    ...   5,10V
  // 0,1,1     4,77V    ...   6,04V
  // 1,0,0     5,70V    ...   7,22V
  // 1,0,1     6,87V    ...   8,70V
  // 1,1,0     8,36V    ...  10,58V
  // 1,1,1    10,24V    ...  12,96V
  uSleep(20000);
  Write(INVD | IVOF, true); // Set Normal (not Inverted) Mode
#endif
  Write(INDL | 0x00, true); // Set Line 0 as initial Display Line
  Write(SPAD | 0x00, true); // Set Page Address to 0
  Write(SCAM | 0x00, true); // Set Column Address MSB to 0
  Write(SCAL | 0x00, true); // Set Column Address LSB to 0
  Write(DION, true);

  // XXX to test
  for (int y = 0; y < 8; y++)
  {
    Write(SPAD | (y&0x07), true); // Set Page Address
    Write(SCAM | 0x00, true); // Set Column Address MSB to 0
    Write(SCAL | 0x00, true); // Set Column Address LSB to 0
    for(int x = 0; x < 128; x++)
    {
      Write(x, false);
    }
  }

  return 0;
}


int cGraphLCDDriverKS0713::Init()
{
  int   i;
  struct timeval tv1, tv2;

  if (!setup->PPDevice)
  {
    // use DirectIO
    if (port->Open(setup->PPPort) != 0)
      return -1;
    uSleep(10);
  }
  else
  {
    // use ppdev
    if (port->Open(setup->PPDevice) != 0)
      return -1;
  }

  if (nSleepInit() != 0)
  {
    dsyslog("graphlcd plugin: INFO: cannot change wait parameters (cGraphLCDDriver::Init)\n");
    useSleepInit = false;
  }
  else
  {
    useSleepInit = true;
  }

  dsyslog("graphlcd plugin: benchmark started.\n");
  gettimeofday(&tv1, 0);
  for (i = 0; i < 1000; i++)
  {
    port->WriteData(i % 0x100);
  }
  gettimeofday(&tv2, 0);
  if (useSleepInit)
    nSleepDeInit();
  timeForPortCmdInNs = ((tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec));
  dsyslog("graphlcd plugin: benchmark stopped. Time for Port Command: %ldns\n", timeForPortCmdInNs);

  // initialize graphic mode
  InitGraphic();

  port->Release();

  oldSetup->CopyFrom(setup);

  // clear display
  Clear();
  Refresh(true);

  isyslog("graphlcd plugin: KS0713 initialized.\n");
  return 0;
}


int cGraphLCDDriverKS0713::DeInit()
{
  if (port->Close() != 0)
    return -1;
  return 0;
}


void cGraphLCDDriverKS0713::Write(byte data, bool C_nD)
{
  if (useSleepInit)
    nSleepInit();

  if (C_nD)
  { // command
    // set RS high (instruction) and E high
    port->WriteControl(RSHI | CEHI);
    nSleep(100 - timeForPortCmdInNs + 100 * setup->AdjustTiming);

    // Output the command
    port->WriteData(data);
    nSleep(100 - timeForPortCmdInNs + 100 * setup->AdjustTiming);

    // set RS high (instruction) and E low
    port->WriteControl(RSHI | CELO);
    nSleep(200 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
  }
  else
  { // data
    // set RS low (data) and E high
    port->WriteControl(RSLO | CEHI);
    nSleep(100 - timeForPortCmdInNs + 100 * setup->AdjustTiming);

    // Output the data
    port->WriteData(data);
    nSleep(100 - timeForPortCmdInNs + 100 * setup->AdjustTiming);

    // set RS low (data) and E low
    port->WriteControl(RSLO | CELO);
    nSleep(200 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
  }

  if (useSleepInit)
    nSleepDeInit();
}


void cGraphLCDDriverKS0713::CheckSetup()
{
  if (setup->DiffersImportantlyWith(oldSetup))
  {
    DeInit();
    Init();
  }
}


void cGraphLCDDriverKS0713::Clear(void)
{
  for (int x = 0; x < (sizeX + 7) / 8; x++)
    memset(LCD[x], 0, sizeY);
}


void cGraphLCDDriverKS0713::SetPixel(int x, int y, eColor color)
{
  byte c;

  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  c = 0x01;
  if (setup->UpsideDown)
  {
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
  }

  c = c << (x % 8);
  if (color == colorBlack)
  {
    LCD[x / 8][y] = LCD[x / 8][y] | c;
  }
  else
  {
    LCD[x / 8][y] = LCD[x / 8][y] & ~c;
  }
}


void cGraphLCDDriverKS0713::Set8Pixels(int x, int y, byte data, eColor color)
{
  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  if (!setup->UpsideDown)
  {
    // normal orientation
    if (color == colorBlack)
    {
      LCD[x / 8][y] = LCD[x / 8][y] | ReverseBits(data);
    }
    else
    {
      LCD[x / 8][y] = LCD[x / 8][y] & ReverseBits(~data);
    }
  }
  else
  {
    // upside down orientation
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
    if (color == colorBlack)
    {
      LCD[x / 8][y] = LCD[x / 8][y] | data;
    }
    else
    {
      LCD[x / 8][y] = LCD[x / 8][y] & ~data;
    }
  }
}


void cGraphLCDDriverKS0713::Refresh(bool refreshAll)
{
  int x,y;
  int xx,yy;
  byte dByte, oneBlock[8];

  CheckSetup();

  if (setup->RefreshDisplay > 0)
  {
    refreshCounter = (refreshCounter + 1) % setup->RefreshDisplay;
    if (!refreshAll && !refreshCounter)
      refreshAll = true;
  }

  refreshAll = true; // differential update is not yet supported

  if (refreshAll)
  {
    // draw all

    // convert the linear lcd array to the paged array for the display
    for (y = 0; y < (sizeY + 7) / 8; y++)
    {
      for (x = 0; x < (sizeX + 7) / 8; x++)
      {
        for (yy = 0; yy < 8; yy++)
        {
          oneBlock[yy] = LCD[x][yy + (y * 8)] ^ ((setup->Invert != 0) ? 0xff : 0x00);
        }
        for (xx = 0; xx < 8; xx++)
        {
          dByte = 0;
          for(yy = 0; yy < 8; yy++)
          {
            if (oneBlock[yy] & bitmask[xx])
            {
              dByte += (1 << yy);
            }
          }
          LCD_page[x * 8 + xx][y] = dByte;
        }
      }
    }

    port->Claim();

    // send LCD_page data to display
    // set page and start address
    for (y = 0; y < (sizeY + 7) / 8; y++)
    {
      Write(SPAD | (y & 0x0f), true);
      Write(SCAM | 0x00, true); // Set Column Address MSB to 0
      Write(SCAL | 0x00, true); // Set Column Address LSB to 0

      for (x = 0; x < sizeX; x++)
      {
        Write(LCD_page[x][y], false);
      }
    }

    port->Release();
  }
  else
  {
    // draw only the changed bytes
  }
}
