/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  port.c  -  parallel port class with low level routines
 *
 *  (c) 2004 Andreas Regel <andreas.regel AT powarman de>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <errno.h>
#include <unistd.h>
#include <sys/io.h>
#include <sys/ioctl.h>
#include <linux/ppdev.h>
#include <linux/parport.h>

#include <vdr/tools.h>

#include "port.h"

static inline int port_in(int port)
{
  unsigned char value;
  __asm__ volatile ("inb %1,%0"
                    : "=a" (value)
                    : "d" ((unsigned short)port));
  return value;
}

static inline void port_out(unsigned short int port, unsigned char val)
{
  __asm__ volatile ("outb %0,%1\n"
                    :
                    : "a" (val), "d" (port));
}

cParallelPort::cParallelPort()
: fd(-1),
  port(0),
  usePPDev(false)
{
}

cParallelPort::~cParallelPort()
{
}

int cParallelPort::Open(int portIO)
{
  usePPDev = false;
  port = portIO;

  if (port < 0x400)
  {
    if (ioperm(port, 3, 255) == -1)
    {
      esyslog("graphlcd plugin: ERROR ioperm(0x%X) failed! Err:%s (cParallelPort::Open)\n",
              port, strerror(errno));
      return -1;
    }
  }
  else
  {
    if (iopl(3) == -1)
    {
      esyslog("graphlcd plugin: ERROR iopl failed! Err:%s (cParallelPort::Init)\n",
              strerror(errno));
      return -1;
    }
  }
  return 0;
}

int cParallelPort::Open(char * device)
{
  usePPDev = true;

  fd = open(device, O_RDWR);
  if (fd == -1)
  {
    esyslog("graphlcd plugin: ERROR cannot open %s. Err:%s (cParallelPort::Init)\n",
            device, strerror(errno));
    return -1;
  }

  if (ioctl(fd, PPCLAIM, NULL) == -1)
  {
    esyslog("graphlcd plugin: ERROR cannot claim %s. Err:%s (cParallelPort::Init)\n",
            device, strerror(errno));
    close(fd);
    return -1;
  }

  int mode = PARPORT_MODE_PCSPP;
  if (ioctl(fd, PPSETMODE, &mode) == -1)
  {
    esyslog("graphlcd plugin: ERROR cannot setmode %s. Err:%s (cParallelPort::Init)\n",
            device, strerror(errno));
    close(fd);
    return -1;
  }

  return 0;
}

int cParallelPort::Close()
{
  if (usePPDev)
  {
    if (fd != -1)
    {
      ioctl(fd, PPRELEASE);
      close(fd);
	   fd = -1;
    }
    else
    {
      return -1;
    }
  }
  else
  {
    if (port < 0x400)
    {
      if (ioperm(port, 3, 0) == -1)
      {
        return -1;
      }
    }
    else
    {
      if (iopl(0) == -1)
      {
        return -1;
      }
    }
  }
  return 0;
}

void cParallelPort::Claim()
{
  if (usePPDev)
    ioctl(fd, PPCLAIM);
}

void cParallelPort::Release()
{
  if (usePPDev)
    ioctl(fd, PPRELEASE);
}

void cParallelPort::SetDirection(int direction)
{
  if (usePPDev)
  {
    if (ioctl(fd, PPDATADIR, &direction) == -1)
    {
      perror("ioctl(PPDATADIR)");
      exit(1);
    }
  }
  else
  {
    if (direction == kForward)
      port_out(port + 2, port_in(port + 2) & 0xdf);
    else
      port_out(port + 2, port_in(port + 2) | 0x20);
  }
}

byte cParallelPort::ReadControl()
{
  byte value;

  if (usePPDev)
  {
    if (ioctl(fd, PPRCONTROL, &value) == -1)
    {
      perror("ioctl(PPRCONTROL)");
      exit(1);
    }
  }
  else
  {
    value = port_in(port + 2);
  }

  return value;
}

byte cParallelPort::ReadStatus()
{
  byte value;

  if (usePPDev)
  {
    if (ioctl(fd, PPRSTATUS, &value) == -1)
    {
      perror("ioctl(PPRSTATUS)");
      exit(1);
    }
  }
  else
  {
    value = port_in(port + 1);
  }

  return value;
}

void cParallelPort::WriteControl(byte value)
{
  if (usePPDev)
  {
    if (ioctl(fd, PPWCONTROL, &value) == -1)
    {
      perror("ioctl(PPWCONTROL)");
      exit(1);
    }
  }
  else
  {
    port_out(port + 2, value);
  }
}

byte cParallelPort::ReadData()
{
  byte data;

  if (usePPDev)
  {
    if (ioctl(fd, PPRDATA, &data) == -1)
    {
      perror("ioctl(PPRDATA)");
      exit(1);
    }
  }
  else
  {
    data = port_in(port);
  }

  return data;
}

void cParallelPort::WriteData(byte data)
{
  if (usePPDev)
  {
    if (ioctl(fd, PPWDATA, &data) == -1)
    {
      perror("ioctl(PPWDATA)");
      exit(1);
    }
  }
  else
  {
    port_out(port, data);
  }
}
