/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  sed1330.c  -  SED1330 device
 *
 *  (c) 2003 Roland Praml <praml.roland AT t-online de>
 *   based on: hd61830.c
 *     (c) 2001-2004 Carsten Siebholz <c.siebholz AT t-online de>
 *
 *   changes for Seiko-Epson displays : Mar 2004
 *     (c) 2004 Heinz Gressenberger <heinz.gressenberger AT stmk.gv.at>
 *
 *   init sequence taken from Thomas Baumann's LCD-Test program
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <sys/time.h>
#include <vdr/tools.h>

#include "sed1330.h"
#include "fct.h"


// Commands

// Command Codes, Bytes for Param

#define C_SYSTEMSET		0x40
#define C_SLEEPIN		0x53
#define C_DISPON    	0x59
#define C_DISPOFF   	0x58
#define C_SCROLL    	0x44
#define C_CSRFORM   	0x5D
#define C_CGRAMADR  	0x5C
#define C_CSRDIR_R  	0x4C
#define C_CSRDIR_L  	0x4D
#define C_CSRDIR_U  	0x4E
#define C_CSRDIR_D  	0x4F
#define C_HDOTSCR   	0x5A
#define C_OVLAY     	0x5B
#define C_CSRW      	0x46
#define C_CSRR      	0x47
#define C_MWRITE    	0x42
#define C_MREAD     	0x43

#define M0 0
// 0-internal CG ROM, 1-external CG-ROM/RAM

#define M1 0
// 0 - 32 char CG-RAM, 1 - 64 char

#define M2 0
// 0 - 8x8, 1 - 8x16 matrix in CG-RAM/ROM

#define FX 8
// character-with
#define FY 8
// character-height
#define BPC 1
// byte per character, 1 - FX<=8, 2 - FX=9..16

#define SAD1 0x0000
// startadress first screen

const int kInterface6800 = 0;
const int kInterface8080 = 1;

const int kWiringOriginal = 0;
const int kWiringPowerLCD = 1;
const int kWiringLCDProc  = 2;
const int kWiringTweakers = 3;
const int kWiringYASEDW   = 4;

const byte kOriginalA0HI = kInitHigh;
const byte kOriginalA0LO = kInitLow;
const byte kOriginalRDHI = kStrobeHigh;
const byte kOriginalRDLO = kStrobeLow;
const byte kOriginalWRHI = kAutoHigh;
const byte kOriginalWRLO = kAutoLow;
const byte kOriginalCSHI = kSelectHigh;
const byte kOriginalCSLO = kSelectLow;

const byte kPowerLCDA0HI = kInitHigh;
const byte kPowerLCDA0LO = kInitLow;
const byte kPowerLCDRDHI = kSelectHigh;
const byte kPowerLCDRDLO = kSelectLow;
const byte kPowerLCDWRHI = kStrobeHigh;
const byte kPowerLCDWRLO = kStrobeLow;
const byte kPowerLCDCSHI = kAutoHigh;
const byte kPowerLCDCSLO = kAutoLow;

const byte kLCDProcA0HI = kSelectHigh;
const byte kLCDProcA0LO = kSelectLow;
const byte kLCDProcRDHI = kInitHigh;
const byte kLCDProcRDLO = kInitLow;
const byte kLCDProcWRHI = kAutoHigh;
const byte kLCDProcWRLO = kAutoLow;
const byte kLCDProcCSHI = kStrobeHigh;
const byte kLCDProcCSLO = kStrobeLow;

const byte kTweakersA0HI = kSelectHigh;
const byte kTweakersA0LO = kSelectLow;
const byte kTweakersRDHI = kAutoHigh;
const byte kTweakersRDLO = kAutoLow;
const byte kTweakersWRHI = kInitHigh;
const byte kTweakersWRLO = kInitLow;
const byte kTweakersCSHI = kStrobeHigh;
const byte kTweakersCSLO = kStrobeLow;

const byte kYASEDWA0HI = kAutoHigh;
const byte kYASEDWA0LO = kAutoLow;
const byte kYASEDWRDHI = kInitHigh;
const byte kYASEDWRDLO = kInitLow;
const byte kYASEDWWRHI = kStrobeHigh;
const byte kYASEDWWRLO = kStrobeLow;
const byte kYASEDWCSHI = kSelectHigh;
const byte kYASEDWCSLO = kSelectLow;

cGraphLCDDriverSED1330::cGraphLCDDriverSED1330(cGraphLCDSetup * setup)
: setup(setup)
{
  int x;

  oldSetup = new cGraphLCDSetup();
  oldSetup->CopyFrom(setup);

  port = new cParallelPort();

  sizeX = setup->SizeX;
  if (sizeX < 0)
    sizeX = 0;
  sizeY = setup->SizeY;
  if (sizeY < 0)
    sizeY = 0;

  useSleepInit = false;

  // setup lcd array (wanted state)
  newLCD = new byte*[(sizeX + 7) / 8];
  if (newLCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      newLCD[x] = new byte[sizeY];
      memset(newLCD[x], 0, sizeY);
    }
  }
  // setup lcd array (current state)
  oldLCD = new byte*[(sizeX + 7) / 8];
  if (oldLCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      oldLCD[x] = new byte[sizeY];
      memset(oldLCD[x], 0, sizeY);
    }
  }
  refreshCounter = 0;
  timeForPortCmdInNs = 0;
}


cGraphLCDDriverSED1330::~cGraphLCDDriverSED1330()
{
  int x;

  delete port;

  // free lcd array (wanted state)
  if (newLCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      delete[] newLCD[x];
    }
    delete[] newLCD;
  }
  // free lcd array (current state)
  if (oldLCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      delete[] oldLCD[x];
    }
    delete[] oldLCD;
  }
  delete oldSetup;
}


int cGraphLCDDriverSED1330::InitGraphic()
{
  // initialize setup with two graphic screens
  // most parts taken from Thomas Baumann's LCD-Test program
  int cr;
  int memGraph;
  int sad1l, sad1h, sad2l, sad2h;
  
  cr = (sizeX / FX - 1) * BPC;    
  memGraph = ((cr + 1) * sizeY);  // required memory for a graphic layer
  sad1l = SAD1 & 0xFF;
  sad1h = SAD1 >> 8 & 0xFF; 
  sad2l = ((SAD1 + memGraph) & 0xFF);
  sad2h = ((SAD1 + memGraph) >> 8 & 0xFF);

  WriteCmd(C_SYSTEMSET);
  WriteData(0x30 + M0 + (M1 << 1) + (M2 << 2));
  WriteData(0x80 + (FX - 1));
  WriteData(0x00 + (FY - 1));
  WriteData(cr);   // C/R .. display adresses per line
  WriteData((setup->OscillatorFrequency * 1000 / (70 * sizeY) - 1) / 9);  // TC/R .. , fFR=70Hz
  WriteData(sizeY - 1); // L/F .. display lines per screen
  WriteData(cr + 1);   // adresses per virtual display line, low byte
  WriteData(0x00);   // adresses per virtual display line, high byte
    // currently we don't use virtual screens greater then the display,
    // therefore the high byte should always be zero

  WriteCmd(C_SCROLL);
  WriteData(sad1l);  // low-byte startadress first layer
  WriteData(sad1h);  // high-byte startadress first layer
  WriteData(sizeY); // lines per screen
  WriteData(sad2l);  // low-byte startadress second layer
  WriteData(sad2h);  // high-byte startadress second layer
  WriteData(sizeY); // lines per screen
  WriteData(0x00); // low-byte startadress third layer, not used
  WriteData(0x00); // high-byte startadress third layer, not used
  WriteData(0x00); // low-byte startadress fourth layer, not used
  WriteData(0x00); // high-byte startadress fourth layer, not used

  WriteCmd(C_CSRFORM);
  WriteData(0x00); // cursor with: 1 pixel
  WriteData(0x86); // cursor height: 7 lines, block mode
  
  WriteCmd(C_CSRDIR_R);  // automatic cursor increment to the right

  WriteCmd(C_OVLAY);
  WriteData(0x0C); // two layer composition with Priority-OR

  WriteCmd(C_DISPON);  // display ON with
  WriteData(0x04); // cursor OFF and first layer ON without flashing

  WriteCmd(C_CSRW);
  WriteData(0x00); // initializing cursor adress, low byte
  WriteData(0x00); // high byte

  return 0;
}


int cGraphLCDDriverSED1330::Init()
{
  int x;
  struct timeval tv1, tv2;

  if (setup->Wiring == kWiringOriginal)
  {
    A0HI = kOriginalA0HI;
    A0LO = kOriginalA0LO;
    RDHI = kOriginalRDHI;
    RDLO = kOriginalRDLO;
    WRHI = kOriginalWRHI;
    WRLO = kOriginalWRLO;
    CSHI = kOriginalCSHI;
    CSLO = kOriginalCSLO;
    ENHI = RDHI;
    ENLO = RDLO;
    RWHI = WRHI;
    RWLO = WRLO;
  }
  else if (setup->Wiring == kWiringPowerLCD)
  {
    A0HI = kPowerLCDA0HI;
    A0LO = kPowerLCDA0LO;
    RDHI = kPowerLCDRDHI;
    RDLO = kPowerLCDRDLO;
    WRHI = kPowerLCDWRHI;
    WRLO = kPowerLCDWRLO;
    CSHI = kPowerLCDCSHI;
    CSLO = kPowerLCDCSLO;
    ENHI = RDHI;
    ENLO = RDLO;
    RWHI = WRHI;
    RWLO = WRLO;
  }
  else if (setup->Wiring == kWiringLCDProc)
  {
    A0HI = kLCDProcA0HI;
    A0LO = kLCDProcA0LO;
    RDHI = kLCDProcRDHI;
    RDLO = kLCDProcRDLO;
    WRHI = kLCDProcWRHI;
    WRLO = kLCDProcWRLO;
    CSHI = kLCDProcCSHI;
    CSLO = kLCDProcCSLO;
    ENHI = RDHI;
    ENLO = RDLO;
    RWHI = WRHI;
    RWLO = WRLO;
  }
  else if (setup->Wiring == kWiringTweakers)
  {
    A0HI = kTweakersA0HI;
    A0LO = kTweakersA0LO;
    RDHI = kTweakersRDHI;
    RDLO = kTweakersRDLO;
    WRHI = kTweakersWRHI;
    WRLO = kTweakersWRLO;
    CSHI = kTweakersCSHI;
    CSLO = kTweakersCSLO;
    ENHI = RDHI;
    ENLO = RDLO;
    RWHI = WRHI;
    RWLO = WRLO;
  }
  else if (setup->Wiring == kWiringYASEDW)
  {
    A0HI = kYASEDWA0HI;
    A0LO = kYASEDWA0LO;
    RDHI = kYASEDWRDHI;
    RDLO = kYASEDWRDLO;
    WRHI = kYASEDWWRHI;
    WRLO = kYASEDWWRLO;
    CSHI = kYASEDWCSHI;
    CSLO = kYASEDWCSLO;
    ENHI = RDHI;
    ENLO = RDLO;
    RWHI = WRHI;
    RWLO = WRLO;
  }

  if (!setup->PPDevice)
  {
    // use DirectIO
    if (port->Open(setup->PPPort) != 0)
      return -1;
    uSleep(10);
  }
  else
  {
    // use ppdev
    if (port->Open(setup->PPDevice) != 0)
      return -1;
  }

  if (nSleepInit() != 0)
  {
    dsyslog("graphlcd plugin: INFO: cannot change wait parameters (cGraphLCDDriver::Init)\n");
    useSleepInit = false;
  }
  else
  {
    useSleepInit = true;
  }

  dsyslog("graphlcd plugin: benchmark started.\n");
  gettimeofday(&tv1, 0);
  for (x = 0; x < 1000; x++)
  {
    port->WriteData(x % 0x100);
  }
  gettimeofday(&tv2, 0);
  if (useSleepInit)
    nSleepDeInit();
  timeForPortCmdInNs = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
  dsyslog("graphlcd plugin: benchmark stopped. Time for Command: %ldns\n", timeForPortCmdInNs);

  // initialize graphic mode
  InitGraphic();

  port->Release();

  oldSetup->CopyFrom(setup);

  // clear display
  // The SED1330 can have up to 64k memory. If there is less memory
  // it will be overwritten twice or more times.
  WriteCmd(C_MWRITE);
  for (x = 0; x < 65536; x++)
    WriteData(0x00);

  WriteCmd(C_CSRW);
  WriteData(0x00); // initializing cursor adress, low byte
  WriteData(0x00); // high byte
  
  isyslog("graphlcd plugin: SED1330 initialized.\n");
  return 0;
}


int cGraphLCDDriverSED1330::DeInit()
{
  port->Close();
  return 0;
}


void cGraphLCDDriverSED1330::WriteCmd(byte cmd)
{
//  if (useSleepInit)
//    nSleepInit();

  if (setup->Interface == kInterface6800)
  {
    // set A0 high (instruction), RW low (write) and E low
    port->WriteControl(A0HI | RWLO | ENLO);
//    nSleep(140 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
    // Output the actual command
    port->WriteData(cmd);
//    nSleep(140 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
    // set E high
    port->WriteControl(A0HI | RWLO | ENHI);
//    nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
    // set E low
    port->WriteControl(A0HI | RWLO | ENLO);
//    nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
  }
  else
  {
    // set A0 high (instruction), CS low, RD and WR high
    port->WriteControl(A0HI | CSLO | RDHI | WRHI);
//    nSleep(140 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
    // Output the actual command
    port->WriteData(cmd);
//    nSleep(140 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
    // set WR low
    port->WriteControl(A0HI | CSLO | RDHI | WRLO);
//    nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
    // set WR high
    port->WriteControl(A0HI | CSLO | RDHI | WRHI);
//    nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
  }

//  if (useSleepInit)
//    nSleepDeInit();
}


void cGraphLCDDriverSED1330::WriteData(byte data)
{
//  if (useSleepInit)
//    nSleepInit();

  if (setup->Interface == kInterface6800)
  {
    // set A0 low (data), RW low (write) and E low
    port->WriteControl(A0LO | RWLO | ENLO);
//    nSleep(140 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
    // Output the actual data
    port->WriteData(data);
//    nSleep(140 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
    // set E high
    port->WriteControl(A0LO | RWLO | ENHI);
//    nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
    // set E low
    port->WriteControl(A0LO | RWLO | ENLO);
//    nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
  }
  else
  {
    // set A0 low (data), CS low, RD and WR high
    port->WriteControl(A0LO | CSLO | RDHI | WRHI);
//    nSleep(140 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
    // Output the actual data
    port->WriteData(data);
//    nSleep(140 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
    // set WR low
    port->WriteControl(A0LO | CSLO | RDHI | WRLO);
//    nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
    // set WR high
    port->WriteControl(A0LO | CSLO | RDHI | WRHI);
//    nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
  }

//  if (useSleepInit)
//    nSleepDeInit();
}


void cGraphLCDDriverSED1330::CheckSetup()
{
  if (setup->DiffersImportantlyWith(oldSetup))
  {
    DeInit();
    Init();
  }
}


void cGraphLCDDriverSED1330::Clear()
{
  for (int x = 0; x < (sizeX + 7) / 8; x++)
    memset(newLCD[x], 0, sizeY);
}


void cGraphLCDDriverSED1330::SetPixel(int x, int y, eColor color)
{
  byte c;

  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  if (setup->UpsideDown)
  {
    x = sizeX-1-x;
    y = sizeY-1-y;
  }

  c = 0x80;
  c = c >> (x % 8);
  if (color == colorBlack)
  {
    newLCD[x / 8][y] = newLCD[x / 8][y] | c;
  }
  else
  {
    newLCD[x / 8][y] = newLCD[x / 8][y] & ~c;
  }
}


void cGraphLCDDriverSED1330::Set8Pixels(int x, int y, byte data, eColor color)
{
  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  if (!setup->UpsideDown)
  {
    // normal orientation
    if (color == colorBlack)
    {
      newLCD[x / 8][y] = newLCD[x / 8][y] | data;
    }
    else
    {
      newLCD[x / 8][y] = newLCD[x / 8][y] & ~data;
    }
  }
  else
  {
    // upside down orientation
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
    if (color == colorBlack)
    {
      newLCD[x / 8][y] = newLCD[x / 8][y] | ReverseBits(data);
    }
    else
    {
      newLCD[x / 8][y] = newLCD[x / 8][y] & ReverseBits(~data);
    }
  }
}


void cGraphLCDDriverSED1330::Refresh(bool refreshAll)
{
  int x;
  int y;
  int pos = SAD1;

  CheckSetup();

  if (setup->RefreshDisplay > 0)
  {
    refreshCounter = (refreshCounter + 1) % setup->RefreshDisplay;
    if (!refreshAll && !refreshCounter)
      refreshAll = true;
  }

  port->Claim();

  if(refreshAll)
  {
    // draw all
    // set cursor to startadress
    WriteCmd(C_CSRW);
    WriteData(pos & 0xFF);
    WriteData(pos >> 8);

    for (y = 0; y < sizeY; y++)
    {
      for(x = 0; x < (sizeX + 7) / 8; x++)
      {
        WriteCmd(C_MWRITE); // cursor increments automatically
        WriteData(newLCD[x][y] ^ ((setup->Invert == 0) ? 0xff : 0x00));
        oldLCD[x][y] = newLCD[x][y];
        pos++;
      }
    }
    // and reset RefreshCounter
    refreshCounter = 0;
  }
  else
  {
    // draw only the changed bytes
    bool cs = false;
    for (y = 0; y < sizeY; y++)
    {
      for (x = 0; x < (sizeX + 7) / 8; x++)
      {
        if (newLCD[x][y] != oldLCD[x][y])
        {
          if (!cs)
          {
            WriteCmd(C_CSRW);
    	      WriteData(pos & 0xFF);
    	      WriteData(pos >> 8);
            WriteCmd(C_MWRITE);
            cs = true;
          }
          WriteData(newLCD[x][y] ^ ((setup->Invert == 0) ? 0xff : 0x00));
          oldLCD[x][y] = newLCD[x][y];
        }
        else
        {
          cs = false;
        }
        pos++;
      }
    }
  }
  port->Release();
}

