/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  sed1330.h  -  SED1330 device
 *
 *  (c) 2003 Roland Praml <praml.roland AT t-online de>
 *   based on: hd61830.c
 *     (c) 2001-2004 Carsten Siebholz <c.siebholz AT t-online de>
 *
 *   changes for Seiko-Epson displays : Mar 2004
 *     (c) 2004 Heinz Gressenberger <heinz.gressenberger AT stmk.gv.at>
 *
 *   init sequence taken from Thomas Baumann's LCD-Test program
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

/**
 * The LCD should be connected to the parallel printer port of the PC.
 *
 *
 * The connections are:
 * (Note: This pin number might differ with other displays)
 *                       vv
 *  printerport   LCD          other
 *  -----------   ----------   ----------
 *  GND   (18)    GND   (01)   GND          (Ground)
 *                VDD   (02)   +5V          (V Controller)
 *                V0    (03)   +5V..-22V    (Contrast)
 *  INIT  (16)    RS    (04)                (Register Select)
 *  nLF   (14)    R/W   (05)                (Read/Write)
 *  nSTRB (01)    EN    (06)                (Enable)
 *  D0    (02)    D0    (07)                (DataBit0)
 *  D1    (03)    D1    (08)                (DataBit1)
 *  D2    (04)    D2    (09)                (DataBit2)
 *  D3    (05)    D3    (10)                (DataBit3)
 *  D4    (06)    D4    (11)                (DataBit4)
 *  D5    (07)    D5    (12)                (DataBit5)
 *  D6    (08)    D6    (13)                (DataBit6)
 *  D7    (09)    D7    (14)                (DataBit7)
 *                nCS   (15)   GND          (Chip Select)
 *                nRES  (16)   +5V          (Reset)
 *                VEE   (17)   -22V         (V LCD)
 *                nDO   (18)                (not connected)
 **/

/**
 * For the following Seiko-Epson displays with SED1330 controller:
 * G121C (128x128, 8K RAM), G2446 (240x64, 8K RAM),
 * G242C (240x128, 8K RAM), G321D (320x200, 32K RAM)
 * you should use one of this schemes (the 6800-mode is the default):
 *
 *  MODE6800
 *  printerport  LCD         other       function in 6800-mode
 *  -----------  ----------  ----------  ---------------------
 *               /RES  (01)  +5V         /RES (Reset)
 *  nSTRB (01)   /RD   (02)              E (Enable)
 *  nLF   (14)   /WR   (03)              R/W (Read/Write)
 *               SEL2  (04)  GND         Interface select 2
 *               SEL1  (05)  +5V         Interface select 1
 *               /CS   (06)  GND         /CS (Chip select)
 *  INIT  (16)   A0    (07)              A0 (Access type)
 *  D0    (02)   DB0   (08)              D0 (Data bus 0)
 *  D1    (03)   DB1   (09)              D1 (Data bus 1)
 *  D2    (04)   DB2   (10)              D2 (Data bus 2)
 *  D3    (05)   DB3   (11)              D3 (Data bus 3)
 *  D4    (06)   DB4   (12)              D4 (Data bus 4)
 *  D5    (07)   DB5   (13)              D5 (Data bus 5)
 *  D6    (08)   DB6   (14)              D6 (Data bus 6)
 *  D7    (09)   DB7   (15)              D7 (Data bus 7)
 *               VDD   (16)  +5V         VDD (Logic power supply)
 *  GND   (18)   VSS   (17)  GND         GND (Ground)
 *               V0    (18)  +5V..-22V   V0 (LCD contrast)
 *               VLC   (19)  -22V        VEE (LCD drive voltage)
 *               FGND  (20)  GND         Frame ground
 *
 *  MODE8080
 *  printerport  LCD         other       function in 8080-mode
 *  -----------  ----------  ----------  ---------------------
 *               /RES  (01)  +5V         /RES (Reset)
 *  nSEL  (17)   /RD   (02)              /RD (Read)
 *  nSTRB (01)   /WR   (03)              /WR (Write)
 *               SEL2  (04)  GND         Interface select 2
 *               SEL1  (05)  GND         Interface select 1
 *  nLF   (14)   /CS   (06)              /CS (Chip select)
 *  INIT  (16)   A0    (07)              A0 (Access type)
 *  D0    (02)   DB0   (08)              D0 (Data bus 0)
 *  D1    (03)   DB1   (09)              D1 (Data bus 1)
 *  D2    (04)   DB2   (10)              D2 (Data bus 2)
 *  D3    (05)   DB3   (11)              D3 (Data bus 3)
 *  D4    (06)   DB4   (12)              D4 (Data bus 4)
 *  D5    (07)   DB5   (13)              D5 (Data bus 5)
 *  D6    (08)   DB6   (14)              D6 (Data bus 6)
 *  D7    (09)   DB7   (15)              D7 (Data bus 7)
 *               VDD   (16)  +5V         VDD (Logic power supply)
 *  GND   (18)   VSS   (17)  GND         GND (Ground)
 *               V0    (18)  +5V..-22V   V0 (LCD contrast)
 *               VLC   (19)  -22V        VEE (LCD drive voltage)
 *               FGND  (20)  GND         Frame ground
 *
 * SEL1 on GND selects 8080 mode, SEL1 on VDD selects 6800 mode.
 * Be carefull, the manual to my G2446 shows SEL1 on pin 4 and
 * SEL2 on pin 5.
 * After a lot of grey hair because of a not working display,
 * measuring the connections between controller and connector
 * showed me the changed pinning and after a little soldering
 * pictures on my LCD :-)
 **/
 

#ifndef _LIBGRAPHLCD_SED1330_H_
#define _LIBGRAPHLCD_SED1330_H_

#include "device.h"
#include "port.h"
#include "../setup.h"

class cGraphLCDDriverSED1330 : public cGraphLCDDevice
{
private:
  cParallelPort * port;
  byte ** newLCD; // wanted state
  byte ** oldLCD; // current state
  int refreshCounter;
  long timeForPortCmdInNs;
  cGraphLCDSetup * setup;
  cGraphLCDSetup * oldSetup;
  bool useSleepInit;

  unsigned char A0HI;
  unsigned char A0LO;
  unsigned char RDHI;
  unsigned char RDLO;
  unsigned char ENHI;
  unsigned char ENLO;
  unsigned char WRHI;
  unsigned char WRLO;
  unsigned char RWHI;
  unsigned char RWLO;
  unsigned char CSHI;
  unsigned char CSLO;

  int InitGraphic();
  void WriteCmd(byte cmd);
  void WriteData(byte data);
  void CheckSetup();

public:
  cGraphLCDDriverSED1330(cGraphLCDSetup * setup);
  virtual ~cGraphLCDDriverSED1330();

  virtual int Init();
  virtual int DeInit();

  virtual void Clear();
  virtual void SetPixel(int x, int y, eColor color);
  virtual void Set8Pixels(int x, int y, byte data, eColor color);
  virtual void Refresh(bool refreshAll = false);
};

#endif

