/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  sed1520.c  -  SED1520 device
 *
 *  (c) 2003 andreas 'randy' weinberger <vdr AT smue org>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <sys/time.h>
#include <vdr/tools.h>

#include "sed1520.h"
#include "fct.h"


// commands
const byte kSEAD = 0x00;	// Set (X) Column Address
const byte kSEPA = 0xb8;	// Set (Y) Page Address
const byte kSEDS = 0xc0; 	// Set Display Start Line
const byte kDION = 0xaf;	// Display on
const byte kDIOF = 0xae;	// Display off

// control bits for DirectIO
#define CE1     0x01
#define CE1HI   0x01  // Chip Enable 1 on
#define CE1LO   0x00  // Chip Enable 1 off

const byte kCS1HI = 0x00;	// Chip Select 1
const byte kCS1LO = 0x01;
const byte kCS2HI = 0x04;	// Chip Select 2
const byte kCS2LO = 0x00;
const byte kCDHI = 0x08;	// Command/Data Register Select
const byte kCDLO = 0x00;
const byte kLEDHI = 0x02;	// LED Backlight (not supported currently)
const byte kLEDLO = 0x00;

extern byte bitmask[];

cGraphLCDDriverSED1520::cGraphLCDDriverSED1520(cGraphLCDSetup * setup)
: setup(setup)
{
  int x;

  oldSetup = new cGraphLCDSetup();
  oldSetup->CopyFrom(setup);


  port = new cParallelPort();

  if (!(setup->SizeX % 8) == 0) {
    sizeX = setup->SizeX+(8-(setup->SizeX % 8));
  } else {
    sizeX = setup->SizeX;
  }

  if (!(setup->SizeY % 8) == 0) {
    sizeY = setup->SizeY+(8-(setup->SizeY % 8));
  } else {
    sizeY = setup->SizeY;
  }

  if (sizeX < 0)
    sizeX = 0;
  if (sizeY < 0)
    sizeY = 0;

  SEAD = kSEAD;
  SEPA = kSEPA;
  SEDS = kSEDS;
  DION = kDION;
  DIOF = kDIOF;
  LED  = kLEDHI;
  CDHI = kCDHI;
  CDLO = kCDLO;
  CS1HI = kCS1HI;
  CS1LO = kCS1LO;
  CS2HI = kCS2HI;
  CS2LO = kCS2LO;

  useSleepInit = false;
  // setup linear lcd array
  LCD = new byte*[(sizeX + 7) / 8];
  if (LCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      LCD[x] = new byte[sizeY];
      memset(LCD[x], 0, sizeY);
    }
  }
  // setup the lcd array for the paged sed1520
  LCD_page = new byte*[sizeX];
  if (LCD_page)
  {
    for (x = 0; x < sizeX; x++)
    {
      LCD_page[x] = new byte[(sizeY + 7) / 8];
      memset(LCD_page[x], 0, (sizeY + 7) / 8);
    }
  }
  refreshCounter = 0;
  timeForPortCmdInNs = 0;
}


cGraphLCDDriverSED1520::~cGraphLCDDriverSED1520()
{
  int x;

  // free linear lcd array
  if (LCD)
  {
    for (x = 0; x < (sizeX + 7) / 8; x++)
    {
      delete[] LCD[x];
    }
    delete[] LCD;
  }
  // free the lcd array for the paged sed1520
  if (LCD_page)
  {
    for (x = 0; x < sizeX; x++)
    {
      delete[] LCD_page[x];
    }
    delete[] LCD_page;
  }
  delete oldSetup;
  delete port;
}


int cGraphLCDDriverSED1520::InitGraphic()
{
  // initialize controller1, set display start 0, set page 0, set y address 0, display on
  SED1520Cmd(SEDS, 1);
  SED1520Cmd(SEPA, 1);
  SED1520Cmd(SEAD, 1);
  SED1520Cmd(DION, 1);

  // initialize controller2, set display start 0, set page 0, set y address 0, display on
  SED1520Cmd(SEDS, 2);
  SED1520Cmd(SEPA, 2);
  SED1520Cmd(SEAD, 2);
  SED1520Cmd(DION, 2);

  return 0;
}


int cGraphLCDDriverSED1520::Init()
{
  int i;
  struct timeval tv1, tv2;

  if (!setup->PPDevice)
  {
    // use DirectIO
    if (port->Open(setup->PPPort) != 0)
      return -1;
    uSleep(10);
  }
  else
  {
    // use ppdev
    if (port->Open(setup->PPDevice) != 0)
      return -1;
  }

  if (nSleepInit() != 0)
  {
    dsyslog("graphlcd plugin: INFO: cannot change wait parameters (cGraphLCDDriver::Init)\n");
    useSleepInit = false;
  }
  else
  {
    useSleepInit = true;
  }

  dsyslog("graphlcd plugin: benchmark started.\n");
  gettimeofday(&tv1, 0);
  for (i = 0; i < 1000; i++)
  {
    port->WriteData(i % 0x100);
  }
  gettimeofday(&tv2, 0);
  if (useSleepInit)
    nSleepDeInit();
  timeForPortCmdInNs = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);

  dsyslog("graphlcd plugin: benchmark stopped. Time for Command: %ldns\n", timeForPortCmdInNs);

  // initialize graphic mode
  InitGraphic();

  port->Release();

  oldSetup->CopyFrom(setup);

  // clear display
  Clear();
  Refresh(true);

  dsyslog("graphlcd plugin: SED1520 initialized.\n");
  return 0;
}


int cGraphLCDDriverSED1520::DeInit()
{
  if (port->Close() != 0)
    return -1;
  return 0;
}

void cGraphLCDDriverSED1520::SED1520Cmd(byte data, int cmdcs)
{
  if (useSleepInit)
    nSleepInit();

  switch (cmdcs)
  {
    case 1:
      port->WriteControl(CDHI | CS1LO | CS2LO | LEDHI);
      nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
      port->WriteData(data);
      nSleep(650 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
      port->WriteControl(CDHI | CS1HI | CS2LO | LEDHI);
      nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
      break;
    case 2:
      port->WriteControl(CDHI | CS1LO | CS2LO | LED);
      nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
      port->WriteData(data);
      nSleep(650 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
      port->WriteControl(CDHI | CS1LO | CS2HI | LED);
      nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
      break;
  }
  if (useSleepInit)
    nSleepDeInit();
}

void cGraphLCDDriverSED1520::SED1520Data(byte data, int datacs)
{
  if (useSleepInit)
    nSleepInit();

  switch (datacs)
  {
    case 1:
      port->WriteControl(CDLO | CS1LO | CS2LO | LEDHI);
      nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
      port->WriteData(data);
      nSleep(650 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
      port->WriteControl(CDLO | CS1HI | CS2LO | LEDHI);
      nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
      break;
    case 2:
      port->WriteControl(CDLO | CS1LO | CS2LO | LED);
      nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
      port->WriteData(data);
      nSleep(650 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
      port->WriteControl(CDLO | CS1LO | CS2HI | LED);
      nSleep(450 - timeForPortCmdInNs + 100 * setup->AdjustTiming);
      break;
  }
  if (useSleepInit)
    nSleepDeInit();
}


void cGraphLCDDriverSED1520::CheckSetup()
{
  if (setup->DiffersImportantlyWith(oldSetup))
  {
    DeInit();
    Init();
  }
}


void cGraphLCDDriverSED1520::Clear()
{
  for (int x = 0; x < (sizeX + 7) / 8; x++)
    memset(LCD[x], 0, sizeY);
}


void cGraphLCDDriverSED1520::SetPixel(int x, int y, eColor color)
{
  byte c;

  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  if (setup->UpsideDown)
  {
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
  }

  c = 0x80;
  c = c >> (x % 8);
  if (color == colorBlack)
  {
    LCD[x / 8][y] = LCD[x / 8][y] | c;
  }
  else
  {
    LCD[x / 8][y] = LCD[x / 8][y] & ~c;
  }
}


void cGraphLCDDriverSED1520::Set8Pixels (int x, int y, byte data, eColor color)
{
  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  if (!setup->UpsideDown)
  {
    // normal orientation
    if (color == colorBlack)
    {
      LCD[x / 8][y] = LCD[x / 8][y] | data;
    }
    else
    {
      LCD[x / 8][y] = LCD[x / 8][y] & ~data;
    }
  }
  else
  {
    // upside down orientation
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
    if (color == colorBlack)
    {
      LCD[x / 8][y] = LCD[x / 8][y] | ReverseBits(data);
    }
    else
    {
      LCD[x / 8][y] = LCD[x / 8][y] & ReverseBits(~data);
    }
  }
}


void cGraphLCDDriverSED1520::Refresh(bool refreshAll)
{
  int x,y,xx,yy;
  byte dByte, oneBlock[8];

  CheckSetup();

  if(setup->RefreshDisplay > 0)
  {
    refreshCounter = (refreshCounter + 1) % setup->RefreshDisplay;
    if(!refreshAll && !refreshCounter)
      refreshAll=true;
  }

  refreshAll = true; // differential update is not yet supported

  if (refreshAll)
  {
    // draw all

    // convert the linear lcd array to the paged array for the display
    for (y = 0; y < (sizeY + 7) / 8; y++)
    {
      for (x = 0; x < (sizeX + 7) / 8; x++)
      {
        for (yy = 0; yy < 8; yy++)
        {
          oneBlock[yy] = LCD[x][yy + (y * 8)] ^ ((setup->Invert != 0) ? 0xff : 0x00);
        }
        for (xx = 0; xx < 8; xx++)
        {
          dByte = 0;
          for (yy = 0; yy < 8; yy++)
          {
            if (oneBlock[yy] & bitmask[xx])
            {
              dByte += (1 << yy);
            }
          }
          LCD_page[x * 8 + xx][y] = dByte;
        }
      }
    }

    port->Claim();

    // send lcd_soll data to display, controller 1
    // set page and start address
    for (y = 0; y < (sizeY + 7) / 8; y++)
    {
      SED1520Cmd(SEAD, 1);
      SED1520Cmd(SEPA + y, 1);
      SED1520Data(0x00 ^ ((setup->Invert != 0) ? 0xff : 0x00), 1); // fill first row with zero

      for (x = 0; x < sizeX / 2 + 1; x++)
      {
        SED1520Data(LCD_page[x][y], 1);
      }
      
      SED1520Cmd(SEAD, 2);
      SED1520Cmd(SEPA + y, 2);

      for (x = sizeX / 2; x < sizeX; x++)
      {
        SED1520Data(LCD_page[x][y], 2);
      }

      SED1520Data(0x00 ^ ((setup->Invert != 0) ? 0xff : 0x00), 2); // fill last row with zero
    }
    port->Release();
  }
  else
  {
    // draw only the changed bytes
  }
}
