/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  t6963.c  -  T6963 device
 *
 *  (c) 2003, 2004 Andreas Regel <andreas.regel AT gmx de>
 *   based on: HD61830 device
 *              (c) 2001-2004 Carsten Siebholz
 *             lcdproc 0.5 driver
 *              (c) 2001 Manuel Stahl
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <sys/time.h>
#include <vdr/tools.h>

#include "t6963.h"
#include "fct.h"

// T6963 commands

const byte kSetCursorPointer       = 0x21;
const byte kSetOffsetRegister      = 0x22;
const byte kSetAddressPointer      = 0x24;

const byte kSetTextHomeAddress     = 0x40;
const byte kSetTextArea            = 0x41;
const byte kSetGraphicHomeAddress  = 0x42;
const byte kSetGraphicArea         = 0x43;

const byte kSetMode                = 0x80;
const byte kSetDisplayMode         = 0x90;
const byte kSetCursorPattern       = 0xA0;

const byte kDataWriteInc           = 0xC0;
const byte kDataReadInc            = 0xC1;
const byte kDataWriteDec           = 0xC2;
const byte kDataReadDec            = 0xC3;
const byte kDataWrite              = 0xC4;
const byte kDataRead               = 0xC5;

const byte kAutoWrite              = 0xB0;
const byte kAutoRead               = 0xB1;
const byte kAutoReset              = 0xB2;


// T6963 Parameters

const byte kModeOr                 = 0x00;
const byte kModeXor                = 0x01;
const byte kModeAnd                = 0x03;
const byte kModeTextAttribute      = 0x04;
const byte kModeInternalCG         = 0x00;
const byte kModeExternalCG         = 0x08;

const byte kTextAttributeNormal    = 0x00;
const byte kTextAttributeInverse   = 0x05;
const byte kTextAttributeNoDisplay = 0x03;
const byte kTextAttributeBlink     = 0x08;

const byte kDisplayModeBlink       = 0x01;
const byte kDisplayModeCursor      = 0x02;
const byte kDisplayModeText        = 0x04;
const byte kDisplayModeGraphic     = 0x08;

const word kGraphicBase            = 0x0000;
const word kTextBase               = 0x1500;
const word kCGRAMBase              = 0x1800;


// T6963 Wirings

const int kWiringStandard = 0;
const int kWiringWindows  = 1;
const int kWiringFeegy    = 2;

const byte kStandardWRHI = 0x00;   // 01 / nSTRB 
const byte kStandardWRLO = 0x01;   // 
const byte kStandardRDHI = 0x00;   // 17 / nSELECT
const byte kStandardRDLO = 0x08;   // 
const byte kStandardCEHI = 0x00;   // 14 / nLINEFEED
const byte kStandardCELO = 0x02;   // 
const byte kStandardCDHI = 0x04;   // 16 / INIT
const byte kStandardCDLO = 0x00;   // 

const byte kWindowsWRHI = 0x00;   // 16 / INIT
const byte kWindowsWRLO = 0x04;   //
const byte kWindowsRDHI = 0x00;   // 14 / nLINEFEED
const byte kWindowsRDLO = 0x02;   //
const byte kWindowsCEHI = 0x00;   // 01 / nSTRB
const byte kWindowsCELO = 0x01;   //
const byte kWindowsCDHI = 0x08;   // 17 / nSELECT
const byte kWindowsCDLO = 0x00;   // 

// feegy == same as windows, but inverted wr/cd
const byte kFeegyWRHI = 0x04;   // 16 / INIT
const byte kFeegyWRLO = 0x00;   // 
const byte kFeegyRDHI = 0x00;   // 14 / nLINEFEED
const byte kFeegyRDLO = 0x02;   // 
const byte kFeegyCEHI = 0x00;   // 01 / nSTRB
const byte kFeegyCELO = 0x01;   // 
const byte kFeegyCDHI = 0x00;   // 17 / nSELECT
const byte kFeegyCDLO = 0x08;   // 



cGraphLCDDriverT6963::cGraphLCDDriverT6963(cGraphLCDSetup * setup)
: setup(setup)
{
  oldSetup = new cGraphLCDSetup();
  oldSetup->CopyFrom(setup);

  port = new cParallelPort();
  
  sizeX = setup->SizeX;
  if (sizeX < 0)
    sizeX = 0;
  sizeY = setup->SizeY;
  if (sizeY < 0)
    sizeY = 0;

  refreshCounter = 0;
  displayMode = 0;
  bidirectLPT = 1;
  autoWrite = false;
}


cGraphLCDDriverT6963::~cGraphLCDDriverT6963()
{
  delete port;
  delete oldSetup;
}


int cGraphLCDDriverT6963::Init()
{
  int x;

  FS = setup->FontSize == 0 ? 6 : 8;

  if (setup->Wiring == kWiringStandard)
  {
    WRHI = kStandardWRHI;
    WRLO = kStandardWRLO;
    RDHI = kStandardRDHI;
    RDLO = kStandardRDLO;
    CEHI = kStandardCEHI;
    CELO = kStandardCELO;
    CDHI = kStandardCDHI;
    CDLO = kStandardCDLO;
  }
  else if (setup->Wiring == kWiringWindows)
  {
    WRHI = kWindowsWRHI;
    WRLO = kWindowsWRLO;
    RDHI = kWindowsRDHI;
    RDLO = kWindowsRDLO;
    CEHI = kWindowsCEHI;
    CELO = kWindowsCELO;
    CDHI = kWindowsCDHI;
    CDLO = kWindowsCDLO;
  }
  else if (setup->Wiring == kWiringFeegy)
  {
    WRHI = kFeegyWRHI;
    WRLO = kFeegyWRLO;
    RDHI = kFeegyRDHI;
    RDLO = kFeegyRDLO;
    CEHI = kFeegyCEHI;
    CELO = kFeegyCELO;
    CDHI = kFeegyCDHI;
    CDLO = kFeegyCDLO;
  }

  // setup lcd array (wanted state)
  newLCD = new byte*[(sizeX + (FS - 1)) / FS];
  if (newLCD)
  {
    for (x = 0; x < (sizeX + (FS - 1)) / FS; x++)
    {
      newLCD[x] = new byte[sizeY];
      memset(newLCD[x], 0, sizeY);
    }
  }
  // setup lcd array (current state)
  oldLCD = new byte*[(sizeX + (FS - 1)) / FS];
  if (oldLCD)
  {
    for (x = 0; x < (sizeX + (FS - 1)) / FS; x++)
    {
      oldLCD[x] = new byte[sizeY];
      memset(oldLCD[x], 0, sizeY);
    }
  }

  if (!setup->PPDevice)
  {
    // use DirectIO
    if (port->Open(setup->PPPort) != 0)
      return -1;
    uSleep(10);
  }
  else
  {
    // use ppdev
    if (port->Open(setup->PPDevice) != 0)
      return -1;
  }

/*
  bool useSleepInit;
  int timeForPortCmdInNs;
  struct timeval tv1, tv2;
  if (nSleepInit() != 0)
  {
    printf("graphlcd plugin: INFO: cannot change wait parameters (cGraphLCDDriver::Init)\n");
    useSleepInit = false;
  }
  else
  {
    useSleepInit = true;
  }

  printf("graphlcd plugin: benchmark started.\n");
  gettimeofday(&tv1, 0);
  for (int i = 0; i < 1000; i++)
  {
    port->WriteData(1 % 0x100);
  }
  gettimeofday(&tv2, 0);
  if (useSleepInit)
    nSleepDeInit();
  timeForPortCmdInNs = (tv2.tv_sec-tv1.tv_sec) * 1000000 + (tv2.tv_usec-tv1.tv_usec);
  printf("graphlcd plugin: benchmark stopped. Time for Port Command: %ldns\n", timeForPortCmdInNs);
*/

  // disable chip
  // disable reading from LCD
  // disable writing to LCD
  // command/status mode
  T6963SetControl(WRHI | CEHI | CDHI | RDHI);
  port->SetDirection(kForward);       // make 8-bit parallel port an output port

  // Test ECP mode
  if (bidirectLPT == 1)
  {
    dsyslog("graphlcd plugin: Testing ECP mode...\n");
    int i = 0;
    int ecp_input;
    port->SetDirection(kReverse);
    for (int i = 0; i < 100; i++)
    {
      T6963SetControl(WRHI | CEHI | CDHI | RDHI);    // wr, ce, cd, rd
      T6963SetControl(WRHI | CELO | CDHI | RDLO);
      T6963SetControl(WRHI | CELO | CDHI | RDLO);
      T6963SetControl(WRHI | CELO | CDHI | RDLO);
      ecp_input = port->ReadData();
      T6963SetControl(WRHI | CEHI | CDHI | RDHI);
      if ((ecp_input & 0x03) == 0x03)
        break;
    }
    port->SetDirection(kForward);
    if (i >= 100)
    {
      dsyslog("graphlcd plugin: ECP mode not working! -> is now disabled\n");
      bidirectLPT = 0;
    }
    else
      dsyslog("graphlcd plugin: working!\n");
  }

  T6963CommandWord(kSetGraphicHomeAddress, kGraphicBase);
  if (sizeX % FS == 0)
    T6963CommandWord(kSetGraphicArea, sizeX / FS);
  else
    T6963CommandWord(kSetGraphicArea, sizeX / FS + 1);

  T6963Command(kSetMode | kModeOr | kModeInternalCG);

  T6963DisplayMode(kDisplayModeText, false);
  T6963DisplayMode(kDisplayModeGraphic, true);
  T6963DisplayMode(kDisplayModeCursor, false);
  T6963DisplayMode(kDisplayModeBlink, false);

  port->Release();

  oldSetup->CopyFrom(setup);

  // clear display
  Clear();
  Refresh(true);

  isyslog("graphlcd plugin: T6963 initialized.\n");
  return 0;
}


int cGraphLCDDriverT6963::DeInit()
{
  int x;

  // free lcd array (wanted state)
  if (newLCD)
  {
    for (x = 0; x < (sizeX + (FS - 1)) / FS; x++)
    {
      delete[] newLCD[x];
    }
    delete[] newLCD;
  }
  // free lcd array (current state)
  if (oldLCD)
  {
    for (x = 0; x < (sizeX + (FS - 1)) / FS; x++)
    {
      delete[] oldLCD[x];
    }
    delete[] oldLCD;
  }

  if (port->Close() != 0)
    return -1;
  return 0;
}


void cGraphLCDDriverT6963::CheckSetup()
{
  if (setup->DiffersImportantlyWith(oldSetup))
  {
    DeInit();
    Init();
  }
}


void cGraphLCDDriverT6963::Clear()
{
  for (int x = 0; x < (sizeX + (FS - 1)) / FS; x++)
    memset(newLCD[x], 0, sizeY);
}


void cGraphLCDDriverT6963::SetPixel(int x, int y, eColor color)
{
  byte c;

  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  if (FS == 6)
    c = 0x20;
  else
    c = 0x80;
  if(setup->UpsideDown)
  {
    x = sizeX - 1 - x;
    y = sizeY - 1 - y;
  }

  c = c >> (x % FS);
  if (color == colorBlack)
  {
    newLCD[x / FS][y] = newLCD[x / FS][y] | c;
  }
  else
  {
    newLCD[x / FS][y] = newLCD[x / FS][y] & ~c;
  }
}


void cGraphLCDDriverT6963::Set8Pixels(int x, int y, byte data, eColor color)
{
  clip(x, 0, sizeX - 1);
  clip(y, 0, sizeY - 1);

  if (FS == 6)
  {
    byte data1 = 0;
    byte data2 = 0;
    byte data3 = 0;

    if (!setup->UpsideDown)
    {
      // normal orientation
      x = x - (x % 8);

      data1 = data >> (2 + (x % 6));
      if (x % 6 == 5)
      {
        data2 = data >> 1;
        data3 = data << 5;
      }
      else
        data2 = data << (4 - (x % 6));

      if (color == colorBlack)
      {
        newLCD[x / 6][y] = newLCD[x / 6][y] | data1;
        if (x / 6 + 1 < (sizeX + 5) / 6)
          newLCD[x / 6 + 1][y] = newLCD[x / 6 + 1][y] | data2;
        if (x / 6 + 2 < (sizeX + 5) / 6)
          if (x % 6 == 5)
            newLCD[x / 6 + 2][y] = newLCD[x / 6 + 2][y] | data3;
      }
      else
      {
        newLCD[x / 6][y] = newLCD[x / 6][y] & ~data1;
        newLCD[x / 6 + 1][y] = newLCD[x / 6 + 1][y] & ~data2;
        if (x % 6 == 5)
          newLCD[x / 6 + 2][y] = newLCD[x / 6 + 2][y] & ~data3;
      }
    }
    else
    {
      // upside down orientation
      x = sizeX - 1 - x;
      y = sizeY - 1 - y;

      x = x - (x % 8);
      data = ReverseBits(data);

      data1 = data >> (2 + (x % 6));
      if (x % 6 == 5)
      {
        data2 = data >> 1;
        data3 = data << 5;
      }
      else
        data2 = data << (4 - (x % 6));

      if (color == colorBlack)
      {
        newLCD[x / 6][y] = newLCD[x / 6][y] | data1;
        newLCD[x / 6 + 1][y] = newLCD[x / 6 + 1][y] | data2;
        if(x % 6 == 5)
          newLCD[x / 6 + 2][y] = newLCD[x / 6 + 2][y] | data3;
      }
      else
      {
        newLCD[x / 6][y] = newLCD[x / 6][y] & ~data1;
        newLCD[x / 6 + 1][y] = newLCD[x / 6 + 1][y] & ~data2;
        if(x % 6 == 5)
          newLCD[x / 6 + 2][y] = newLCD[x / 6 + 2][y] & ~data3;
      }
    }
  }
  else
  {
    if (!setup->UpsideDown)
    {
      if (color == colorBlack)
      {
        newLCD[x / 8][y] = newLCD[x / 8][y] | data;
      }
      else
      {
        newLCD[x / 8][y] = newLCD[x / 8][y] & ~data;
      }
    }
    else
    {
      x = sizeX - 1 - x;
      y = sizeY - 1 - y;
      if (color == colorBlack)
      {
        newLCD[x / 8][y] = newLCD[x / 8][y] | ReverseBits(data);
      }
      else
      {
        newLCD[x / 8][y] = newLCD[x / 8][y] & ReverseBits(~data);
      }
    }
  }
}


void cGraphLCDDriverT6963::Refresh(bool refreshAll)
{
  int x,y;
  int addr = 0;
//  struct timeval tv1, tv2;
  
  CheckSetup();

  if(setup->RefreshDisplay > 0)
  {
    refreshCounter = (refreshCounter + 1) % setup->RefreshDisplay;
    if(!refreshAll && !refreshCounter)
      refreshAll = true;
  }

//  gettimeofday(&tv1, 0);
  port->Claim();
  if (refreshAll)
  {
    // draw all
    T6963CommandWord(kSetAddressPointer, kGraphicBase);
    if (setup->AutoMode)
    {
      T6963Command(kAutoWrite);
      autoWrite = true;
    }
    for (y = 0; y < sizeY; y++)
    {
      for(x = 0; x < (sizeX + (FS - 1)) / FS; x++)
      {
        if (autoWrite)
          T6963Data((newLCD[x][y]) ^ ((setup->Invert == 0) ? 0xff : 0x00));
        else
          T6963CommandByte(kDataWriteInc, (newLCD[x][y]) ^ ((setup->Invert == 0) ? 0xff : 0x00));
        oldLCD[x][y] = newLCD[x][y];
      }
    }
    if (autoWrite)
    {
      T6963Command(kAutoReset);
      autoWrite = false;
    }
    // and reset RefreshCounter
    refreshCounter = 0;

  }
  else
  {
    // draw only the changed bytes

    bool cs = false;
    for (y = 0; y < sizeY; y++)
    {
      for (x = 0; x < (sizeX + (FS - 1)) / FS; x++)
      {
        if (oldLCD[x][y] != newLCD[x][y])
        {
          if (!cs)
          {
            if (sizeX % FS == 0)
              addr = (y * (sizeX / FS)) + x;
            else
              addr = (y * (sizeX / FS + 1)) + x;
            T6963CommandWord(kSetAddressPointer, kGraphicBase + addr);
            if (setup->AutoMode)
            {
              T6963Command(kAutoWrite);
              autoWrite = true;
            }
            cs = true;
          }
          if (autoWrite)
            T6963Data((newLCD[x][y]) ^ ((setup->Invert == 0) ? 0xff : 0x00));
          else
            T6963CommandByte(kDataWriteInc, (newLCD[x][y]) ^ ((setup->Invert == 0) ? 0xff : 0x00));
          oldLCD[x][y] = newLCD[x][y];
        }
        else
        {
          if (autoWrite)
          {
            T6963Command(kAutoReset);
            autoWrite = false;
          }
          cs = false;
        }
      }
    }
    if (autoWrite)
    {
      T6963Command(kAutoReset);
      autoWrite = false;
    }
  }
  port->Release();

//  gettimeofday(&tv2, 0);
//  int timeRefresh = ((tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec));
//  printf("graphlcd plugin: Time for refresh: %ldus\n", timeRefresh);
}


void cGraphLCDDriverT6963::T6963SetControl(byte flags)
{
  unsigned char status = port->ReadControl();
  status &= 0xF0;      // mask 4 bits
  status |= flags;     // add new flags
  port->WriteControl(status);
}


void cGraphLCDDriverT6963::T6963DSPReady()
{
  int input;

  port->SetDirection(kReverse);
  if (bidirectLPT == 1)
  {
    for (int i = 0; i < 10; i++)
    {
      T6963SetControl(WRHI | CEHI | CDHI | RDHI);
      T6963SetControl(WRHI | CELO | CDHI | RDLO);
      input = port->ReadData();
      T6963SetControl(WRHI | CEHI | CDHI | RDHI);
      if (!autoWrite && (input & 3) == 3)
        break;
      if (autoWrite && (input & 8) == 8)
        break;
    }
  }
  else
  {
    T6963SetControl(WRHI | CEHI | CDHI | RDHI);
    T6963SetControl(WRHI | CELO | CDHI | RDLO);
    T6963SetControl(WRHI | CEHI | CDHI | RDHI);
  }
  port->SetDirection(kForward);
}


void cGraphLCDDriverT6963::T6963Data(byte data)
{
  if (setup->StatusCheck)
    T6963DSPReady();
  T6963SetControl(WRHI | CEHI | CDLO | RDHI);        // CD down (data)
  T6963SetControl(WRLO | CELO | CDLO | RDHI);        // CE & WR down
  port->WriteData(data);
  T6963SetControl(WRHI | CEHI | CDLO | RDHI);        // CE & WR up again
  T6963SetControl(WRHI | CEHI | CDHI | RDHI);        // CD up again
}


void cGraphLCDDriverT6963::T6963Command(byte cmd)
{
  if (setup->StatusCheck)
    T6963DSPReady();
  T6963SetControl(WRHI | CEHI | CDHI | RDHI);        // CD up (command)
  T6963SetControl(WRLO | CELO | CDHI | RDHI);        // CE & WR down
  port->WriteData(cmd);
  T6963SetControl(WRHI | CEHI | CDHI | RDHI);        // CE & WR up again
  T6963SetControl(WRHI | CEHI | CDLO | RDHI);        // CD down again
}


void cGraphLCDDriverT6963::T6963CommandByte(byte cmd, byte data)
{
  T6963Data(data);
  T6963Command(cmd);
}


void cGraphLCDDriverT6963::T6963Command2Bytes(byte cmd, byte data1, byte data2)
{
  T6963Data(data1);
  T6963Data(data2);
  T6963Command(cmd);
}


void cGraphLCDDriverT6963::T6963CommandWord(byte cmd, word data)
{
  T6963Data(data % 256);
  T6963Data(data >> 8);
  T6963Command(cmd);
}


void cGraphLCDDriverT6963::T6963DisplayMode(byte mode, bool enable)
{
  if (enable)
    displayMode |= mode;
  else
    displayMode &= ~mode;
  T6963Command(kSetDisplayMode | displayMode);
}

