/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  font.c  -  font handling
 *
 *  (c) 2001-2004 Carsten Siebholz <c.siebholz AT t-online de>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "font.h"
#include "global.h"

#include <vdr/tools.h>

#define FONTFILE_SIGN "FNT1"


#pragma pack(1)
struct tFONTHEADER
{
  char sign[4];  // = FONTFILE_SIGN
  byte totalY;   // Y-size (totally)
  byte sizeY;    // Y-size of 'normal' characters (without overlengths)
  word count;    // number of chars in this file
};
#pragma pack()


cGraphLCDFont::cGraphLCDFont()
{
  for (int i = 0; i < 256; i++)
    characters[i] = NULL;
}


cGraphLCDFont::~cGraphLCDFont()
{
  for (int i = 0; i < 256; i++)
  {
    if (Exists(i))
    {
      if (characters[i]->data)
      {
        delete[] characters[i]->data;
      }
      delete characters[i];
    }
  }
}


bool cGraphLCDFont::Load(const string & fileName)
{
  bool  ret = false;
  FILE * fntFile;

  tFONTHEADER fhdr;
  byte indexArray[256];
  int i;
  unsigned int begData, fontDataSize;
  byte l, r, j, maxw=0;
  word b,b2;

  fntFile = fopen(fileName.c_str(), "rb");
  if (fntFile)
  {
    fread(&fhdr, sizeof(fhdr), 1, fntFile);

    fread(&indexArray, sizeof(indexArray), 1, fntFile);
    begData = sizeof(fhdr) + sizeof(indexArray);
    fontDataSize = 2 * fhdr.totalY;

    for (i = 0; i < 256; i++)
    {
      if (indexArray[i] != 0xFF)
      {
        characters[i] = new tFontChar;
        if (characters[i])
        {
          memset(characters[i], 0, sizeof(tFontChar));
          characters[i]->data = new word[fontDataSize / 2];
          if (characters[i]->data)
          {
            fseek(fntFile, begData + indexArray[i] * fontDataSize, SEEK_SET);
            fread(characters[i]->data, fontDataSize, 1, fntFile);
            characters[i]->h = fhdr.totalY; //XXX ToDo: correct calculation?

            // calculate width of this character
            b = 0;
            for (j = 0; j < fhdr.totalY; j++)
            {
              b = b | characters[i]->data[j];
            }

            b2 = b; l = 99; r = 99;
            for (j = 0; (j < 16) && (l == 99); j++)
            {
              if (b2 & 0x8000)
              {
                l = j;
              }
              b2 = b2 << 1;
            }
            b2 = b;
            for (j = 0; (j < 16) && (r == 99); j++)
            {
              if (b2 & 0x0001)
              {
                r = 15 - j;
              }
              b2 = b2 >> 1;
            }

            if (l == 99)
            {
              // space
              characters[i]->w = SPACEWIDTH;
            }
            else
            {
              // align left
              for (j = 0; j < fhdr.totalY; j++)
              {
                characters[i]->data[j] = characters[i]->data[j] << l;
              }
              characters[i]->w = r - l + 1;
            }
            if (characters[i]->w > maxw)
              maxw = characters[i]->w;
            ret = true;
          }
          else
          {
            delete characters[i];
          }
        }
      }
    }
    this->totalW = maxw;
    this->totalH = fhdr.totalY;
    this->sizeH  = fhdr.sizeY;
    fclose(fntFile);
  }
  return ret;
}


int cGraphLCDFont::WidthC(unsigned char ch)
{
  if (Exists(ch))
    return characters[ch]->w;
  else
    return 0;
}


int cGraphLCDFont::HeightC(unsigned char ch)
{
  if(Exists(ch))
    return characters[ch]->h;
  else
    return 0;
}


int cGraphLCDFont::WidthS(const string & str)
{
  unsigned int i;
  int sum = 0;

  for (i = 0; i < str.length(); i++)
  {
    sum += WidthC(str[i]);
  }
  if (str.length() > 1)
  {
    sum += str.length() - 1;
  }
  return sum;
}


int cGraphLCDFont::WidthS(const string & str, unsigned int len)
{
  unsigned int i;
  int sum = 0;

  for (i = 0; i < str.length() && i < len; i++)
  {
    sum += WidthC(str[i]);
  }
  if (min(str.length(), len) > 1)
  {
    sum += min(str.length(), len) - 1;
  }
  return sum;
}


int cGraphLCDFont::HeightS(const string & str)
{
  unsigned int i;
  int sum = 0;

  for (i = 0; i < str.length(); i++)
    sum = max(sum, HeightC(str[i]));
  return sum;
}


int cGraphLCDFont::HeightS(const string & str, unsigned int len)
{
  unsigned int i;
  int sum = 0;

  for (i = 0; i < str.length() && i < len; i++)
    sum = max(sum, HeightC(str[i]));
  return sum;
}


bool cGraphLCDFont::Exists(unsigned char ch)
{
  if (characters[ch])
    return true;
  return false;
}


word cGraphLCDFont::GetDataLine(const unsigned char ch, int line)
{
  if (Exists(ch) && (line <= Height()))
    return characters[ch]->data[line];
  return 0;
}

