/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  graphlcd.c - main classes
 *
 *  (c) 2001-2004 Carsten Siebholz <c.siebholz AT t-online de>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

// SLOW_IO_BY_JUMPING tells Linux to use jumps instead of a
//   outb(0x80) to slow down the outb(port) a little bit.
#define SLOW_IO_BY_JUMPING
// REALLY_SLOW_IO tells Linux to use the jmp/outb command
//   4 times instead of only 1 time
#undef  REALLY_SLOW_IO


#include <getopt.h>

#include "status.h"
#include "menu.h"
#include "global.h"
#include "i18n.h"

#include <vdr/plugin.h>


static const char *VERSION        = "0.1.1";
static const char *DESCRIPTION    = "Output to graphic LCD";
static const char *MAINMENUENTRY  = NULL;


// default parallel port address
#define DEFAULT_PPPORT	 0x378


class cPluginGraphLCD : public cPlugin
{
private:
  // Add any member variables or functions you may need here.
  cGraphLCDStatus * graphLCDstatus;
  char * cfgDir;

public:
  cPluginGraphLCD();
  virtual ~cPluginGraphLCD();
  virtual const char * Version() { return VERSION; }
  virtual const char * Description() { return tr(DESCRIPTION); }
  virtual const char * CommandLineHelp();
  virtual bool ProcessArgs(int argc, char * argv[]);
  virtual bool Initialize();
  virtual bool Start();
  virtual void Housekeeping();
  virtual const char * MainMenuEntry() { return MAINMENUENTRY; }
  virtual cOsdObject * MainMenuAction();
  virtual cMenuSetupPage * SetupMenu();
  virtual bool SetupParse(const char * Name, const char * Value);
};


cPluginGraphLCD::cPluginGraphLCD()
: graphLCDstatus(NULL),
  cfgDir(NULL)
{
  // Initialize any member variables here.
  // DON'T DO ANYTHING ELSE THAT MAY HAVE SIDE EFFECTS, REQUIRE GLOBAL
  // VDR OBJECTS TO EXIST OR PRODUCE ANY OUTPUT!
}


cPluginGraphLCD::~cPluginGraphLCD()
{
  // Clean up after yourself!
  delete[] cfgDir;
  if (graphLCDstatus)
    delete graphLCDstatus;
}


const char * cPluginGraphLCD::CommandLineHelp()
{
  // Return a string that describes all known command line options.
  return "  -c CTR,   --contr=CTR    LCD-Controller type\n"
         "  -p IOP,   --port=IOP     IO port\n"
         "  -d DEV,   --device=DEV   sets the parport device\n"
         "  -x SZX,   --sizex=SZX    horizontal size in pixels\n"
         "  -y SZY,   --sizey=SZY    vertical   size in pixels\n"
         "  -z Z,     --zoom=Z       zoom = 1 double pixels\n";
}


bool cPluginGraphLCD::ProcessArgs(int argc, char * argv[])
{
  // Implement command line argument processing here if applicable.
  static struct option long_options[] =
  {
    { "contr",    required_argument, NULL, 'c'},
    { "port",     required_argument, NULL, 'p'},
    { "device",   required_argument, NULL, 'd'},
    { "sizex",    required_argument, NULL, 'x'},
    { "sizey",    required_argument, NULL, 'y'},
    { "zoom",     required_argument, NULL, 'y'},
    { NULL}
  };

  int c;
  int option_index = 0;
  int driverCount;
  int driverNumber;
  tDriver * drivers = GetAvailableDrivers(driverCount);
  driverNumber = driverCount;
  while ((c = getopt_long(argc, argv, "c:p:d:x:y:z:", long_options, &option_index)) != -1)
  {
    switch (c)
    {
      case 'c':
        for (int i = 0; i < driverCount; i++)
        {
          if (!strcasecmp(optarg, drivers[i].alias))
          {
            GraphLCDSetup.DriverID = drivers[i].driverID;
            driverNumber = i;
            break;
          }
        }
        break;

      case 'p':
        GraphLCDSetup.PPPort = (unsigned int) strtol(optarg, (char**) NULL, 0);
        break;

      case 'd':
        GraphLCDSetup.PPDevice = strdup(optarg);
        break;

      case 'x':
        GraphLCDSetup.SizeX = atoi(optarg);
        break;

      case 'y':
        GraphLCDSetup.SizeY = atoi(optarg);
        break;

      case 'z':
        GraphLCDSetup.Zoom = atoi(optarg);
        break;

      default:
        return false;
    }
  }

  if (kDriverUnknown == GraphLCDSetup.DriverID ||
      driverNumber == driverCount )
  {
    // no driver specified:
    // => choose SIMLCD
    GraphLCDSetup.DriverID = kDriverSimLCD;
    driverNumber = 0;
  }

  if (!GraphLCDSetup.PPDevice && GraphLCDSetup.PPPort == -1)
  {
    // neither PPDevice nor PPPort given:
    // => choose PPPort and set it to default (0x378)
    GraphLCDSetup.PPPort = DEFAULT_PPPORT;
  }
  
  if (GraphLCDSetup.SizeX == 0)
    GraphLCDSetup.SizeX = drivers[driverNumber].defaultSizeX;
  if (GraphLCDSetup.SizeY == 0)
    GraphLCDSetup.SizeY = drivers[driverNumber].defaultSizeY;

  return true;
}


bool cPluginGraphLCD::Initialize()
{
  // Initialize any background activities the plugin shall perform.
  return true;
}


bool cPluginGraphLCD::Start()
{
  bool ret = false;
  const char * str;

  // Start any background activities the plugin shall perform.
  RegisterI18n(Phrases);

  str = ConfigDirectory(PLUGIN_NAME);
  if (!str)
    return false;

  cfgDir = strdup(str);

  graphLCDstatus = new cGraphLCDStatus();
  if (graphLCDstatus)
  {
    ret = graphLCDstatus->Init(cfgDir) == 0;
  }
  return ret;
}


void cPluginGraphLCD::Housekeeping()
{
  // Perform any cleanup or other regular tasks.
}


cOsdObject * cPluginGraphLCD::MainMenuAction()
{
  // Perform the action when selected from the main VDR menu.
  return NULL;
}


cMenuSetupPage * cPluginGraphLCD::SetupMenu()
{
  // Return a setup menu in case the plugin supports one.
  return new cGraphLCDMenuSetup();
}


bool cPluginGraphLCD::SetupParse(const char * Name, const char * Value)
{
  // Parse your own setup parameters and store their values.
  if     (!strcasecmp(Name, "Plugin active"))    GraphLCDSetup.PluginActive    = atoi(Value);
  else if(!strcasecmp(Name, "Orientation"))      GraphLCDSetup.UpsideDown      = atoi(Value);
  else if(!strcasecmp(Name, "Display"))          GraphLCDSetup.Invert          = atoi(Value);
  else if(!strcasecmp(Name, "Brightness"))       GraphLCDSetup.Brightness      = atoi(Value);
  else if(!strcasecmp(Name, "RefreshDisplay"))   GraphLCDSetup.RefreshDisplay  = atoi(Value);
  else if(!strcasecmp(Name, "AdjustTiming"))     GraphLCDSetup.AdjustTiming    = atoi(Value);
  else if(!strcasecmp(Name, "WaitMethod"))       GraphLCDSetup.WaitMethod      = atoi(Value);
  else if(!strcasecmp(Name, "WaitPrio"))         GraphLCDSetup.WaitPrio        = atoi(Value);
  else if(!strcasecmp(Name, "Wiring"))           GraphLCDSetup.Wiring          = atoi(Value);
  else if(!strcasecmp(Name, "FontSize"))         GraphLCDSetup.FontSize        = atoi(Value);
  else if(!strcasecmp(Name, "AutoMode"))         GraphLCDSetup.AutoMode        = atoi(Value);
  else if(!strcasecmp(Name, "StatusCheck"))      GraphLCDSetup.StatusCheck     = atoi(Value);
  else if(!strcasecmp(Name, "OscillatorFrequency")) GraphLCDSetup.OscillatorFrequency = atoi(Value);
  else if(!strcasecmp(Name, "Interface")) GraphLCDSetup.Interface = atoi(Value);
  else if(!strcasecmp(Name, "ShowDateTime"))     GraphLCDSetup.ShowDateTime    = atoi(Value);
  else if(!strcasecmp(Name, "ShowChannel"))      GraphLCDSetup.ShowChannel     = atoi(Value);
  else if(!strcasecmp(Name, "ShowLogo"))         GraphLCDSetup.ShowLogo        = atoi(Value);
  else if(!strcasecmp(Name, "ShowSymbols"))      GraphLCDSetup.ShowSymbols     = atoi(Value);
  else if(!strcasecmp(Name, "ShowETSymbols"))    GraphLCDSetup.ShowETSymbols   = atoi(Value);
  else if(!strcasecmp(Name, "ShowProgram"))      GraphLCDSetup.ShowProgram     = atoi(Value);
  else if(!strcasecmp(Name, "ShowTimebar"))      GraphLCDSetup.ShowTimebar     = atoi(Value);
  else if(!strcasecmp(Name, "ShowMenu"))         GraphLCDSetup.ShowMenu        = atoi(Value);
  else if(!strcasecmp(Name, "ShowMessages"))     GraphLCDSetup.ShowMessages    = atoi(Value);
  else if(!strcasecmp(Name, "ShowColorButtons")) GraphLCDSetup.ShowColorButtons= atoi(Value);
  else if(!strcasecmp(Name, "ShowVolume"))       GraphLCDSetup.ShowVolume      = atoi(Value);
  else if(!strcasecmp(Name, "ShowNotRecording")) GraphLCDSetup.ShowNotRecording = atoi(Value);
  else if(!strcasecmp(Name, "IdentifyReplayType")) GraphLCDSetup.IdentifyReplayType = atoi(Value);
  else if(!strcasecmp(Name, "ModifyReplayString")) GraphLCDSetup.ModifyReplayString = atoi(Value);
  else if(!strcasecmp(Name, "ReplayLogo"))       GraphLCDSetup.ReplayLogo = atoi(Value);
  else if(!strcasecmp(Name, "ScrollMode")) GraphLCDSetup.ScrollMode = atoi(Value);
  else if(!strcasecmp(Name, "ScrollSpeed")) GraphLCDSetup.ScrollSpeed = atoi(Value);
  else if(!strcasecmp(Name, "ScrollTime")) GraphLCDSetup.ScrollTime = atoi(Value);
#ifdef INCLUDE_SERDISPLIB
  else if(!strcasecmp(Name, "Contrast"))         GraphLCDSetup.Contrast      = atoi(Value);
  else if(!strcasecmp(Name, "Backlight"))        GraphLCDSetup.Backlight      = atoi(Value);
#endif /* INCLUDE_SERDISPLIB */

  else return false;
  return true;
}

VDRPLUGINCREATOR(cPluginGraphLCD); // Don't touch this!
