/**
 *  GraphLCD plugin for the Video Disk Recorder 
 *
 *  graphlib.c  -  graphic library
 *
 *  (c) 2001-2004 Carsten Siebholz <c.siebholz AT t-online de>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "graphlib.h"
#include "setup.h"
#include "global.h"

#include <vdr/tools.h>


#define SORT(a,b) if ((b) < (a)) {int tmp;tmp = b;b = a;a = tmp;}

byte bitmask[8]  = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
byte bitmaskl[8] = {0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff};
byte bitmaskr[8] = {0xff, 0x7f, 0x3f, 0x1f, 0x0f, 0x07, 0x03, 0x01};


cGraphLCDGraphLib::cGraphLCDGraphLib(cGraphLCDDevice * device)
: pLCDDevice(device)
{
}


int cGraphLCDGraphLib::Abs2(int value)
{
  if (value < 0)
    return -value;
  else
    return value;
}


// Draw a line from (x1,y1) to (x2,y2) in a given color (color)
// No restrictions are placed on these input values. (so x1>x2 is no problem)
// (Bressenham line algorithm uses only integer arithmetic)
void cGraphLCDGraphLib::Line(int x1, int y1, int x2, int y2, eColor color)
{
  int d, sx, sy, dx, dy;
  unsigned int ax, ay;

  clip(x1, 0, pLCDDevice->SizeX() - 1);
  clip(y1, 0, pLCDDevice->SizeY() - 1);
  clip(x2, 0, pLCDDevice->SizeX() - 1);
  clip(y2, 0, pLCDDevice->SizeY() - 1);

  dx = x2 - x1;
  ax = Abs2(dx) << 1;
  if (dx < 0)
    sx = -1;
  else
    sx = 1;

  dy = y2 - y1;
  ay = Abs2(dy) << 1;
  if (dy < 0)
    sy = -1;
  else
    sy = 1;

  pLCDDevice->SetPixel(x1, y1, color);
  if (ax > ay)
  {
    d = ay - (ax >> 1);
    while (x1 != x2)
    {
      if (d >= 0)
      {
        y1 += sy;
        d -= ax;
      }
      x1 += sx;
      d += ay;
      pLCDDevice->SetPixel(x1, y1, color);
    }
  }
  else
  {
    d = ax - (ay >> 1);
    while (y1 != y2)
    {
      if (d >= 0)
      {
        x1 += sx;
        d -= ay;
      }
      y1 += sy;
      d += ax;
      pLCDDevice->SetPixel(x1, y1, color);
    }
  }
}



// Draw a hozizontal line from (x1,y) to (x2,y) in a given color (color)
// No restrictions are placed on these input values. (so x1>x2 is no problem)
void cGraphLCDGraphLib::HLine(int x1, int y, int x2, eColor color)
{
  clip(x1, 0, pLCDDevice->SizeX() - 1);
  clip(y, 0, pLCDDevice->SizeY() - 1);
  clip(x2, 0, pLCDDevice->SizeX() - 1);

  SORT(x1,x2);
  if (x1 / 8 == x2 / 8)
  { // start and end in the same byte
    pLCDDevice->Set8Pixels(x1, y, bitmaskr[x1 % 8] & bitmaskl[x2 % 8], color);
  }
  else
  { // start and end in different bytes
    pLCDDevice->Set8Pixels(x1, y, bitmaskr[x1 % 8], color);
    x1 = ((x1 + 8) / 8) * 8;
    while (x1 < (x2 / 8) * 8)
    {
      pLCDDevice->Set8Pixels(x1, y, 0xff, color);
      x1 += 8;
    }
    pLCDDevice->Set8Pixels(x2, y, bitmaskl[x2 % 8], color);
  }
}


// Draw a vertical line from (x,y1) to (x,y2) in a given color (color)
// No restrictions are placed on these input values. (so x1>x2 is no problem)
void cGraphLCDGraphLib::VLine(int x, int y1, int y2, eColor color)
{
  int y;

  clip(x, 0, pLCDDevice->SizeX() - 1);
  clip(y1, 0, pLCDDevice->SizeY() - 1);
  clip(y2, 0, pLCDDevice->SizeY() - 1);

  SORT(y1,y2);
  for (y = y1; y <= y2; y++)
    pLCDDevice->SetPixel(x, y, color);
}



// Draw a rectangle in a given color (color)
// (x1,y1) defines the upperleft corner, (x2,y2) the lowerright.
// No restrictions are placed on these input values. (so x1>x2 is no problem)
void cGraphLCDGraphLib::Rectangle(int x1, int y1, int x2, int y2, bool filled, eColor color)
{
  int y;

  clip(x1, 0, pLCDDevice->SizeX() - 1);
  clip(y1, 0, pLCDDevice->SizeY() - 1);
  clip(x2, 0, pLCDDevice->SizeX() - 1);
  clip(y2, 0, pLCDDevice->SizeY() - 1);

  SORT(x1,x2);
  SORT(y1,y2);

  if (x1 != x2 && y1 != y2)
  {
    if (!filled)
    {
      HLine(x1, y1, x2, color);
      VLine(x1, y1, y2, color);
      HLine(x1, y2, x2, color);
      VLine(x2, y1, y2, color);
    }
    else
    {
      for (y = y1; y <= y2; y++)
      {
        HLine(x1, y, x2, color);
      }
    }
  }
}


void cGraphLCDGraphLib::RoundEdges(int x1, int y1, int x2, int y2, int type)
{
  switch (type)
  {
    case 0:
      break;

    case 1:
      pLCDDevice->SetPixel(x1, y1, colorWhite);
      pLCDDevice->SetPixel(x2, y1, colorWhite);
      pLCDDevice->SetPixel(x1, y2, colorWhite);
      pLCDDevice->SetPixel(x2, y2, colorWhite);
      break;

    default:
      HLine(x1, y1, x1 + type - 1, colorWhite);
      VLine(x1, y1, y1 + type - 1, colorWhite);

      HLine(x2, y1, x2 - type + 1, colorWhite);
      VLine(x2, y1, y1 + type - 1, colorWhite);

      HLine(x1, y2, x1 + type - 1, colorWhite);
      VLine(x1, y2, y2 - type + 1, colorWhite);

      HLine(x2, y2, x2 - type + 1, colorWhite);
      VLine(x2, y2, y2 - type + 1, colorWhite);
      break;
  }
}


void cGraphLCDGraphLib::RoundRectangle(int x1, int y1, int x2, int y2, bool filled, eColor color, int type)
{
  int y;

  clip(x1, 0, pLCDDevice->SizeX() - 1);
  clip(y1, 0, pLCDDevice->SizeY() - 1);
  clip(x2, 0, pLCDDevice->SizeX() - 1);
  clip(y2, 0, pLCDDevice->SizeY() - 1);

  SORT(x1,x2);
  SORT(y1,y2);

  if (type > (x2 - x1) / 2)
    type = (x2 - x1) / 2;
  if (type > (y2 - y1) / 2)
    type = (y2 - y1) / 2;

  if ((x1 != x2) && (y1 != y2))
  {
    if (filled)
    {
      HLine(x1 + type, y1, x2 - type, color);
      for (y = y1 + 1; y < y1 + type; y++)
        HLine(x1 + 1, y, x2 - 1, color);
      for (y = y1 + type; y <= y2 - type; y++)
        HLine(x1, y, x2, color);
      for (y = y2 - type + 1; y < y2; y++)
        HLine(x1 + 1, y, x2 - 1, color);
      HLine(x1 + type, y2, x2 - type, color);
      if (type == 4)
      {
        // round the ugly fat box...
        pLCDDevice->SetPixel(x1 + 1, y1 + 1, color == colorWhite ? colorBlack : colorWhite);
        pLCDDevice->SetPixel(x1 + 1, y2 - 1, color == colorWhite ? colorBlack : colorWhite);
        pLCDDevice->SetPixel(x2 - 1, y1 + 1, color == colorWhite ? colorBlack : colorWhite);
        pLCDDevice->SetPixel(x2 - 1, y2 - 1, color == colorWhite ? colorBlack : colorWhite);
      }
    }
    else
    {
      HLine(x1 + type, y1, x2 - type, color);
      VLine(x1, y1 + type, y2 - type, color);
      VLine(x2, y1 + type, y2 - type, color);
      HLine(x1 + type, y2, x2 - type, color);
      if (type > 1)
      {
        HLine(x1 + 1, y1 + 1, x1 + type - 1, color);
        HLine(x2 - type + 1, y1 + 1, x2 - 1, color);
        HLine(x1 + 1, y2 - 1, x1 + type - 1, color);
        HLine(x2 - type + 1, y2 - 1, x2 - 1, color);
        VLine(x1 + 1, y1 + 1, y1 + type - 1, color);
        VLine(x1 + 1, y2 - 1, y2 - type + 1, color);
        VLine(x2 - 1, y1 + 1, y1 + type - 1, color);
        VLine(x2 - 1, y2 - 1, y2 - type + 1, color);
      }
    }
  }
}


// Draw a circle with center (x,y) and radius (r) in a given color (color)
// No restrictions are placed on these input values. (so x1>x2 is no problem)
// (Bressenham circle algorithm uses only integer arithmetic)
void cGraphLCDGraphLib::Circle(int x, int y, int r, bool filled, eColor color)
{
  int xx, yy, dd;

  clip(x, 0, pLCDDevice->SizeX() - 1);
  clip(y, 0, pLCDDevice->SizeY() - 1);

  xx = 0;
  yy = r;
  dd = 2 * (1 - r);

  while (yy >= 0)
  {
    if (!filled)
    {
      pLCDDevice->SetPixel(x + xx, y - yy, color);
      pLCDDevice->SetPixel(x + xx, y + yy, color);
      pLCDDevice->SetPixel(x - xx, y - yy, color);
      pLCDDevice->SetPixel(x - xx, y + yy, color);
    }
    else
    {
      HLine(x - xx, y - yy, x + xx, color);
      HLine(x - xx, y + yy, x + xx, color);
    }
    if (dd + yy > 0)
    {
      yy--;
      dd = dd - (2 * yy + 1);
    }
    if (xx > dd)
    {
      xx++;
      dd = dd + (2 * xx + 1);
    }
  }
}


// print a text (text) at position (x,y) with font (pF)
// and given fore- and background colors (fgcolor, bgcolor).
// (xmax) cuts the output horizontally
// if (prop) is true, then the text will be proportional
int cGraphLCDGraphLib::Text(int x, int y, int xmax, const string & text, class cGraphLCDFont * font,
                            bool prop, eColor fgcolor, eColor bgcolor, int skipPixels)
{
  int xt, yt;
  int i;
  unsigned char c;
  int start;

  clip(x, 0, pLCDDevice->SizeX() - 1);
  clip(y, 0, pLCDDevice->SizeY() - 1);

  xt = x;
  yt = y;
  start = 0;

  if (text.length() > 0)
  {
    if (skipPixels > 0)
    {
      if (!prop)
      {
        if (skipPixels >= (int) text.length() * font->Width())
          start = text.length();
        else
          while (skipPixels > font->Width())
          {
            skipPixels -= font->Width();
            start++;
          }
      }
      else
      {
        if (skipPixels >= font->WidthS(text))
          start = text.length();
        else
          while (skipPixels > font->WidthC(text[start]))
          {
            skipPixels -= font->WidthC(text[start]);
            skipPixels--;
            start++;
          }
      }
    }
    for (i = start; i < (int) text.length(); i++)
    {
      c = text[i];
      if (xt > xmax)
      {
        i = text.length();
      }
      else
      {
        if (!prop)
        {
          if (skipPixels > 0)
          {
            Char(xt, yt, xmax, c, font, fgcolor, bgcolor, skipPixels);
            xt += (font->Width()) - skipPixels;
            skipPixels = 0;
          }
          else
          {
            Char(xt, yt, xmax, c, font, fgcolor, bgcolor);
            xt += (font->Width());
          }
        }
        else
        {
          if (skipPixels > 0)
          {
            xt += CharProp(xt, yt, xmax, c, font, fgcolor, bgcolor, skipPixels);
            skipPixels = 0;
          }
          else
          {
            xt += CharProp(xt, yt, xmax, c, font, fgcolor, bgcolor);
          }
          if (xt <= xmax)
          {
            VLine(xt, yt, yt + (font->Height()) - 1, bgcolor);
            xt++;
          }
        }
      }
    }
  }
//printf("x=%d y=%d max=%d text=%s ret=%d\n",x,y,maxwidth,text,xt);
  return xt;
}


// print a fixed font character (c) at position (x,y) with font (pF)
// and given fore- and background colors (fgcolor, bgcolor).
// (xmax) cuts the output horizontally
void cGraphLCDGraphLib::Char(int x, int y, int xmax, unsigned char c, class cGraphLCDFont * font,
                             eColor fgcolor, eColor bgcolor, int skipPixels)
{
  int z;
  word b, d;
  byte data, mask;

  clip(x, 0, pLCDDevice->SizeX() - 1);
  clip(y, 0, pLCDDevice->SizeY() - 1);

  if (font && font->Exists(c))
  {
    // character is defined
    for (z = 0; z < font->HeightC(c); z++)
    {
      b = font->GetDataLine(c, z);
      b = b << skipPixels;
      d = b;
      if (d)
      {
        data = (byte) (d >> 8 + (x % 8));
        mask = 0xff;
        if ((x & 0xfff8) == (xmax & 0xfff8))
        {
          mask = bitmaskl[xmax % 8];
          data = data & mask;
        }
        pLCDDevice->Set8Pixels(x & 0xfff8, y + z, data, fgcolor);
        d = b << (8 - (x % 8));
        d = d >> (8 - (x % 8));
        if (d && (x + 8 < ((pLCDDevice->SizeX() + 7) & 0xfff8)) && mask == 0xff)
        {
          data = (byte) (d >> (x % 8));
          if (((x + 8) & 0xfff8) == (xmax & 0xfff8))
          {
            mask = bitmaskl[xmax % 8];
            data = data & mask;
          }
          pLCDDevice->Set8Pixels((x + 8) & 0xfff8, y + z, data, fgcolor);
          d = (b << (8 - (x % 8))) & 0x00ff;
          d = d >> (8 - (x % 8));
        }
        if (d && (x + 16 < ((pLCDDevice->SizeX() + 7) & 0xfff8)) && mask == 0xff)
        {
          data = (byte) (d << (8 - (x % 8)));
          if (((x + 16) & 0xfff8) == xmax & 0xfff8)
            data = data & bitmaskl[xmax % 8];
          pLCDDevice->Set8Pixels((x + 16) & 0xfff8, y + z, data, fgcolor);
        }
      }
    }
  }
  else
  {
    // character not defined
    dsyslog("graphlcd plugin: char not defined: %c,%d,%x", c, (int) c, c);
  }
}


// print a proportional font character (c) at position (x,y) with font (pF)
// and given fore- and background colors (fgcolor, bgcolor).
// (xmax) cuts the output horizontally
int cGraphLCDGraphLib::CharProp(int x, int y, int xmax, unsigned char c, class cGraphLCDFont * font,
                                eColor fgcolor, eColor bgcolor, int skipPixels)
{
  int z;
  word b, d;
  byte data, mask;

  clip(x, 0, pLCDDevice->SizeX() - 1);
  clip(y, 0, pLCDDevice->SizeY() - 1);

  if (font && font->Exists(c))
  {
    // character is defined
    for (z = 0; z < font->HeightC(c); z++)
    {
      b = font->GetDataLine(c, z);
      b = b << skipPixels;
      d = b;
      if (d)
      {
        data = (byte) (d >> 8 + (x % 8));
        mask = 0xff;
        if ((x & 0xfff8) == (xmax & 0xfff8))
        {
          mask = bitmaskl[xmax % 8];
          data = data & mask;
        }
        pLCDDevice->Set8Pixels(x & 0xfff8, y + z, data, fgcolor);
        d = b << (8 - (x % 8));
        d = d >> (8 - (x % 8));
        if (d && (x + 8 < ((pLCDDevice->SizeX() + 7) & 0xfff8)) && mask == 0xff)
        {
          data = (byte) (d >> (x % 8));
          if (((x + 8) & 0xfff8) == (xmax & 0xfff8))
          {
            mask = bitmaskl[xmax % 8];
            data = data & mask;
          }
          pLCDDevice->Set8Pixels((x + 8) & 0xfff8, y + z, data, fgcolor);
          d = (b << (8 - (x % 8))) & 0x00ff;
          d = d >> (8 - (x % 8));
        }
        if (d && (x + 16 < ((pLCDDevice->SizeX() + 7) & 0xfff8)) && mask == 0xff)
        {
          data = (byte) (d << (8 - (x % 8)));
          if (((x + 16) & 0xfff8) == (xmax & 0xfff8)) data = data & bitmaskl[xmax % 8];
          pLCDDevice->Set8Pixels((x + 16) & 0xfff8, y + z, data, fgcolor);
        }
      }
    }
    return font->WidthC(c) - skipPixels;
  }
  else
  {
    // character not defined
    dsyslog("graphlcd plugin: char not defined: %c,%d,%x", c, (int) c, c);
    return 0;
  }
}


// draws a bitmap (pcBmp) in color (color)
// at the given position (x,y)
void cGraphLCDGraphLib::Bitmap(int x, int y, cGraphLCDLogo * bitmap, eColor color)
{
  byte cl = 0;
  int xt, yt;
  byte * data = (byte *) bitmap->DataPtr();
  word temp;
  int h, w;

  clip(x, 0, pLCDDevice->SizeX() - 1);
  clip(y, 0, pLCDDevice->SizeY() - 1);

  w = bitmap->Width();
  h = bitmap->Height();

  if (data)
  {
    if (!(x % 8))
    {
      // Bitmap is byte alligned (0,8,16,...)
      for (yt = 0; yt < h; yt++)
      {
        for (xt = 0; xt < (w / 8); xt++)
        {
          cl = (byte) * data;
          pLCDDevice->Set8Pixels(x + (xt * 8), y + yt, cl, color);
          data++;
        }
        if (w % 8)
        {
          cl = (byte) *data;
          pLCDDevice->Set8Pixels(x + ((w / 8) * 8), y + yt, cl & bitmaskl[w % 8 - 1], color);
          data++;
        }
      }
    }
    else
    {
      // Bitmap is not byte alligned
      for (yt = 0; yt < h; yt++)
      {
        temp = 0;
        for(xt = 0; xt < (w + (x % 8)) / 8; xt++)
        {
          cl = (byte) *(data + yt * ((w + 7) / 8) + xt);
          temp = temp | ((word) cl << (8 - (x % 8)));
          cl = (temp & 0xff00) >> 8;
          if (!xt)
          {
            // first byte
            pLCDDevice->Set8Pixels(x - (x % 8) + (xt * 8), y + yt, cl & bitmaskr[x % 8], color);
          } else {
            // not the first byte
            pLCDDevice->Set8Pixels(x - (x % 8) + (xt * 8), y + yt, cl, color);
          }
          temp <<= 8;
        }
        if ((w + (x % 8) + 7) / 8 != (w + (x % 8)) / 8)
        {
          // print the rest
          cl = (byte) *(data + (yt + 1) * ((w + 7) / 8) - 1);
          temp = temp | ((word) cl << (8 - (x % 8)));
          cl = (temp & 0xff00) >> 8;
          pLCDDevice->Set8Pixels(x - (x % 8) + (((w + (x % 8)) / 8) * 8), y + yt, cl & bitmaskl[(w + x) % 8 - 1], color);
        }
      }
    }
  }
}

