/**
 *  crtfont.c  -  a tool to create *.fnt files for graphlcd plugin
                  and SiMPeL
 *
 *  (C) 2001 by Carsten Siebholz <c.siebholz@t-online.de>
 **/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program;                                              *
 *   if not, write to the Free Software Foundation, Inc.,                  *
 *   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *
 *                                                                         *
 ***************************************************************************/


#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static const char *prgname = "crtfont";
static const char *version = "0.1.3";


typedef unsigned char   byte;
typedef unsigned short  word;
typedef unsigned int    dword;


#define SORT(a,b) if ((b) < (a)) {int tmp;tmp = b;b = a;a = tmp;}
#define FREENULL(p) (free (p), p = NULL)

#define ABORT       { fclose(InFile); return NULL; }
#define GETANDCHECK { t=fgetc(InFile);if(t==EOF) {fclose(InFile);return false;};}

#define FONTFILE_SIGN "FNT1"
#define MAX_LINELEN		512

#pragma pack(1)
typedef struct tagFONTHEADER {
  char            sign[4];  // = FONTFILE_SIGN
  unsigned char   y_total;  // Y-size (totally)
  unsigned char   y_size;   // Y-size of 'normal' characters (without overlengths)
  unsigned short  count;    // number of chars in this file
} FONTHEADER;
#pragma pack()


struct {
	unsigned short  tag;
	unsigned short  type;
	unsigned long   length;
/*     1 = BYTE.      8-bit unsigned integer.                                                                  */
/*     2 = ASCII.     8-bit bytes that store ASCII codes; the last byte must be null.                          */
/*     3 = SHORT.     A 16-bit (2-byte) unsigned integer.                                                      */
/*     4 = LONG.      A 32-bit (4-byte) unsigned integer.                                                      */
/*     5 = RATIONAL.  Two LONGs: the first represents the numerator of a fraction, the second the denominator. */
	unsigned long   off_val;
} tifftag;

enum PicFormat { undef, TIFF, BMP };
char bitmask[8]  = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
byte bitmaskl[8] = {0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff};

typedef struct {
  int    w,h;
  void*  pData;
} BITMAP;


BITMAP*   LoadTIFF(const char* picName);
BITMAP*   LoadBMP (const char* picName);
void      usage(void);



void    LEndToBEnd ( void* ptr, unsigned short iAnz ) {
	int             i=0;
	unsigned char   c=0;
	unsigned char * pc=(unsigned char*)ptr;

	for( i=0; i<(iAnz/2); i++ ) {
		c=pc[i];
		pc[i]=pc[iAnz-i-1];
		pc[iAnz-i-1]=c;
	}
}




char *trimleft (char *str) {
	char *s = str;
	while (*s==' ' || *s=='\t')
		s++;
	strcpy(str, s);
	return str;
}

char *trimright (char *str) {
	char *s = str + strlen(str) - 1;
	while (s>=str && (*s==' ' || *s=='\t' || *s=='\n' || *s=='\r'))
		*s-- = 0;
	return str;
}

char *trim (char *str) {
	return trimleft(trimright(str));
}





int main(int argc, char *argv[]) {
  int         iRet = 0;
  PicFormat   picFormat = undef;
  FONTHEADER  fhdr;
  unsigned char Index_Array [256];
  char*       picName = NULL;
  char*       fonName = NULL;
  char*       fntName = NULL;
  FILE       *fonFile, *fntFile;
  char        szLine[MAX_LINELEN];
  int         line=0, AnzChars=0;
  char*       pszT;
  int         BegOff,EndOff;
  unsigned char ch, data;
  BITMAP*     pBitmap = NULL;
  unsigned short d,bit;
  int         x,y;
  bool        bError = false;


  memcpy(fhdr.sign, FONTFILE_SIGN, 4);
  memset(&Index_Array, 0xFF, sizeof(Index_Array));


  static struct option long_options[] = {
    { "format",   required_argument, NULL, 'f'},
    { "bmpfile",  required_argument, NULL, 'b'},
    { "descfile", required_argument, NULL, 'd'},
    { "outfile",  required_argument, NULL, 'o'},
    { NULL}
  };


  int c, option_index = 0;
  while((c=getopt_long(argc,argv,"f:b:d:o:",long_options, &option_index))!=-1) {
    switch(c) {
      case 'f':
        if(!strcasecmp(optarg, "TIFF")) {
          picFormat = TIFF;
        }
        if(!strcasecmp(optarg, "BMP")) {
          picFormat = BMP;
        }
        break;

      case 'b':
        picName = strdup(optarg);
        break;

      case 'd':
        fonName = strdup(optarg);
        break;

      case 'o':
        fntName = strdup(optarg);
        break;

      default:
        return 1;
    }
  }


  if(picFormat == undef) {
    fprintf(stderr, "ERROR: You have to specify the format   (-f <format>)\n");
    bError = true;
  }
  if(!picName) {
    fprintf(stderr, "ERROR: You have to specify the bmpfile  (-b bmpfile)\n");
    bError = true;
  }
  if(!fonName) {
    fprintf(stderr, "ERROR: You have to specify the descfile (-d descfile)\n");
    bError = true;
  }
  if(!fntName) {
    fprintf(stderr, "ERROR: You have to specify the outfile  (-o outfile)\n");
    bError = true;
  }

  if(bError) {
    usage();
    return 1;
  }


  fonFile = fopen(fonName,"r");
  if(!fonFile) {
    fprintf(stderr,"Cannot open file: %s\n",fonName);
    return 2;
  }
  fntFile = fopen(fntName,"wb");
  if(!fntFile) {
    fprintf(stderr,"Cannot open file: %s\n",fntName);
    return 2;
  }

  // Load Picture
  switch(picFormat) {
    case TIFF:

      pBitmap = LoadTIFF(picName);
/*       pTIF = Read_TIF(picName);                           */
/*       if(!pTIF) {                                         */
/*         fprintf(stderr,"Cannot open file: %s\n",picName); */
/*         return 2;                                         */
/*       }                                                   */
      break;

    case BMP:
      pBitmap = LoadBMP (picName);
      break;

    default:
      return 2;
  }

  if(!pBitmap) return 3;


  fgets(szLine, sizeof(szLine), fonFile);
  trim(szLine);
  fhdr.y_total = atoi(szLine);
  fgets(szLine, sizeof(szLine), fonFile);
  trim(szLine);
  fhdr.y_size = atoi(szLine);

  fwrite(&fhdr, sizeof(FONTHEADER), 1, fntFile);
  fwrite(&Index_Array, sizeof(Index_Array), 1, fntFile);

  while(!feof(fonFile)) {
    fgets(szLine, sizeof(szLine), fonFile);
    trim(szLine);

    pszT = strtok(szLine, " ");
    if(pszT) {
      BegOff=atoi(pszT);

      // get character
      pszT = strtok(NULL, " ");
      while(pszT) {
        if(strlen(pszT)==1) {
          ch = pszT[0];
        } else {
          ch = atoi(pszT);
        }

        // get EndOffset
        pszT = strtok(NULL, " ");
        EndOff=atoi(pszT);

        for(y=fhdr.y_total*line; y<fhdr.y_total*(line+1); y++) {
          d=0;
          bit=0x8000;
          for(x=BegOff; x<EndOff; x++) {
            data = *((unsigned char*)pBitmap->pData + (y*((pBitmap->w+7)/8)) + (x/8));
            if(data & bitmask [x%8]) d = d | bit;
            bit >>= 1;
          }
          fwrite(&d, sizeof(unsigned short), 1, fntFile);
        }

        BegOff = EndOff;
        Index_Array[(int)ch]=AnzChars;
        AnzChars++;

        // get next character
        pszT = strtok(NULL, " ");

      }
    }
    line++;
  }

  fhdr.count = AnzChars;

  fseek(fntFile,0,SEEK_SET);
  fwrite(&fhdr, sizeof(FONTHEADER), 1, fntFile);
  fwrite(&Index_Array, sizeof(Index_Array), 1, fntFile);

  fclose(fntFile);
  fclose(fonFile);

  fprintf(stdout,"Font '%s' created successfully\n", fntName);

  return iRet;
}


BITMAP*
LoadTIFF(const char* picName) {
  struct {
    unsigned short  tag;
    unsigned short  type;
    unsigned long   length;
    /*     1 = BYTE.      8-bit unsigned integer.                                                                  */
    /*     2 = ASCII.     8-bit bytes that store ASCII codes; the last byte must be null.                          */
    /*     3 = SHORT.     A 16-bit (2-byte) unsigned integer.                                                      */
    /*     4 = LONG.      A 32-bit (4-byte) unsigned integer.                                                      */
    /*     5 = RATIONAL.  Two LONGs: the first represents the numerator of a fraction, the second the denominator. */
    unsigned long   off_val;
  } tifftag;

  BITMAP* pBitmap = NULL;

  FILE *InFile;
  unsigned int tiff_header, tiff_anztags, tiff_data;
  unsigned char cl,ch,y,i;
  unsigned char height, width, strip;
  unsigned char fLittleEndian=0;
  unsigned int  j;
  int t;

  if(picName) {
    InFile = fopen(picName, "rb");
    if(InFile) {

  //    isyslog("graphlcd plugin: try to load logo %s.", picName);
      if(fseek(InFile, 0, SEEK_SET)==EOF) ABORT;
      GETANDCHECK; cl=(unsigned char)t;
      GETANDCHECK; ch=(unsigned char)t;
      if((cl==0x49) && (ch==0x49)) {
        fLittleEndian=1;
      }

      if(fseek(InFile, 4, SEEK_SET)==EOF) ABORT;
      GETANDCHECK; cl=(unsigned char)t;
      GETANDCHECK; ch=(unsigned char)t;
      tiff_header = cl+256*ch;
  //printf("tiff_header:%d %x\n", tiff_header, tiff_header);

      if(fseek(InFile, tiff_header, SEEK_SET)==EOF) ABORT;
      GETANDCHECK; cl=(unsigned char)t;
      GETANDCHECK; ch=(unsigned char)t;
      tiff_anztags = cl+256*ch;
  //printf("tiff_anztags:%d %x\n", tiff_anztags, tiff_anztags);

      height=0;
      width=0;
      strip=0;
      for(i=0; (i<tiff_anztags)&&(!height||!width||!strip); i++) {
        if(fread(&tifftag, sizeof(tifftag), 1, InFile)!=1) ABORT;
        if(tifftag.tag==0x0100) width=tifftag.off_val;
        if(tifftag.tag==0x0101) height=tifftag.off_val;
        if(tifftag.tag==0x0111) strip=tifftag.off_val;
  //printf("tag%d: %d %d %ld %ld\n", i,tifftag.tag, tifftag.type, tifftag.length, tifftag.off_val );
      }

      if(fseek(InFile,strip, SEEK_SET)==EOF) ABORT;
      GETANDCHECK; cl=(unsigned char)t;
      GETANDCHECK; ch=(unsigned char)t;
      tiff_data = cl+256*ch;
  //printf("tiff_data:%d %x\n", tiff_data, tiff_data);

      if(fseek(InFile, tiff_data, SEEK_SET)==EOF) ABORT;


      pBitmap = (BITMAP*)malloc(sizeof(BITMAP));
      if(pBitmap) {
        pBitmap->w = width;
        pBitmap->h = height;
        pBitmap->pData = (unsigned char*)malloc( height*((width+7)/8) );
        if(pBitmap->pData) {
          if(fread(pBitmap->pData, height*((width+7)/8), 1, InFile)!=1) {
            free(pBitmap->pData);
            ABORT;
          }

          for(j=0; j<(unsigned int)height*(((unsigned int)width+7)/8); j++) {
            (*((unsigned char*)pBitmap->pData+j)) = (*((unsigned char*)pBitmap->pData+j))^0xff;
          }
          // cut the rest of the line
          if(width%8) {
            for(y=1; y<=height; y++) {
              j=y*((width+7)/8)-1;
              (*((unsigned char*)pBitmap->pData+j)) = ((*((unsigned char*)pBitmap->pData+j))>>(8-width%8))<<(8-width%8);
            }
          }
        } else {
          fprintf(stderr, "ERROR: cannot allocate memory\n");
        }
      } else {
        fprintf(stderr, "ERROR: cannot allocate memory\n");
      }
      fclose(InFile);
    } else {
      fprintf(stderr, "ERROR: cannot open picture %s\n", picName);
    }
  } else {
    fprintf(stderr, "ERROR: no filename given!\n");
  }

  return pBitmap;
}



BITMAP*
LoadBMP(const char* picName) {

  #pragma pack(1)
  typedef struct BMPH {
    word    bmpIdentifier;
    dword   bmpFileSize;
    dword   bmpReserved;
    dword   bmpBitmapDataOffset;
    dword   bmpBitmapHeaderSize;
    dword   bmpWidth;
    dword   bmpHeight;
    word    bmpPlanes;
    word    bmpBitsPerPixel;
    dword   bmpCompression;
    dword   bmpBitmapDataSize;
    dword   bmpHResolution;
    dword   bmpVResolution;
    dword   bmpColors;
    dword   bmpImportantColors;
  } BMPHEADER;

  typedef struct RGBQ {
      byte    rgbBlue;
      byte    rgbGreen;
      byte    rgbRed;
      byte    rgbReserved;
  } RGBQUAD;
  #pragma pack()

  BITMAP* pBitmap = NULL;

  FILE         *InFile;
  BMPHEADER     bmpHeader;
  RGBQUAD      *pPalette;
  char         *pByte;
  char          Dummy;
  long          iNumColors;
  long          iSize;
  dword         x, y;
  word          iRead;
  bool          bInvert = false;


  if(picName) {
    InFile = fopen(picName, "rb");
    if(InFile) {

      if(fread(&bmpHeader, sizeof(BMPHEADER), 1, InFile)!=1) ABORT;

      // check for Windows BMP
      if(bmpHeader.bmpBitmapHeaderSize != 0x00000028 ) {
        fprintf(stderr, "ERROR: only Windows BMP images are allowed.\n");
        ABORT;
      }

      // check for 2 color
      iNumColors = (1 << bmpHeader.bmpBitsPerPixel);
      if(iNumColors != 2) {
        fprintf(stderr, "ERROR: the image has %ld colors, but only images with 2 colors are allowed.\n", iNumColors);
        ABORT;
      }


      iSize = bmpHeader.bmpHeight * bmpHeader.bmpWidth;

      pPalette = (RGBQUAD *) malloc( iNumColors*sizeof(RGBQUAD));
      if(!pPalette) {
        fprintf(stderr, "ERROR: cannot allocate memory\n");
        ABORT;
      }

      if(fread( pPalette, iNumColors*sizeof(RGBQUAD), 1, InFile)!=1) {
        free(pPalette);
        ABORT;
      }

      // check colors
      if(pPalette->rgbBlue+pPalette->rgbGreen+pPalette->rgbRed <
         (pPalette+1)->rgbBlue+(pPalette+1)->rgbGreen+(pPalette+1)->rgbRed ) {
        // index 0 represents 'black', index 1 'white'
        bInvert = true;
      } else {
        // index 0 represents 'white', index 1 'black'
      }

      pBitmap = (BITMAP*)malloc(sizeof(BITMAP));
      if(!pBitmap) {
        fprintf(stderr, "ERROR: cannot allocate memory\n");
        free(pPalette);
        ABORT;
      }

      if(fseek(InFile, bmpHeader.bmpBitmapDataOffset, SEEK_SET)==EOF) {
        free(pBitmap);
        free(pPalette);
        ABORT;
      }


      switch (bmpHeader.bmpCompression) {
        case 0: // BI_RGB       no compression
          pBitmap->h = bmpHeader.bmpHeight;
          pBitmap->w = bmpHeader.bmpWidth;

          pBitmap->pData = (unsigned char*)malloc( bmpHeader.bmpHeight*((bmpHeader.bmpWidth+7)/8) );
          if(!pBitmap->pData) {
            fprintf(stderr, "ERROR: cannot allocate memory\n");
            free(pBitmap);
            free(pPalette);
            ABORT;
          }

          for(y=bmpHeader.bmpHeight; y>0; y--) {
            pByte = (char*)pBitmap->pData + (y-1)*((bmpHeader.bmpWidth+7)/8);
            iRead = 0;
            for(x=0; x<bmpHeader.bmpWidth/8; x++) {

              if(fread( pByte , sizeof(char), 1, InFile)!=1) {
                free(pBitmap->pData);
                free(pBitmap);
                free(pPalette);
                ABORT;
              }
              iRead++;
              if(bInvert) *pByte = *pByte^0xff;
              pByte++;
            }

            if(bmpHeader.bmpWidth%8) {
              if(fread( pByte , sizeof(char), 1, InFile)!=1) {
                free(pBitmap->pData);
                free(pBitmap);
                free(pPalette);
                ABORT;
              }
              iRead++;
              if(bInvert) *pByte = *pByte^0xff;
              *pByte = *pByte & bitmaskl[bmpHeader.bmpWidth%8];
              pByte++;
            }

            // Scan line must be 4-byte-alligned
            while (iRead%4) {
              if(fread(&Dummy, sizeof(char), 1, InFile)!=1) {
                free(pBitmap->pData);
                free(pBitmap);
                free(pPalette);
                ABORT;
              }
              iRead++;
            }
          }
          break;

        case 1: // BI_RLE4      RLE 4bit/pixel
        case 2: // BI_RLE8      RLE 8bit/pixel
        case 3: // BI_BITFIELDS
        default:
          fprintf(stderr, "ERROR: only uncompressed RGB images are allowed.\n");
          free(pBitmap);
          free(pPalette);
          ABORT;
      }
      fclose(InFile);
    } else {
      fprintf(stderr, "ERROR: cannot open picture %s\n", picName);
    }
  } else {
    fprintf(stderr, "ERROR: no filename given!\n");
  }

  return pBitmap;
}

#if 0
unsigned char* Read_TIF (const char* fn){
	unsigned char* pRet = NULL;
	FILE *InFile;
	unsigned int tiff_header, tiff_anztags, tiff_data;
	unsigned char cl,ch,i;
	unsigned long height=0, width=0, strip=0, xt, yt;
	unsigned char fLittleEndian=0;

	InFile = fopen(fn, "rb");
	if( InFile ) {
		fseek(InFile, 0, SEEK_SET);
		cl=fgetc(InFile);
		ch=fgetc(InFile);
		if( (cl==0x49) && (ch==0x49) ) {
			fLittleEndian=1;
		}

		fseek(InFile, 4, SEEK_SET);
		cl=fgetc(InFile);
		ch=fgetc(InFile);
		tiff_header = cl+256*ch;

		fseek(InFile, tiff_header, SEEK_SET);
		cl=fgetc(InFile);
		ch=fgetc(InFile);
		tiff_anztags = cl+256*ch;

		for( i=0; (i<tiff_anztags)&&(!height||!width||!strip); i++ ) {
			fread(&tifftag, sizeof(tifftag), 1, InFile);
			if( fLittleEndian==0 ) {
				LEndToBEnd (&tifftag.tag    , 2);
				LEndToBEnd (&tifftag.type   , 2);
				LEndToBEnd (&tifftag.length , 4);
				LEndToBEnd (&tifftag.off_val, 4);
			}

			if( tifftag.tag==0x0100 ) width=tifftag.off_val;
			if( tifftag.tag==0x0101 ) height=tifftag.off_val;
			if( tifftag.tag==0x0111 ) strip=tifftag.off_val;
		}

		fseek(InFile,strip, SEEK_SET);
		cl=fgetc(InFile);
		ch=fgetc(InFile);
		tiff_data = cl+256*ch;

		fseek(InFile, tiff_data, SEEK_SET);

		TIF_width=width; TIF_height=height;

		pRet = (unsigned char*) malloc(height*width);
		if(pRet) {
			memset(pRet, 0, height*width);
		
			for( yt=0; yt<height; yt++ ) {
				for( xt=0; xt<width; xt++ ) {
					if( !(xt%8) ) {
						cl=fgetc(InFile);
					}
					if( !(cl&bitmask[xt%8]) ) {	// black in tiff
						*(pRet+(yt*width)+xt)=1;
//						printf("X");
					} else {					// white in tiff
						*(pRet+(yt*width)+xt)=0;
//						printf(".");
					}
				}
//				printf("\n");
			}
		}
		fclose(InFile);
	}
	return pRet;
}
#endif



void
usage(void) {
  fprintf(stdout, "\n");
  fprintf(stdout, "%s v%s\n", prgname, version);
  fprintf(stdout, "%s is a tool to create *.fnt files that are used by the\n", prgname);
  fprintf(stdout, "        graphlcd plugin for VDR and SiMPeL.\n\n");
  fprintf(stdout, "  Usage: %s -f <format> -b bmpfile -d descfile -o outfile\n\n", prgname);
  fprintf(stdout, "  -f  --format      specifies the format of the bitmap. Possible values are:\n");
  fprintf(stdout, "                    TIFF : file is an uncompressed RGB-encoded 2-color TIFF\n" );
  fprintf(stdout, "                    BMP  : file is an uncompressed RGB-encoded 2-color Windows BMP\n" );
  fprintf(stdout, "  -b  --bmpfile     specifies the name of the bitmap file (*.tif, *.bmp)\n");
  fprintf(stdout, "  -d  --descfile    specifies the name of the description file   (*.fon)\n");
  fprintf(stdout, "  -o  --outfile     specifies the name of the output file        (*.fnt)\n");
  fprintf(stdout, "\n" );
  fprintf(stdout, "  example: %s -f BMP -b f12.bmp -d f12.fon -o f12.fnt\n", prgname );
}

